/* Copyright  2016 Vermilion Software ltd.  This software is the confidential and proprietary information of Vermilion Software ltd. (Confidential Information). You are not 
 * permitted to use, copy, disclose or broadcast such Confidential Information, save that Vermilion Software Ltd has permitted certain parties to use it on the terms of certain licence 
 * agreements.  If you are not party to such an agreement, you are requested to report the matter to Vermilion Software Ltd on +44 (0)845 241 9966.  Failure to do so may result in a criminal 
 * offence and legal liability to compensate Vermilion Software ltd.  Vermilion Software ltd will rigorously enforce all its rights in relation to the Confidential Information.
 */
package com.vermilion.vrs.dgs.customprocessors;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.vermilion.vrs.common.DateHelper;
import com.vermilion.vrs.common.StringHelper;
import com.vermilionsoftware.vrs.common.exception.VermilionFatalException;
import com.vermilionsoftware.vrs.common.exception.VermilionNonFatalException;
import com.vermilionsoftware.vrs.common.util.UserConstants;
import com.vermilionsoftware.vrs.core.biz.bd.CoreBD;
import com.vermilionsoftware.vrs.dgs.api.DynamicDGSCall;
import com.vermilionsoftware.vrs.dgs.api.MetaDataField;
import com.vermilionsoftware.vrs.dgs.api.ParameterList;
import com.vermilionsoftware.vrs.dgs.api.ParameterRecord;
import com.vermilionsoftware.vrs.dgs.api.RecordRow;
import com.vermilionsoftware.vrs.dgs.api.RecordSet;
import com.vermilionsoftware.vrs.dgs.api.RecordSetList;
import com.vermilionsoftware.vrs.dgs.api.RecordSetMetaData;
import com.vermilionsoftware.vrs.dgs.api.WarningMessage;
import com.vermilionsoftware.vrs.dgs.api.ri.RIDeleteComponent;
import com.vermilionsoftware.vrs.dgs.api.ri.RIDeleteSection;
import com.vermilionsoftware.vrs.dgs.biz.bd.DgsBD;
import com.vermilionsoftware.vrs.dgs.biz.data.ViewDSTData;
import com.vermilionsoftware.vrs.dgs.biz.data.ViewDSTListData;
import com.vermilionsoftware.vrs.reportdefinition.util.ComponentUtility;
import com.vermilionsoftware.vrs.rg.biz.data.TranslationData;
import com.vermilionsoftware.vrs.rg.util.TranslateXML;

/**
 * @author Vermilion
 *
 *         This is the DragThroughProcessor common set of instructions (DTP).
 *         This provides the methods for the instructions
 *         Please update the version in getName() when you update the golden copy in SVN
 *
 */

public abstract class DragThroughProcessorInstructions extends
		DragThroughProcessorParent{

    /**
     * Returns the name of the custom class.
     *
     * @return String Name of the custom class.
     */
    public String getName() {
        //Change log:
        //3.8.16: dynamic calls on resolveDataVariables check to make sure DS exists first.
    	//3.8.17: new instructions rotateData, rotateDataAddHeaderColumn, and rotateDataRemoveHeaderColumn
    	//        to replace rotation processors.
        //3.8.18: added joinWhere instruction
        //3.8.19: changed resolveDataVariables to cache dynamic calls
        //3.8.20: added data formatting to rotate data
        //3.8.21: fixes to warningIf bug
        //3.8.22: Ordered processor calls by documentation
        //3.8.23: try catch added to getJoinMergeRows to assist any datatype problems
        //3.8.24: fix on getJoinMergeRows when source cells are empty
        //3.8.25: added parameter passing to data substitution variables
        //3.8.26: added value-matching case insensitivity to data variable row selection with new instruction
        //3.8.27: added transformMonthlyMatrix instruction to convert recordset date & return % value (0-1 range) into a year/month matrix of values, plus optional YTD column
        //3.8.28: added concatenateColumn to merge multiple [typically] text records together into one paragraph
        //3.8.29: added column name recognition to setData functionality
        //3.8.30: added concatenateColumnRemoveLinebreaks option (default Y) to concatenateColumn so that line breaks are removed during concatenation
        //3.8.31: removed auto-dragthrough from WITHIN the filterColumn instruction
        //3.8.32: helptext for compareTwoValuesMessage was incorrcetly written for compareMessage, which is not ever read by any module. fixed.
    	//3.8.33: joinWhere fix to (...) OR (....) logic and to resolve auto subsitution variables in the joinWhere instruction before processing
    	//3.8.34: added LEFT OUTER JOIN option for joinWhere, allowing left side datasouces to always return their rows even if the join doesn't link in the other datasources
    	//3.8.35: added setValue instruction, supporting substitution variable (and function) substitution, 
    	//        and changed groupColumns/groupTotals to support any output recordset - just filling out the instructed groupColumns/groupTotals columns
        //3.8.36: change replaceDataVariables to format numbers ONLY if output field is text/html.
    	//3.8.37: fixed joinDuplicates iterator bug

        return "Drag Through Processor (common v3.8.35) (parent:"+getParentName()+")";
    }
    
    public String getRevision() {
		return "$Author: Vermilion $";
	}
    
	protected boolean runInstructions(String key)
			throws VermilionNonFatalException, VermilionFatalException {
		//Try to keep the following in the same order as the documentation. Comments are documentation sections.
		
		//Add Blank Line
			if (key.equalsIgnoreCase("addBlankLine")) {
				processAddBlankLine();
		//Add Total Row
			} else if (key.equalsIgnoreCase("addTotalRow")) {
				processAddTotalRow();
		//Check Opening Balance
			} else if (key.equalsIgnoreCase("checkOpeningBalance")) {
				processCheckOpeningBalance(); //requires documentation
		//Compare Previous Term
			} else if (key.equalsIgnoreCase("comparePreviousTerm")) {
				processComparePreviousTerm();
		//Compare Two Values
			} else if (key.equalsIgnoreCase("compareTwoValues")) {
				processCompareTwoValues(); //requires documentation
		//Count By
			} else if (key.equalsIgnoreCase("countByField")) {
				processCountBy();
		//DragThrough
			} else if (key.equalsIgnoreCase("dragThrough")) {
				processDragThroughRSorDS(getAndRemoveInstruction("dragThrough"));
			} else if (key.equalsIgnoreCase("dragThroughDS")) {
				processDragThroughRSorDS(getAndRemoveInstruction("dragThroughDS"));
			} else if (key.equalsIgnoreCase("dragThroughRS")) {
				processDragThroughRSorDS(getAndRemoveInstruction("dragThroughRS"));
		//Filtering
			} else if (key.equalsIgnoreCase("filterColumn")) {
				processStrFilter();
			} else if (key.equalsIgnoreCase("filterColumnList")) {
				processStrFilterList();
			} else if (key.equalsIgnoreCase("filterRemoveColumn")) {
				processStrFilterRemove();
			} else if (key.equalsIgnoreCase("hideRow")) {
				processHideRow();
		//Force Snap Shots
			} else if (key.equalsIgnoreCase("forceSnapshotFg")
					|| key.equalsIgnoreCase("forceSnapshotIfDSSignedOff")) {
				processForceSnapshotFg();
		//Group Data
			} else if (key.equalsIgnoreCase("groupColumns")) {
				processGroupColumns();
		//Joins
			} else if (key.equalsIgnoreCase("joinWhere")) {
				processJoinWhere();
			} else if (key.equalsIgnoreCase("joinPrimaryDS")) {
				processJoinOnKey();
			} else if (key.equalsIgnoreCase("joinLeftDS")) {
				processJoinWithoutKey(); //requires documentation
			} else if (key.equalsIgnoreCase("joinRightDS")) {
				processJoinWithoutKey(); //requires documentation
			} else if (key.equalsIgnoreCase("mergeRecords")) {
				processMergeRecords();
			} else if (key.equalsIgnoreCase("concatenateColumn")) {
				processConcatenateColumn();
			} else if (key.equalsIgnoreCase("unionDatasources")) {
				processUnionDS();
		//Manual Overrides
			} else if (key.equalsIgnoreCase("manualOverride")) {
				processManualOverride();
		//Paragraph Breaks
			} else if (key.equalsIgnoreCase("paragraphBreaks")
					|| key.equalsIgnoreCase("paragraphBreak")) {
				processParagraphBreaker();	
			} else if (key.equalsIgnoreCase("bulletBreakFlagged")) {
				processBulletBreakFlagged();
		//Rendering Instructions	
			} else if (key.equalsIgnoreCase("removeComponentRowCount")) {
				processRemoveComponentRowCount();
			} else if (key.equalsIgnoreCase("showComponentRowCount")) {
				processShowComponentRowCount();
			} else if (key.equalsIgnoreCase("removeSectionRowCount")) {
				processRemoveSectionRowCount();
			} else if (key.equalsIgnoreCase("showSectionRowCount")) {
				processShowSectionRowCount();
		//Replace Names
			} else if (key.equalsIgnoreCase("replaceNamesDS")) {
				processReplaceNamesDS();
		//Return Specific Rows (by Order)
			} else if (key.equalsIgnoreCase("returnFirstX")) {
				processReturnFirstX();
			} else if (key.equalsIgnoreCase("returnLastX")) {
				processReturnLastX();
			} else if (key.equalsIgnoreCase("returnRowNumX")) {
				processRowNum();
		//Small Percentage Value Rows Removal (the Other Bucket)
			} else if (key.equalsIgnoreCase("smallPercentage")) {
				processSmallPercentage();
		//Sort By Column
			} else if (key.equalsIgnoreCase("sortColumn")) {
				processSortColumn();
		//Strip HTML Tags
			} else if (key.equalsIgnoreCase("stripTags")) {
				processStripTags();
		//Substitute Column
			} else if (key.equalsIgnoreCase("substituteColumn")) {
				processSubstituteColumn();
		//Substitute Variables
			} else if (key.equalsIgnoreCase("resolveSubVarColNm")) {
				processSubstituteValues();
			} else if (key.equalsIgnoreCase("resolveDataVariables")) {
				processResolveDataSubVar();
		//Subtract One Column Value from Another
			} else if (key.equalsIgnoreCase("subtractColumns")) {
				processSubtractColumns();
		//Sum Column Values into Another Column
			} else if (key.equalsIgnoreCase("sumColumns")) {
				processSumColumns();
		//Business Validation Rules (BVRs)
			} else if (key.equalsIgnoreCase("thresholdCheck")) {
				processThresholdCheck();
			} else if (key.equalsIgnoreCase("warningIf")) {
				processWarningIf();
			} else if (key.equalsIgnoreCase("warningIfNot")) {
				processWarningIfNot();
			} else if (key.equalsIgnoreCase("warningIfColSame")) {
				processWarningIfColSame();
			} else if (key.equalsIgnoreCase("warningIfColDiffer")) {
				processWarningIfColDiffer();
			} else if (key.equalsIgnoreCase("warningIfSumDifference")) {
				processWarningIfSumDifference();
			} else if (key.equalsIgnoreCase("warningIfBetween")) {
				processWarningIfBetween();
			} else if (key.equalsIgnoreCase("warningIfNotBetween")) {
				processWarningIfNotBetween();
			} else if (key.equalsIgnoreCase("warningIfRowCount")) {
				processWarningIfRowCount();
			} else if (key.equalsIgnoreCase("warningIfRowCountNot")) {
				processWarningIfRowCountNot();
			} else if (key.equalsIgnoreCase("warningIfSumDiffer")) {
				processWarningIfSumDiffer();
		// Data transformation
			} else if (key.equalsIgnoreCase("transformMonthlyMatrix")) {
				processTransformMonthlyMatrix();
			
		//Not Yet In Documentation
			} else if (key.equalsIgnoreCase("undoHtmlEscapes")){
				processUndoHtmlEscapes();
			} else if (key.equalsIgnoreCase("logShowSignOffStatusFg")) {
				processLogShowSignOffStatusFg();
			} else if (key.equalsIgnoreCase("distinctValues")) {
				processDistinctValues();
			} else if (key.equalsIgnoreCase("addUniqueId")) {
				processAddUniqueId();
			} else if (key.equalsIgnoreCase("loopDSOverEntityRS")) {
				processloopDSOverEntityRS();
			} else if (key.equalsIgnoreCase("suppressWarnings")) {
				suppressWarnings();
			} else if (key.equalsIgnoreCase("setData")) {
				processSetData();
			} else if (key.equalsIgnoreCase("setValue")) {
				processSetValue();
			} else if (key.equalsIgnoreCase("splitIntoBalancedColumns")) {
				processBalanceColumns("split");
			} else if (key.equalsIgnoreCase("wrapIntoBalancedColumns")) {
				processBalanceColumns("wrap");
			} else if (key.equalsIgnoreCase("padRowsToMultiple")) {
				processRowPadding();
			} else if(key.equalsIgnoreCase("rotateData")) {
				processRotateData();
			} else {
				//System.out.println("245:Returning FALSE for runInstructions("+key+")");
				return false;
			}
			
			// When adding new Instruction methods, make sure you include a getAndRemoveInstruction call to the instruction triggering it,
			// otherwise the instruction is repeated x times. 

			//System.out.println("249:Returning TRUE from runInstructions("+key+")");
			return true;
		}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void getHelpText() {
		runPos = "Common getHelpText starting";

		// instruction,	work on output DS, default value, help text
		allParams.put("addBlankLine",					new String[] { "Y", "-",				"Top|Bottom (Will add a blank line in the output as either the 1st record or the last record" });
		allParams.put("addTotalRow",					new String[] { "Y", "-",				"col1,col2...coln (Adds a new row containing the sum of each column specified. The first column will contain 'Total' unless putTotalAt instruction is included." });
		allParams.put("addUniqueId",					new String[] { "Y", "-",				" (Adds a unique key to the output in the 1st column - starting at 1" });
		allParams.put("appendParameters",				new String[] { "Y", "Y",				"Adds an additional column to the output recordset of loopDSOverEntityRS containing string representations of the entities used for the DS calls." });
		allParams.put("breakHeaders",					new String[] { "Y", "Y",				"Using paragraphBreak, breaks headers up. Default is Y" });
		allParams.put("bulletBreakFlagged",				new String[] { "Y", "-",				"Breaks bullet and paragraphs into separate rows and adds a flag (1 or 0) to indicate whether there was a bullet on this line." });
		allParams.put("checkOpeningBalance",			new String[] { "Y", "-",				"openBalanceCol,closeBalanceCol (used to check that the opening balance for this term is equal to the closing balance from the previous term" });
		allParams.put("compareTwoValues",				new String[] { "Y", "",					"col1,val1,val2,col2,threshold (will pickup the row containing val1 and val 2 in col1 and compare the values in col2. "+"If they differ more than the threshold it will display a warning" });
		allParams.put("compareTwoValuesMessage",		new String[] { "Y", "",					"message (if the comparison exceeds the threshold) (available substitution values: $CompareValue1$, $CompareValue2$, $RowTag1$, $RowTag2$, $Difference$, $ThresholdValue$" });
		allParams.put("concatenateColumn",              new String[] { "N", "",                 "system Column Name - used to merge the text contents of the given column into one output record/paragraph"});
		allParams.put("concatenateColumnRemoveLinebreaks", new String[] { "N", "Y",              "Y|N (default Y) remove any line breaks (e.g. <BR> or <P> tags) from the given content"});	
		allParams.put("countByCount",					new String[] { "Y", "-",				"The field name that where the count will be placed" });
		allParams.put("countByField",					new String[] { "Y", "-",				"The field name that will be used to determine when the counter is reset to 1" });
		allParams.put("countByTotal",					new String[] { "Y", "-",				"The field name that where the total will be upticked until the counter is reset to 1." });
		allParams.put("countOnOutput",					new String[] { "Y", "N",				"If Y, countBy instruction will count the output RS (after other DTP instructions); by default countBy looks at the" +"first input recordset." });
		allParams.put("dataVarDynamicCalls",			new String[] { "Y", "N",				"Y|N (if something that looks like a data substitution variable references a non-input RS, try to make a dynamic DGS call for it)" });
		allParams.put("dataVarResolveInside",			new String[] { "Y", "N",				"Y|N (try to resolve any formulas or entity substitution variables that might comprise data substitution variables)" });
		allParams.put("dataVarUseDefault",				new String[] { "Y", "Y",				"Y|N (if row selector argument of a data substitution variable is invalid, use default row selection logic as backup; otherwise, don't replace variable." });
		allParams.put("dataVarValueCaseSensitive", 		new String[] { "Y", "Y",				"Y|N (require row selector argument of data substitution variable to match on case. Default=Y)"});
		allParams.put("distinctValues",					new String[] { "N", "-",				"col (Outputs a distinct set of the values in col to the 2nd column in the output" });
		allParams.put("dragByColumnName",				new String[] { "Y", "",					"Y|N (used to limit dragThrough to columns that match in name, either system or preferred)" });
		allParams.put("dragThrough",					new String[] { "N", "",					"code (used to pinpoint the RST or the DST that should be dragged through" });
		allParams.put("dragThroughDS",					new String[] { "N", "",					"code (used to pinpoint the DST that should be dragged through" });
		allParams.put("dragThroughRS",					new String[] { "N", "",					"code (used to pinpoint the RST that should be dragged through" });
		allParams.put("filterCaseSensitive", 			new String[] { "Y", "N",				"Y|N (if Y, filterColumn and filterRemoveColumn instructions will match only if case matches)"});
		allParams.put("filterColumn",					new String[] { "Y", "",					"# (Check whether a drag-through filter is required (see filterColumn, filterRowCount and filterValue)" });
		allParams.put("filterColumnList",				new String[] { "N", "",					"# (This will allow a csv list of values to be specified in the filterValueList and will return all records that match)" });
		allParams.put("filterIgnoreBlank",				new String[] { "Y", "Y",				"# (used to indicate if the filterColumn should ignore blank entries. If N it will not return rows where this filerColumn or filterRemoveColumn is blank." });		
		allParams.put("filterMatchType",				new String[] { "Y", "EQUALS",			"EQUALS|CONTAINS|REGEX (used to determine by what methodlogy the filterColumn (or filterRemoveColumn) matches the "+"filterValue (or filterRemoveValue) provided."});
		allParams.put("filterRemoveColumn",				new String[] { "Y", "",					"# (Remove any rows with the filterRemoveValue in the filterRemoveColumn" });
		allParams.put("filterRemoveValue",				new String[] { "Y", "",					"# (Remove any rows with the filterRemoveValue in the filterRemoveColumn" });
		allParams.put("filterRowCount",					new String[] { "Y", "-1",				"# (used to limit the number of filtered records dragged through (999=n/a) - see filterColumn/Value" });
		allParams.put("filterValue",					new String[] { "Y", "",					"# (Check whether a drag-through filter is required (see filterColumn, filterRowCount and filterValue). This can take /formatdate, but the month must be in CAPS" });
		allParams.put("forceSnapshotFg", 				new String[] { "Y", "N",				"Y|N (used to allow snapshots to be forced in (otherwise VRS will only do it on the first 'touch')" });
		allParams.put("forceSnapshotIfDSSignedOff",		new String[] { "Y", "N",				"Y|N (Used to allow snapshots to be forced in (otherwise VRS will only do it on the first 'touch')" });
		allParams.put("globalVariableCd",				new String[] { "Y", "",					"# (used to pinpoint which global variable should be used when/if resolving variables" });
		allParams.put("globalVariableKey",				new String[] { "Y", "X",				"# (used to pinpoint which global key should be used when/if resolving variables - i.e. the key/desc field value" });
		allParams.put("groupColumns",					new String[] { "N", "-",				"col1,col2...coln (the columns that the data will be grouped on" });
		allParams.put("groupColumns2",					new String[] { "N", "",					"col1,col2...coln (Optional for use with groupColumns; the columns that the data will be grouped on" });
		allParams.put("groupTotals",					new String[] { "N", "-",				"col1,col2...coln (the columns that will be summed under the group columns" });
		allParams.put("hideRow",						new String[] { "N", "-",				"col1=xx,col2,spreadType (This will hide any records that contain xx in col1 and distribute their values across col2. The distribution will be as per the spreadType of either 'even' or 'percent')" });
		allParams.put("joinWhere",						new String[] { "N", "",  				"Join two or more datasource/recordsets togther; joinWhere= will cartesian join the sets, joinWhere=DS2.field=DS1.field will join two datasources based on field matching; also use the following operators: <, >, =, <>, !=, and, or" });
		allParams.put("joinDuplicates",					new String[] { "N", "N",				"If 'Y', multiple matches in secondary DS will duplicate the primary DS rows (optional)." });
		allParams.put("joinLeftDS",						new String[] { "N", "-",				"code" });
		allParams.put("joinPrimaryDS",					new String[] { "N", "-",				"code|<OUTPUT> (the DST that will be used in the join statement - all records from this DST will be in the resulting DST; if <OUTPUT> is used, other instructions can precede the join)" });
		allParams.put("joinPrimaryKeyColumn",			new String[] { "N", "-",				"col1 (the column in the primary table that will be used to join on" });
		allParams.put("joinResultingColumn",			new String[] { "N", "-",				"col1 (the column in the output DST that will populated with the data from the secondary value column" });
		allParams.put("joinSecondaryDS",				new String[] { "N", "-",				"code (the DST that will be used in the join statement - only records from this DST that match the primary DST will be used" });
		allParams.put("joinSecondaryEntityName",		new String[] { "N", "-",				"entity name (the name of the entity that you want to join on" });
		allParams.put("joinSecondaryKeyColumn",			new String[] { "N", "-",				"col1 (the column in the secondary table that will be used to join on" });
		allParams.put("joinSecondaryValueColumn",		new String[] { "N", "-",				"col1 (the column in the secondary table that will be picked up and output in the resulting DST" });
		allParams.put("joinUnfoundKeysValue",			new String[] { "N", null,				"A value to return for any rows that are not found in secondary DS (optional)." });
		allParams.put("logShowSignOffStatusFg", 		new String[] { "Y", "N",				"Y|N (used to force the sign-off status to be highlighted on the warnings pane" });
		allParams.put("loopDSOverEntityRS",				new String[] { "Y", "-",				"DS_CODE (Input record sets must all be entity-as-a-data-sources. DS_CODE will be run over all combinations of input parameters.)" });
		allParams.put("manualOverride",					new String[] { "N", "-",				"DS_CODE1,DS_CODE2 (This will return the first datasource rows if they are populated - otherwise the second datasource rows (assumes the first is the manual and the second is automatic)" });
		allParams.put("mergeRecords",					new String[] { "N", "-",				" (This will take records in pairs and merge them into 1 output record keeping the 1 column the same" });
		allParams.put("padRowsByGroup",					new String[] { "Y", "-",				"{column name} (If using padRowsToMultiple, this will insert them with the same content in the grouped column)" });
		allParams.put("padRowsLocation",				new String[] { "Y", "After",			"Before|After (If using padRowsToMultiple, this will instruct DTP whether to prepend or append additional rows)" });	
		allParams.put("padRowsToMultiple",				new String[] { "Y", "1",				"# (Adds extra blank rows to recordset until total number of rows is a multiple of # (including 1 * # = #)" });
		allParams.put("paragraphBreak",					new String[] { "Y", "Y",				"Breaks paragraphs into separate rows" });
		allParams.put("paragraphBreaks",				new String[] { "Y", "Y",				"Breaks paragraphs into separate rows" });
		allParams.put("period",							new String[] { "Y", "",					"previous (defines that the previous period is required for the filter)" });
		allParams.put("putTotalAt",						new String[] { "Y", "-",				"colName (Puts the 'Total' label in the named column or no column at all if 'N/A' is used here.) " });
		allParams.put("removeComponentRowCount",		new String[] { "Y", "",					"# (used to remove an associated component if the datasource contains this number of rows" });
		allParams.put("removeFormatting",				new String[] { "Y", "Y",				"Using paragraphBreak, removes formatting. Default is Y." });
		allParams.put("removeSectionRowCount",			new String[] { "Y", "",					"# (used to remove an associated section if the datasource contains this number of rows" });
		allParams.put("replaceNamesColumn",				new String[] { "Y", "-",				"col1 (The column in the main DST that contains the name to be replaced" });
		allParams.put("replaceNamesDS",					new String[] { "N", "-",				"dscode1,dscode2 (dscode1 is the primary DST, and dscode2 is the DST containing a mapping between the original name and the replacement name" });
		allParams.put("resolveDataVariables",			new String[] { "Y", "N",				"Y|N (will replace data substitution variables in dragThroughDS or dragThrough RS with other input RS/DS" });
		allParams.put("resolveSubVarColNm",				new String[] { "Y", "",					"# (used to indicate which column should have SV's resolved against" });
		allParams.put("resolveSubVarEscapeQuotes",		new String[] { "Y", "N",				"Y|N (used to replace single and double quotes that do not delimit function arguments with HTML escaped entities before trying to parse functions. Replacement is reversed afterwards.)" });		
		allParams.put("resolveSubVarOutputLang",		new String[] { "Y", "en-gb",			"iso-cd (used to define the locale when substitution variables/functions." });
		allParams.put("returnFirstX",					new String[] { "Y", "-1",				"# (This will return only the first X rows from a datasource." });
		allParams.put("returnLastX",					new String[] { "Y", "-1",				"# (This will return only the last X rows from a datasource." });
		allParams.put("returnRowNumX",					new String[] { "N", "-",				"# (This will return a specific row number from a datasource" });
		allParams.put("rotateData",						new String[] { "Y", "-",				"Y|N (Default -. If Y, transposes the table fo data in the DS."});
		allParams.put("rotateDataAddHeaderColumn",		new String[] { "Y", "-",				"{column name} (Specified a column of the output data to contain the field names from the data to be rotated. " + "Additional columns of rotated data will be shifted over this as if it were inserted.)"});
		allParams.put("rotateDataFormatIntoText",		new String[] { "Y", "Y",				"Y|N (Default Y. If any numeric or date data is rotated into a text column, it will be formatted according to "+ "the input RS formatting (if set). This can result in less precise data.)"});
		allParams.put("rotateDataRemoveHeaderColumn",	new String[] { "Y", "-",				"{column name} (Specifies a column in the data-to-be-rotated that contains discardable header data. " + "If rotating input record set, input column name; if rotating processed output, output cuolumn name "});
		//            instruction,	                     work on output DS, default value,      help text
		allParams.put("setData",						new String[] { "Y", "-",				"row,col,value (sets the data at (row,col) to value, counting from 0. col can be a column number, system name, or preferred name.)" });
		allParams.put("setValue",						new String[] { "Y", "-",				"col=value (sets the data at the given column to a value for each row, counting from 0. col can be a column number, system name, or preferred name.)" });
		allParams.put("showComponentRowCount",			new String[] { "Y", "",					"# (used to show an associated component only if the datasource contains this number of rows" });
		allParams.put("showSectionRowCount",			new String[] { "Y", "",					"# (used to show an associated section only if the datasource contains this number of rows" });
		allParams.put("smallPercentage",				new String[] { "Y", "-",				"col1,col2,entityValue (This will check col2 for a value < the entityValue, and merge all records into 1 new record with 'Other' in col1" });
		allParams.put("smallPercentageException",		new String[] { "Y", "-",				"value1 (This will exclude any records with value1 in col1 (see smallPercentage) from being merged" });
		allParams.put("sortColumn",						new String[] { "Y", "-",				"col1 (The column to sort on" });
		allParams.put("sortOrder",						new String[] { "Y", "asc",				"asc|desc (sort in either ascending or descending order." });
		allParams.put("splitIntoBalancedColumns",		new String[] { "Y", "1",				"# (Splits input DS into # columns. Output RS must be have # repetitions of input RS column types." });
		allParams.put("stripTags",						new String[] { "Y", "-",				"Y|N (Default Y. If Y it will remove all HTML tags from all records except those specified in StripTagsExceptions" });
		allParams.put("stripTagsException",				new String[] { "Y", "-",				"a list of tags that will not be removed from StripTags eg img|/img|a href to leave images and hyperlinks" });
		allParams.put("stripTagsExceptions",			new String[] { "Y", "-",				"a list of tags that will not be removed from StripTags eg img|/img|a href to leave images and hyperlinks" });
		allParams.put("substituteColumn",				new String[] { "Y", "-",				"col,value (Will replace the value in the specified column with the value passed" });
		allParams.put("substituteColumnNulls",			new String[] { "Y", "N",				"Will substitute the column even if the incoming data is null." });
		allParams.put("subtractColumns",				new String[] { "Y", "-",				"col1,col2,resultCol (used to subtract two columns (col1 - col2) and put the result in resultCol" });
		allParams.put("sumColumns",						new String[] { "Y", "-",				"col1,col2,resultCol (used to add two columns together (col1 and col2) and put the result in resultCol" });
		allParams.put("suppressWarnings",				new String[] { "Y", "-",				"warningType (Will suppress all warnings of this type. \"Rendering\" is the most likely type desired.)" });
		allParams.put("thresholdCheck",					new String[] { "Y", "-",				"col1,entityVal,logLevel,operator,custoMMessage (Checks col1 against the entity value, and if it is < the entity value it produces a warning. LogLevel : All/ReportOnly/LogOnly. Operator : </>/<=/>=. CustomMessage uses substitution variables $col1Value$ and $thresholdValue$" });
		allParams.put("transformMonthlyMatrix",			new String[] { "N", "N",				"Y|N (used to turn a vertical list of returns into a monthly/year matrix" });
		allParams.put("transformMonthlyMatrixDateColumn",	new String[] { "N", "-",			"column name or number (starting at 1) of the date column on the input recordset" });
		allParams.put("transformMonthlyMatrixReturnColumn",	new String[] { "N", "-",			"column name or number (starting at 1) of the value column on the input recordset" });
		allParams.put("undoHtmlEscapes",				new String[] { "Y", "-",				"Y|# (Default -. If Y it will remove all HTML escaped characters (such as &gt;) for all column. If a # is specified, it will remove from that column number only." });
		allParams.put("unionDatasources",				new String[] { "N", "-",				"dscode,dscode,... (used to define if multiple datasources should be unioned together (and their order)" });
		allParams.put("warningIf",						new String[] { "Y", "-",				"X=Y (used to create a BVR alert if the column X is equal to the value Z. Can also use !=, <=, >=, ==, <, >" });
		allParams.put("warningIfBetween",				new String[] { "Y", "-",				"X,Y,Z (used to create a BVR alert if the column X is between values Y and Z eg col1,10,-10 will alert if any " +"value in col1 is between 10 and -10" });
		allParams.put("warningIfColDiffer",				new String[] { "Y", "-",				"X=Y (used to create a BVR alert if the column X is not equal to the column Z" });
		allParams.put("warningIfColEqual",				new String[] { "Y", "-",				"X=Y (used to create a BVR alert if the column X is equal to the column Z" });
		allParams.put("warningIfDisplayType",			new String[] { "Y", "All",				"X (used with any of the warningIf commands to specify where the message will be displayed. Options : All, None, Log, Report " });
		allParams.put("warningIfMsg",					new String[] { "Y", "BVR Violation:",	"Lorem Ipsum Dolor, (text message to be put out if warningIf detects a match" });
		allParams.put("warningIfMsgCol",				new String[] { "Y", "-",				"col1 (if the warningIf values equal then display the value in this column against value Z" });
		allParams.put("warningIfNot",					new String[] { "Y", "-",				"X=Y (used to create a BVR alert if the column X is not equal to the value Z" });
		allParams.put("warningIfNotBetween",			new String[] { "Y", "-",				"X,Y,Z (used to create a BVR alert if the column X is not between values Y and Z eg col1,10,-10 will alert if any value in col1 is between 10 and -10" });
		allParams.put("warningIfOverriden",				new String[] { "Y", "Y",				"Display a warning if the main datasource value has been overridden" });
		allParams.put("warningIfRepeat",				new String[] { "Y", "Y",				"Y|N (used if the warning should repeat if invoked for multiple rows of data)" });
		allParams.put("warningIfRowCount",				new String[] { "Y", "-",				"=X (used to create a BVR alert if the number if rows is zero, or a specified number" });
		allParams.put("warningIfRowCountNot",			new String[] { "Y", "-",				"=X (used to create a BVR alert if the number if rows is not zero, or a specified number" });
		allParams.put("warningIfSumDiffer",				new String[] { "Y", "-",				"A=X[,t] (used where A is the col to sum and X is the amount it should be. ',t' is optional parameter to allow a " +"tolerance for close values (typically used to check a set of percentages add up to 100)" });
		allParams.put("warningIfSumDifference",			new String[] { "Y", "-",				"A.B=X.Y (used to define which input columns to compare - A & X are input recordset/datasource codes, B and Y are " +"system column codes" });
		allParams.put("warningIfSumDifferenceMethod",	new String[] { "Y", "-",				"X/Y (used to if the two sides should sum all records (SumAll) and/or use to only use one (last) record (LastRecord)" });
		allParams.put("warningIfSumMsg",				new String[] { "Y", "-",				"txt (used to set the message to be displayed if warningIfSumDifference detects a difference" });
		allParams.put("warningIfTrace",					new String[] { "Y", "N",				"Y|N (used to trace warningIf checks" });
		allParams.put("wrapIntoBalancedColumns",		new String[] { "Y", "1",				"# (Wraps input DS into rows with # items per row. Output RS must be have # repetitions of input RS column types." });
                

	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processDragThroughRSorDS(String code)
			throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processDragThroughRSorDS starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processDragThroughRSorDS starting";

		String dragByColumnNameStr = getAndRemoveInstruction("dragByColumnName");
		boolean dragByColumnNameBool = dragByColumnNameStr != null
				&& dragByColumnNameStr.equalsIgnoreCase("Y");

		RecordSet dragThroughRSorDS = this.getInputRecordSet(code, true);
		if (dragThroughRSorDS != null) {
			if (dragByColumnNameBool) {
				writeOutputRecordsetUsingColName(dragThroughRSorDS);
			} else {
				writeOutputRecordset(dragThroughRSorDS);
			}
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processReturnFirstX() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processReturnFirstX starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processReturnFirstX starting";

		int returnFirstX = Integer
				.parseInt(getAndRemoveInstruction("returnFirstX"));

		if (returnFirstX > outputRecordSet.getRowCount()) {
			return;
		}
		RecordRow[] row = new RecordRow[returnFirstX];
		for (int i = 0; i < returnFirstX; i++) {
			row[i] = outputRecordSet.getRow(i);
		}

		this.deleteOutputRecordset();

		for (int i = 0; i < returnFirstX; i++) {
			outputRecordSet.addRow(row[i]);
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processReturnLastX() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processReturnLastX starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processReturnLastX starting";

		int returnLastX = Integer
				.parseInt(getAndRemoveInstruction("returnLastX"));
		int rowCount = outputRecordSet.getRowCount();
		int startRow = rowCount - returnLastX;

		if (returnLastX > rowCount) {
			return;
		}
		RecordRow[] row = new RecordRow[returnLastX];

		int j = 0;
		for (int i = startRow; i < rowCount; i++) {
			row[j++] = outputRecordSet.getRow(i);
		}

		this.deleteOutputRecordset();

		for (int i = 0; i < returnLastX; i++) {
			outputRecordSet.addRow(row[i]);
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processRowNum() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processRowNum starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processRowNum starting";

		int svc = -1;

		for (int r = 0; r < inputRecordSetList.getRecordSetCount(); r++) { // for r=0, r maximum length is the number of recordset(s)
			if (r > 1) {
				outputRecordSet
						.addWarningMessage("Error: RecordSet is greater than one");
			} else {
				RecordSet irs = inputRecordSetList.getRecordSet(r);

				int row = Integer
						.parseInt(getAndRemoveInstruction("returnRowNumX")) - 1;
					// Have included '-1' because it  will bring indexOutOfBounds error//

				RecordRow recordRow = irs.getRow(row);

				this.deleteOutputRecordset();

				outputRecordSet.addRow(recordRow);

				if (svc > -1) {
					// If resolveSubVarColNm has been set in the DST notes then
					// this will trigger the DragThrough to look at substituting
					// anything it finds in the field for each record...
					try {
						outputRecordSet
								.setValue(
										row,
										svc,
										ComponentUtility.parseText(
												irs.getTid(),
												StringHelper.getResolveSubVars(
														outputRecordSet
																.getString(row,
																		svc),
														parameterList, true),
												getAndRemoveInstruction("resolveSubVarOutputLang")));
						if (outputRecordSet.getString(row, svc).indexOf("$_") > -1) {
							String s = getSVWarning(
									outputRecordSet.getString(row, svc),
									parameterList).trim();
							if (!s.equals("")) {
								outputRecordSet.addWarningMessage(s
										+ " (err1003)");
							}
						}
					} catch (Exception e) {
						outputRecordSet
								.addWarningMessage("Could not resolve substitution variable from within record "
										+ row
										+ " please check syntax (entities available are "
										+ parameterList.getParameterNames());
					}
				}
			}
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processRemoveComponentRowCount()
			throws VermilionNonFatalException, VermilionFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"removeComponentRowCount starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processRemoveComponentRowCount starting";

		if (getAndRemoveInstruction("removeComponentRowCount").equals(
				outputRecordSet.getRowCount() + "")) {
			RIDeleteComponent ri = RIDeleteComponent.getInstance();
			outputRecordSet.addRenderingInstruction(ri);

			if (trace) { 
				WarningMessage warningMessage = new WarningMessage(
						0,
						"Rendering",
						WarningMessage.DONT_OUTPUT,
						"removeComponentRowCount decided to remove component due to input record set row count match ("
								+ outputRecordSet.getRowCount() + " rows seen)");
				outputRecordSet.addWarningMessage(warningMessage);
			}
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processRemoveSectionRowCount()
			throws VermilionNonFatalException, VermilionFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processRemoveSectionRowCount starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processRemoveSectionRowCount starting";

		if (getAndRemoveInstruction("removeSectionRowCount").equals(
				outputRecordSet.getRowCount() + "")) {
			RIDeleteSection ri = RIDeleteSection.getInstance();
			outputRecordSet.addRenderingInstruction(ri);
			if (trace) {
				WarningMessage warningMessage = new WarningMessage(
						0,
						"Rendering",
						WarningMessage.DONT_OUTPUT,
						"removeSectionRowCount decided to remove section due to input record set row count match ("
								+ outputRecordSet.getRowCount() + " rows seen)");
				outputRecordSet.addWarningMessage(warningMessage);
			}
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processShowComponentRowCount()
			throws VermilionNonFatalException, VermilionFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processShowComponentRowCount starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processShowComponentRowCount starting";

		if (!getAndRemoveInstruction("showComponentRowCount").equals(
				outputRecordSet.getRowCount() + "")) {
			RIDeleteComponent ri = RIDeleteComponent.getInstance();
			outputRecordSet.addRenderingInstruction(ri);

			if (trace) {
				WarningMessage warningMessage = new WarningMessage(
						0,
						"Rendering",
						WarningMessage.DONT_OUTPUT,
						"showComponentRowCount decided to remove component due to input record set row count no match ("
								+ outputRecordSet.getRowCount() + " rows seen)");
				outputRecordSet.addWarningMessage(warningMessage);
			}
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processShowSectionRowCount()
			throws VermilionNonFatalException, VermilionFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processShowSectionRowCount starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processShowSectionRowCount starting";

		if (!getAndRemoveInstruction("showSectionRowCount").equals(
				outputRecordSet.getRowCount() + "")) {
			RIDeleteSection ri = RIDeleteSection.getInstance();
			outputRecordSet.addRenderingInstruction(ri);
			if (trace) {
				WarningMessage warningMessage = new WarningMessage(
						0,
						"Rendering",
						WarningMessage.DONT_OUTPUT,
						"showSectionRowCount decided to remove section due to input record set row count no match ("
								+ outputRecordSet.getRowCount() + " rows seen)");
				outputRecordSet.addWarningMessage(warningMessage);
			}
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processStrFilter() throws VermilionNonFatalException {
		processFilter(true);
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processStrFilterRemove() throws VermilionNonFatalException {
		processFilter(false);
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	/**
	 * 
	 * @param inputRecordSetList
	 * @param parameterList
	 * @param includeRows
	 *            - true if the output should only include the filtered rows,
	 *            false is the output should be all rows except the filtered
	 *            ones
	 * @throws VermilionNonFatalException
	 */
	protected void processFilter(boolean includeRows)
			throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processStrFilter starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processStrFilter starting";

		String filterColumn;
		String filterValue;

		if (includeRows) {
			filterColumn = getAndRemoveInstruction("filterColumn");
			filterValue = getAndRemoveInstruction("filterValue");
		} else {
			filterColumn = getAndRemoveInstruction("filterRemoveColumn");
			filterValue = getAndRemoveInstruction("filterRemoveValue");
		}

		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "filtering " + includeRows
							+ " on column " + filterColumn + " with value "
							+ filterValue);
			outputRecordSet.addWarningMessage(msg);
		}

		/* if (outputRecordSet.getRowCount() == 0) {
		writeOutputRecordset();
	}
	// this is now handled in the parent
	*/ 
		String filterIgnoreBlank = getAndRemoveInstruction("filterIgnoreBlank");
		int filterRowCount = Integer
				.parseInt(getAndRemoveInstruction("filterRowCount"));
		String resolveSubVarOutputLang = getAndRemoveInstruction("resolveSubVarOutputLang");
		String filterMatchType = getAndRemoveInstruction("filterMatchType");
		String filterCaseSensitive = getAndRemoveInstruction("filterCaseSensitive");
		
		// There aren't any substitution variables
		if (filterValue.indexOf("$_") == -1) {
			int aCol = getColumnIndexBySystemNm(outputRecordSet.getMetaData(),
					filterColumn);
			if (aCol < 0)
				return;

			for (int r = 0; r < outputRecordSet.getRowCount(); r++) {
				outputRecordSet.setValue(r, aCol, ComponentUtility.parseText(
						outputRecordSet.getTid(), StringHelper
								.getResolveSubVars(
										outputRecordSet.getString(r, aCol),
										parameterList, true),
						resolveSubVarOutputLang));
			}

			setRecordSet(getFilteredRows(filterColumn, filterValue,
					outputRecordSet, filterRowCount, filterIgnoreBlank,
					includeRows, filterMatchType, filterCaseSensitive));
			return;
		}

		// if it gets here there are substitution values in the filterValue
		try {
			// outputRecordSet.addWarningMessage("JCH1 " + filterValue);

			boolean isDate = filterValue.contains("$_Date.SysDate$") ? true
					: false;

			// if this is a date, we need to ensure it is in the correct format
			if (isDate) {
				// /formatdate($_Date.SysDate$,yyyymmdd)
				if (filterValue.startsWith("/formatdate")) {
					String dateFormat = filterValue.substring(28,
							filterValue.length() - 1);
					filterValue = filterValue.substring(12,
							filterValue.indexOf(","));

					filterValue = getFilterDateValue(filterValue,
							resolveSubVarOutputLang, dateFormat);

				} else {
					filterValue = getFilterDateValue(filterValue,
							resolveSubVarOutputLang);
				}

				// outputRecordSet.addWarningMessage("JCH3 " + filterValue);
			} else {
				// System.out.println("FilterValue : " + filterValue);
				filterValue = StringHelper.getResolveSubVars(filterValue,
						parameterList, true);
			}

			if (trace) {
				WarningMessage msg = new WarningMessage(0, "Rendering",
						WarningMessage.DONT_OUTPUT, "substitution value "
								+ filterValue);
				outputRecordSet.addWarningMessage(msg);
			}

			int aCol = getColumnIndexBySystemNm(outputRecordSet.getMetaData(),
					filterColumn);
			if (aCol < 0)
				return;

			for (int r = 0; r < outputRecordSet.getRowCount(); r++) {
				outputRecordSet.setValue(r, aCol, ComponentUtility.parseText(
						outputRecordSet.getTid(), StringHelper
								.getResolveSubVars(
										outputRecordSet.getString(r, aCol),
										parameterList, true),
						resolveSubVarOutputLang));
			}

			setRecordSet(getFilteredRows(filterColumn, filterValue,
					outputRecordSet, filterRowCount, filterIgnoreBlank,
					includeRows, filterMatchType, filterCaseSensitive));
		} catch (Exception e) {
			copyWarningMessages(inputRecordSetList, outputRecordSet);
			outputRecordSet.addWarningMessage(e.toString());
		}
		copyWarningMessages(inputRecordSetList, outputRecordSet);
	}

	private RecordSet getFilteredRows(String strFilterFieldName,
			String strFieldValue, RecordSet outputRS, int returnFirstX,
			String filterIgnoreBlank, boolean includeRows, String filterMatchType,
			String filterCaseSensitive) {
		
		runPos="getFilteredRows method starting";
		
		//used in processFilter
		String strRecordValue = null;
		String strInFieldValue = null;
		int iFilterFieldIndex = 0;
		RecordSet outRs = null;
		int numRecordsWritten = 0;
		int numRecordsDropped = 0;

		try {
			outRs = new RecordSet(outputRS.getMetaData());

			if (strFieldValue != null) {
				// get the column number of the specified field name
				iFilterFieldIndex = getColumnIndexBySystemNm(
						outputRS.getMetaData(), strFilterFieldName.trim());

				// for each row, check whether the filter field has the
				// specified value
				if (iFilterFieldIndex < 0) {
					return outRs;
				}
				
				strInFieldValue = strFieldValue.trim();
				
				if (trace) {
					WarningMessage warningMessage = new WarningMessage(0,
							"Rendering", WarningMessage.DONT_OUTPUT,
							"Converting " + strFilterFieldName.trim() + " = '"
									+ strInFieldValue + "' ( with matching type " + filterMatchType
									+") to appropriate regex comparison string");
					outRs.addWarningMessage(warningMessage);
				}
				
				//based on the filterMatchType, turn strInFieldValue to the appropriate regex match

				if(!filterMatchType.equalsIgnoreCase("REGEX")){
					strInFieldValue = Pattern.quote(strInFieldValue);
					if(filterMatchType.equalsIgnoreCase("CONTAINS")){
						strInFieldValue = ".*"+strInFieldValue+".*";
					}
				}
				if(filterIgnoreBlank.equalsIgnoreCase("Y") == includeRows){
					strInFieldValue="("+strInFieldValue+"|)";
				}
				if(filterCaseSensitive.equalsIgnoreCase("N")){
					strInFieldValue="(?i:"+strInFieldValue+")";
				}
				
				int iRowCount = outputRS.getRowCount();

				int iMaxRows = iRowCount;
				if (returnFirstX != -1)
					iMaxRows = Math.min(returnFirstX, iMaxRows);

				if (trace) {
					WarningMessage warningMessage = new WarningMessage(0,
							"Rendering", WarningMessage.DONT_OUTPUT,
							"Filtering " + strFilterFieldName.trim() + " = '"
									+ strInFieldValue + "' (" + iMaxRows
									+ " rows)");
					outRs.addWarningMessage(warningMessage);
				}


				int iCollectedRows = 0;

				for (int iRow = 0; iRow < iRowCount; iRow++) {
					strRecordValue = outputRS
							.getString(iRow, iFilterFieldIndex);
					if(strRecordValue == null){
						strRecordValue = "";
					}
					strRecordValue = strRecordValue.trim();
					
					// outputRecordSet.addWarningMessage("Comparing '" +
					// strInFieldValue + "' to '" + strRecordValue.trim() +
					// "' is " +
					// strInFieldValue.equalsIgnoreCase(strRecordValue.trim()));

					if (includeRows == strRecordValue.matches(strInFieldValue)){
						if (iCollectedRows < iMaxRows) {
							iCollectedRows++;
							outRs.addRow(outputRS.getRow(iRow));
							++numRecordsWritten;
						} else {
							++numRecordsDropped;
						}
					}else{
						++numRecordsDropped;
						
					}
					if (strRecordValue.indexOf("$_") > -1) {
						outRs.addWarningMessage("Unresolved variable seen : "
								+ strRecordValue);
					}
				}
			}
		} catch (VermilionNonFatalException e) {
			logManager.error("Unexpected getFilteredRows ("
					+ strFilterFieldName + " / " + strFieldValue + " ) error "
					+ e.toString());
			e.printStackTrace();
		}
		//logManager.debug("the final row count is: " + outRs.getRowCount());

		if (trace) {
			WarningMessage warningMessage = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "Passed through "
							+ numRecordsWritten + " records, dropped "
							+ numRecordsDropped + " records");
			outRs.addWarningMessage(warningMessage);
		}
		return outRs;
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processStrFilterList() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processStrFilterList starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processStrFilterList starting";

		String filterColumnList = getAndRemoveInstruction("filterColumnList");
		String filterValueList = getAndRemoveInstruction("filterValueList");

		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "filtering on column "
							+ filterColumnList + " with values "
							+ filterValueList);
			outputRecordSet.addWarningMessage(msg);
		}

		String filterIgnoreBlank = getAndRemoveInstruction("filterIgnoreBlank");
		int filterRowCount = Integer
				.parseInt(getAndRemoveInstruction("filterRowCount"));
		String resolveSubVarOutputLang = getAndRemoveInstruction("resolveSubVarOutputLang");
		RecordSet inRS = inputRecordSetList.getRecordSet(0);
		String filterMatchType = getAndRemoveInstruction("filterMatchType");
		String filterCaseSensitive = getAndRemoveInstruction("filterCaseSensitive");

		// There aren't any substitution variables
		if (filterValueList.indexOf("$_") == -1) {
			int aCol = getColumnIndexBySystemNm(outputRecordSet.getMetaData(),
					filterColumnList);
			if (aCol < 0)
				return;

			/*
			 * for(int r=0; r<outputRecordSet.getRowCount(); r++) {
			 * outputRecordSet.setValue(r, aCol,
			 * ComponentUtility.parseText(outputRecordSet.getTid(),
			 * StringHelper.getResolveSubVars(outputRecordSet.getString(r,aCol),
			 * parameterList, true), resolveSubVarOutputLang)); }
			 */

			// loop around the list of values to get all the records
			String[] filterValues = filterValueList.split(",");
			for (String filterValue : filterValues) {
				RecordSet rs = getFilteredRows(filterColumnList, filterValue,
						inRS, filterRowCount, filterIgnoreBlank, true, filterMatchType, filterCaseSensitive);

				// add all the recordset
				writeOutputRecordset(rs);
			}

			return;
		}

		// if it gets here there are substitution values in the filterValue
		try {
			boolean isDate = filterValueList.equals("$_Date.SysDate$") ? true
					: false;
			// if this is a date, we need to ensure it is in the correct format
			if (isDate) {
				filterValueList = getFilterDateValue(filterValueList,
						resolveSubVarOutputLang);
			} else {
				// System.out.println("FilterValue : " + filterValue);
				filterValueList = StringHelper.getResolveSubVars(
						filterValueList, parameterList, true);
			}

			if (trace) {
				WarningMessage msg = new WarningMessage(0, "Rendering",
						WarningMessage.DONT_OUTPUT, "substitution value "
								+ filterValueList);
				outputRecordSet.addWarningMessage(msg);
			}

			int aCol = getColumnIndexBySystemNm(outputRecordSet.getMetaData(),
					filterColumnList);
			if (aCol < 0)
				return;

			/*
			 * for(int r=0; r<outputRecordSet.getRowCount(); r++) {
			 * outputRecordSet.setValue(r, aCol,
			 * ComponentUtility.parseText(outputRecordSet.getTid(),
			 * StringHelper.getResolveSubVars(outputRecordSet.getString(r,aCol),
			 * parameterList, true), resolveSubVarOutputLang)); }
			 */
			// loop around the list of values to get all the records
			String[] filterValues = filterValueList.split(",");
			for (String filterValue : filterValues) {
				RecordSet rs = getFilteredRows(filterColumnList, filterValue,
						inRS, filterRowCount, filterIgnoreBlank, true, filterMatchType, filterCaseSensitive);

				// add all the recordset
				writeOutputRecordset(rs);
			}
		} catch (Exception e) {
			copyWarningMessages(inputRecordSetList, outputRecordSet);
			outputRecordSet.addWarningMessage(e.toString());
		}
		copyWarningMessages(inputRecordSetList, outputRecordSet);
	}

	protected String getFilterDateValue(String beforeFilterValue,
			String resolveSubVarOutputLang) {
		return getFilterDateValue(beforeFilterValue, resolveSubVarOutputLang,
				DateHelper._SIMPLEDATE_FORMAT_FULL_MONTH);
	}

	protected String getFilterDateValue(String beforeFilterValue,
			String resolveSubVarOutputLang, String dateFormat) {
		String afterFilterValue;

		// If filtering on a date check that the Date entity is passed
		ArrayList<String> pl = parameterList.getParameterNames();
		if (!pl.contains("Date")) {
			outputRecordSet
					.addWarningMessage("When filtering on a date the Date entity must be on the datasource");
			return "";
		}

		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "getResolveSubVars "
							+ StringHelper.getResolveSubVars(beforeFilterValue,
									parameterList, true));
			outputRecordSet.addWarningMessage(msg);
		}

		afterFilterValue = ComponentUtility.parseText(outputRecordSet.getTid(),
				StringHelper.getResolveSubVars(beforeFilterValue,
						parameterList, true), resolveSubVarOutputLang);

		// outputRecordSet.addWarningMessage("afterFilterValue : " +
		// afterFilterValue);

		if (afterFilterValue.equals(beforeFilterValue)) {
			outputRecordSet
					.addWarningMessage("DragThroughProcessor Could not resolve "
							+ beforeFilterValue
							+ " from "
							+ parameterList.getParameterNames());
		}

		SimpleDateFormat sdf1 = new SimpleDateFormat(
				DateHelper._SIMPLEDATE_FORMAT_YYYYMMDD);
		SimpleDateFormat sdf2 = new SimpleDateFormat(dateFormat);
		Date theDate = null;
		try {
			theDate = sdf1.parse(afterFilterValue);

			// outputRecordSet.addWarningMessage("theDate : " +
			// theDate.toString());

			// check whether previous period date is required
			String strPeriod = getAndRemoveInstruction("period");
			if (strPeriod != null && strPeriod.equalsIgnoreCase("previous")) {
				String beforeStrTerm = DateHelper.ENTITYF_TERM_CODE_VARIABLE;
				String strTerm = ComponentUtility.parseText(outputRecordSet
						.getTid(), StringHelper.getResolveSubVars(
						DateHelper.ENTITYF_TERM_CODE_VARIABLE, parameterList,
						true), resolveSubVarOutputLang);
				if (strTerm.equals(beforeStrTerm)) {
					outputRecordSet
							.addWarningMessage("DragThroughProcessor Could not process "
									+ DateHelper.ENTITYF_TERM_CODE_VARIABLE);
				}
				//logManager.debug(">> >> >> >> the term is: " + strTerm);
				theDate = DateHelper.getPreviousTermEnd(theDate, strTerm);
			}
		} catch (ParseException e) {
			outputRecordSet.addWarningMessage("Unexpected DragThrough error "
					+ e.toString());
		}
		// outputRecordSet.addWarningMessage("returning : " +
		// sdf2.format(theDate));

		return sdf2.format(theDate);
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processForceSnapshotFg() throws VermilionNonFatalException {
		// if the valuation ready data source hasn't been signed off, prevent
		// snapshots
		// set the snapshot behaviour
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processForceSnapshotFg starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processForceSnapshotFg starting";

		String flag1 = getAndRemoveInstruction("forceSnapshotFg");
		String flag2 = getAndRemoveInstruction("forceSnapshotIfDSSignedOff");

		if (flag1.toUpperCase().equals("N")
				&& (flag2.toUpperCase().equals("N"))) {
			outputRecordSet
					.addWarningMessage(outputRecordSet.getDataSourceName()
							+ " Not applying snapshot as neither forceSnapshotFg flag nor forceSnapshotIfDSSignedOff flag is set to Y");
			return;
		}

		String forceSnapshotDSCode = getAndRemoveInstruction("forceSnapshotDSCode");

		if (forceSnapshotDSCode == null) {
			outputRecordSet.addWarningMessage(outputRecordSet.getDataSourceName()
					+ " Please provide the forceSnapshotDSCode");
			return;
		}

		
		if(forceSnapshotDSCode==null || forceSnapshotDSCode.equals("") ) {
			if ( inputRecordSetList.getRecordSetCount()!=1) {
				outputRecordSet.addWarningMessage(outputRecordSet.getDataSourceName()+" Please provide the forceSnapshotDSCode");
				return;
			}
			else {
		      	//logManager.error("Check746------------------");
        		forceSnapshotDSCode=inputRecordSetList.getRecordSetCode(0);
        		outputRecordSet.addWarningMessage("forceSnapshot defaulting to "+forceSnapshotDSCode);
	        }
		}
		
		
		RecordSet isReady = outputRecordSet;
		if (inputRecordSetList.getDataSourceCodes().contains(forceSnapshotDSCode)) {
			isReady = inputRecordSetList.getRecordSetByDataSourceCode(forceSnapshotDSCode);
			// set the record set behaviour
			setRecordSet(StringHelper.setSnapshotBehaviour(outputRecordSet,isReady, parameterList));
		} else {
			if ( inputRecordSetList.getRecordSetByRecordSetCode(forceSnapshotDSCode)!=null ) {
              	isReady = outputRecordSet; // inputRecordSetList.getRecordSetByRecordSetCode(forceSnapshotDSCode);
            	// set the record set behaviour
            	setRecordSet( StringHelper.setSnapshotBehaviour(outputRecordSet, isReady, parameterList));
            	outputRecordSet.setSnapshotBehaviour(RecordSet.SNAPSHOT_PREVENT);
            	outputRecordSet.addWarningMessage("Snapshots Prevented768");
	        }
	        else {
	          	//logManager.error("Check763------------------");
	        	String rscodes="";
	        	for(int i=0; i<inputRecordSetList.getRecordSetCount();i++) {
	        		rscodes+=", "+inputRecordSetList.getRecordSetCode(i);
	        	}
	        	rscodes=rscodes.substring(1); // get rid of the leading comma
	            outputRecordSet.addWarningMessage(outputRecordSet.getDataSourceName()+" DragThroughProcessor did not find "
	    				+forceSnapshotDSCode+" DSCode " + " in the input list (datasource list is "
	            		+inputRecordSetList.getDataSourceCodes()+"/recordset list is ["+rscodes+"])");
	        }
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processLogShowSignOffStatusFg()
			throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processLogShowSignOffStatusFg starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processLogShowSignOffStatusFg starting";

		if (!(getAndRemoveInstruction("logShowSignOffStatusFg").toUpperCase()
				.equals("Y") && outputRecordSet.getDataSourceInfo()
				.getSignOffRequired())) {
			return;
		}

		String strStatusCode = outputRecordSet.getApprovedStatusCode();
		boolean dataSignedOff = (strStatusCode != null
				&& strStatusCode.equals("A") ? true : false);

		// if the valuation ready data source hasn't been signed off, prevent
		// snapshots
		String inputEntityValues = StringHelper.getEntityValuesInString(
				outputRecordSet, parameterList);
		if (dataSignedOff) {
			WarningMessage warningMessage = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "The data source: '"
							+ outputRecordSet.getDataSourceName()
							+ "' "
							+ inputEntityValues
							+ " has been signed off ("
							+ new java.text.SimpleDateFormat(
									"dd.MMM.yyyy HH:mm").format(new Date())
							+ ")");
			outputRecordSet.addWarningMessage(warningMessage);
			return;
		}

		String newMsg = "The data source: '"
				+ outputRecordSet.getDataSourceName() + "' "
				+ inputEntityValues + " has not been signed off.";
		boolean found = false;
		int warningCount = outputRecordSet.getWarningMessageCount();
		for (int w = 0; w < warningCount; w++) {
			if (outputRecordSet.getWarningMessageText(w).equals(newMsg)) {
				found = true;
				break;
			}
		}

		if (!found) {
			WarningMessage warningMessage = new WarningMessage(0, "Warning",
					WarningMessage.DONT_OUTPUT, newMsg);
			outputRecordSet.addWarningMessage(warningMessage);
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processWarningIfColDiffer()
			throws VermilionNonFatalException {
		String instr = getAndRemoveInstruction("warningIfColDiffer");
		processWarningIfColCheck(instr, true);
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processWarningIfColSame() throws VermilionNonFatalException {
		String instr = getAndRemoveInstruction("warningIfColSame");
		processWarningIfColCheck(instr, false);
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processWarningIfColCheck(String instr, boolean differ)
			throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "warningIfColCheck starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processWarningIfColCheck starting";

		String warningIfMsgCol = getAndRemoveInstruction("warningIfMsgCol");
		String warningIfMsg = getAndRemoveInstruction("warningIfMsg");
		String warningIfTrace = getAndRemoveInstruction("warningIfTrace");
		String warningIfDisplayType = getAndRemoveInstruction("warningIfDisplayType");

		instr += " ";
		String[] parts = instr.split("=");
		if (parts.length < 2) {
			outputRecordSet
					.addWarningMessage("Missing parts to [warningIfColSame]. Expected 'warningIf=xxx=yyy' - only see "
							+ parts.length + " parts (" + instr + ")");
			return;
		}

		String xCol = parts[0].trim();
		String yCol = parts[1].trim();
		RecordSetMetaData rsmd = outputRecordSet.getMetaData();

		int xColIndex = getColumnIndexBySystemNm(rsmd, xCol);
		int yColIndex = getColumnIndexBySystemNm(rsmd, yCol);
		int msgColIndex = -1;

		if (!warningIfMsgCol.equals("-"))
			msgColIndex = getColumnIndexBySystemNm(rsmd, warningIfMsgCol);

		if (xColIndex < 0 || yColIndex < 0)
			return;

		String xValue;
		String yValue;
		for (int row = 0; row < outputRecordSet.getRowCount(); row++) {

			xValue = outputRecordSet.getRow(row).getString(xColIndex);
			yValue = outputRecordSet.getRow(row).getString(yColIndex);

			if (warningIfTrace.equalsIgnoreCase("Y")) {
				WarningMessage wm = new WarningMessage(10, "WARN",
						WarningMessage.DONT_OUTPUT, "DTP "
								+ outputRecordSet.getDataSourceCode()
								+ ": warningIf Trace - Seen " + xValue);
				outputRecordSet.addWarningMessage(wm);
			}

			if (differ) {
				if ((xValue == null && yValue != null)
						|| (!xValue.equalsIgnoreCase(yValue))) {
					addWarningIfMessage(msgColIndex, row, xValue, warningIfMsg,
							warningIfDisplayType);
				}
			} else {
				if ((xValue == null && (yValue.equals("") || yValue
						.equalsIgnoreCase("<null>")))
						|| (xValue.equalsIgnoreCase(yValue))) {
					addWarningIfMessage(msgColIndex, row, xValue, warningIfMsg,
							warningIfDisplayType);
				}
			}
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processWarningIfBetween() throws VermilionNonFatalException {
		String warningIf = getAndRemoveInstruction("warningIfBetween");
		processWarningBetween(warningIf, true);
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processWarningIfNotBetween()
			throws VermilionNonFatalException {
		String warningIf = getAndRemoveInstruction("warningIfNotBetween");
		processWarningBetween(warningIf, false);
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processWarningBetween(String warningIf, boolean between)
			throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "warningIfBetween starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processWarningBetween starting";

		String warningIfMsgCol = getAndRemoveInstruction("warningIfMsgCol");
		String warningIfMsg = getAndRemoveInstruction("warningIfMsg");
		String warningIfRepeat = getAndRemoveInstruction("warningIfRepeat");
		String warningIfDisplayType = getAndRemoveInstruction("warningIfDisplayType");
		boolean singleWarningOnly = warningIfRepeat.equalsIgnoreCase("N");

		warningIf += " ";
		String[] parts = warningIf.split("\\,");

		String xCol = parts[0].trim();
		String yValue = parts[1].trim();
		String zValue = parts[2].trim();

		yValue = StringHelper.getResolveSubVars(yValue, parameterList, true);
		zValue = StringHelper.getResolveSubVars(zValue, parameterList, true);
		RecordSetMetaData rsmd = outputRecordSet.getMetaData();

		int xColIndex = getColumnIndexBySystemNm(rsmd, xCol);
		int msgColIndex = -1;

		if (!warningIfMsgCol.equals("-"))
			msgColIndex = getColumnIndexBySystemNm(rsmd, warningIfMsgCol);

		if (xColIndex < 0)
			return;

		if (yValue == null || yValue.equalsIgnoreCase("<null>")) {
			outputRecordSet.addWarningMessage("You must specify a value in y");
			return;
		}
		if (zValue == null || zValue.equalsIgnoreCase("<null>")) {
			outputRecordSet.addWarningMessage("You must specify a value in z");
			return;
		}

		Double yDecimal = new Double(yValue);
		Double zDecimal = new Double(zValue);

		try {
			String xValue;
			for (int row = 0; row < outputRecordSet.getRowCount(); row++) {
				xValue = outputRecordSet.getRow(row).getString(xColIndex);

				if (xValue == null) {
					continue;
				}

				Double xDecimal = new Double(xValue);
				if (between) {
					if (xDecimal < yDecimal && xDecimal > zDecimal) {
						addWarningIfMessage(msgColIndex, row, xValue,
								warningIfMsg, warningIfDisplayType);
						if (singleWarningOnly) {
							break;
						}
					}
				} else {
					if (xDecimal > yDecimal || xDecimal < zDecimal) {
						addWarningIfMessage(msgColIndex, row, xValue,
								warningIfMsg, warningIfDisplayType);
						if (singleWarningOnly) {
							break;
						}
					}

				}
			}
		} catch (Exception e) {
			e.printStackTrace();
			WarningMessage wm = new WarningMessage(10, "WARN",
					WarningMessage.DONT_OUTPUT, "DTP "
							+ outputRecordSet.getDataSourceCode()
							+ ": warningIf Exception " + e.getMessage());
			outputRecordSet.addWarningMessage(wm);

		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processWarningIfRowCount() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "warningIfRowCount starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processWarningIfRowCount starting";

		String warningIfRowCount = getAndRemoveInstruction("warningIfRowCount");

		String warningIfMsg = getAndRemoveInstruction("warningIfMsg");
		String warningIfDisplayType = getAndRemoveInstruction("warningIfDisplayType");

		if (outputRecordSet.getRowCount() == Integer
				.parseInt(warningIfRowCount)) {
			addWarningIfMessage(-1, 0, "rowcount", warningIfMsg,
					warningIfDisplayType);
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processWarningIfRowCountNot()
			throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "warningIfRowCountNot starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processWarningIfRowCountNot starting";

		String warningIfRowCountNot = getAndRemoveInstruction("warningIfRowCountNot");

		String warningIfMsg = getAndRemoveInstruction("warningIfMsg");
		String warningIfDisplayType = getAndRemoveInstruction("warningIfDisplayType");

		if (outputRecordSet.getRowCount() != Integer
				.parseInt(warningIfRowCountNot)) {
			addWarningIfMessage(-1, 0, "rowcount", warningIfMsg,
					warningIfDisplayType);
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processWarningIf() throws VermilionNonFatalException {
		String warningIf = getAndRemoveInstruction("warningIf");
		processWarningIfAny(warningIf);
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processWarningIfNot() throws VermilionNonFatalException {
		String warningIfNot = getAndRemoveInstruction("warningIfNot");
		String[] parts;
		String warningIf="";
		if (warningIfNot.indexOf("!=") > -1) {
			parts = warningIfNot.split("!=");
			warningIf = parts[0] + "==" + parts[1];
		} else if (warningIfNot.indexOf("<>") > -1) {
			parts = warningIfNot.split("<>");
			warningIf = parts[0] + "==" + parts[1];
		} else if (warningIfNot.indexOf("==") > -1) {
			parts = warningIfNot.split("==");
			warningIf = parts[0] + "!=" + parts[1];
		} else if (warningIfNot.indexOf(">=") > -1) {
			parts = warningIfNot.split(">=");
			warningIf = parts[0] + "<" + parts[1];
		} else if (warningIfNot.indexOf("<=") > -1) {
			parts = warningIfNot.split("<=");
			warningIf = parts[0] + ">" + parts[1];
		} else if (warningIfNot.indexOf("=") > -1) {
			parts = warningIfNot.split("=");
			warningIf = parts[0] + "!=" + parts[1];
		} else if (warningIfNot.indexOf(">") > -1) {
			parts = warningIfNot.split(">");
			warningIf = parts[0] + "<=" + parts[1];
		} else if (warningIfNot.indexOf("<") > -1) {
			parts = warningIfNot.split("<");
			warningIf = parts[0] + ">=" + parts[1];
		}
		processWarningIfAny(warningIf);
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processWarningIfAny(String warningIf)
			throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "warningIf starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processWarningIfAny starting";

		String warningIfMsgCol = getAndRemoveInstruction("warningIfMsgCol");
		String warningIfMsg = getAndRemoveInstruction("warningIfMsg");
		String warningIfTrace = getAndRemoveInstruction("warningIfTrace");
		String warningIfRepeat = getAndRemoveInstruction("warningIfRepeat");
		String warningIfDisplayType = getAndRemoveInstruction("warningIfDisplayType");

		boolean singleWarningOnly = warningIfRepeat.equalsIgnoreCase("N");

		warningIf += " ";
		String[] parts = getParts(warningIf);
		String comparison = getComparison(warningIf);

		String xCol = parts[0].trim();
		String yValue = parts[1].trim();
		
		yValue = StringHelper.getResolveSubVars(yValue, parameterList, true);
		
		RecordSetMetaData rsmd = outputRecordSet.getMetaData();

		int xColIndex = getColumnIndexBySystemNm(rsmd, xCol);
		int msgColIndex = -1;

		if (!warningIfMsgCol.equals("-"))
			msgColIndex = getColumnIndexBySystemNm(rsmd, warningIfMsgCol);

		if (xColIndex < 0)
			return;

		RecordSetMetaData mf = outputRecordSet.getMetaData();
		String format = (String) mf.getField(xColIndex).getFieldDisplayFormat();
		String type = (String) mf.getField(xColIndex).getFieldType();
		NumberFormat numFormatter = new DecimalFormat(format);
		
		if (type.equalsIgnoreCase("decimal")
				|| type.equalsIgnoreCase("integer")) {
			
			if(yValue.equalsIgnoreCase("") || yValue.equalsIgnoreCase("null")){
				yValue="";
			}
			
			/*if (!yValue.equalsIgnoreCase("<null>")) {
				
				Double Z1 = new Double(yValue);
				yValue = numFormatter.format(Z1);
			}*/
			
			else{
				Double Z1 = new Double(yValue);
				yValue = numFormatter.format(Z1);
			}
			
				
		}

		try {
			String xValue;
			for (int row = 0; row < outputRecordSet.getRowCount(); row++) {
			
				xValue = outputRecordSet.getRow(row).getString(xColIndex);
				
				
				if(xValue.isEmpty()){
					xValue = null;
				}
				
				if (type.equalsIgnoreCase("decimal")
						|| type.equalsIgnoreCase("integer")) {
					
					if (xValue != null) {
						Double colValue1 = new Double(xValue);
						xValue = numFormatter.format(colValue1);
					}
					
					
				}

				if (warningIfTrace.equalsIgnoreCase("Y")) {
					WarningMessage wm = new WarningMessage(10, "WARN",
							WarningMessage.DONT_OUTPUT, "DTP "
									+ outputRecordSet.getDataSourceCode()
									+ ": warningIf Trace - Seen " + xValue);
					outputRecordSet.addWarningMessage(wm);
				}

				boolean fireAlert = false;

				if (comparison.equals("==")) {
					
					if (xValue == null) {
			
						if (yValue.equals("")
								|| yValue.equalsIgnoreCase("<null>") || yValue.isEmpty() || yValue.equalsIgnoreCase("null")) {
							fireAlert = true;
						}
					} else if (xValue.equalsIgnoreCase(yValue)) {
						fireAlert = true;
					}
					
					
					
				} else if (comparison.equals("!=")) {
					if (xValue == null) {
						if (!(yValue.equals("") || yValue
								.equalsIgnoreCase("<null>")) || yValue.isEmpty()) {
							fireAlert = true;
						}
					} else if (!xValue.equalsIgnoreCase(yValue)) {
						fireAlert = true;
					}
				} else {
					if ((type.equalsIgnoreCase("decimal") || type
							.equalsIgnoreCase("integer"))
							&& StringHelper.isNumeric(xValue)
							&& StringHelper.isNumeric(yValue)) {
						fireAlert = checkNumericValues(comparison, type,
								xValue, yValue);
						
					} else {
						fireAlert = checkStringValues(comparison, type, xValue,
								yValue);
					}
				}

				if (fireAlert) {
					addWarningIfMessage(msgColIndex, row, xValue, warningIfMsg,
							warningIfDisplayType);
					if (singleWarningOnly) {
						break;
					}
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
			WarningMessage wm = new WarningMessage(10, "WARN",
					WarningMessage.DONT_OUTPUT, "DTP "
							+ outputRecordSet.getDataSourceCode()
							+ ": warningIf Exception " + e.getMessage());
			outputRecordSet.addWarningMessage(wm);

		}
	}

	private String[] getParts(String warningIf) {
		// used for processWarningIfAny
		String[] parts = {};
		if (warningIf.indexOf("!=") > -1) {
			parts = warningIf.split("!=");
		} else if (warningIf.indexOf("<>") > -1) {
			parts = warningIf.split("<>");
		} else if (warningIf.indexOf("==") > -1) {
			parts = warningIf.split("==");
		} else if (warningIf.indexOf(">=") > -1) {
			parts = warningIf.split(">=");
		} else if (warningIf.indexOf("<=") > -1) {
			parts = warningIf.split("<=");
		} else if (warningIf.indexOf("=") > -1) {
			parts = warningIf.split("=");
		} else if (warningIf.indexOf(">") > -1) {
			parts = warningIf.split(">");
		} else if (warningIf.indexOf("<") > -1) {
			parts = warningIf.split("<");
		}

		if (parts.length < 2) {
			outputRecordSet
					.addWarningMessage("Missing parts to [warningIf]. Expected 'warningIf=xxx[operator]yyy' - only see "
							+ parts.length + " parts (" + warningIf + ")");
		}

		return parts;
	}

	private String getComparison(String warningIf) {
		// used for processWarningIfAny
		String comparison = "==";
		if (warningIf.indexOf("!=") > -1) {
			comparison = "!=";
		} else if (warningIf.indexOf("<>") > -1) {
			comparison = "!=";
		} else if (warningIf.indexOf("==") > -1) {
			comparison = "==";
		} else if (warningIf.indexOf(">=") > -1) {
			comparison = ">=";
		} else if (warningIf.indexOf("<=") > -1) {
			comparison = "<=";
		} else if (warningIf.indexOf("=") > -1) {
			comparison = "==";
		} else if (warningIf.indexOf(">") > -1) {
			comparison = ">";
		} else if (warningIf.indexOf("<") > -1) {
			comparison = "<";
		} else {
			outputRecordSet
					.addWarningMessage("The operator for [warningIf] is invalid. Valid operators are : =, ==, !=, <=, >=, <, >, <> "
							+ "Expected 'warningIf=xxx[operator]yyy' - only see "
							+ warningIf + ")");

		}

		return comparison;
	}

	private boolean checkNumericValues(String comparison, String type,
			String xValue, String yValue) {
		// used for processWarningIfAny
		boolean fireAlert = false;
		
		BigDecimal xDecimal = new BigDecimal(xValue);
		BigDecimal yDecimal = new BigDecimal(yValue);

		if (comparison.equals("<=")) {
			return (xDecimal.compareTo(yDecimal) <= 0);
		} else if (comparison.equals(">=")) {
			return (xDecimal.compareTo(yDecimal) >= 0);
		} else if (comparison.equals(">")) {
			return (xDecimal.compareTo(yDecimal) > 0);
		} else if (comparison.equals("<")) {
			return (xDecimal.compareTo(yDecimal) < 0);
		}

		return fireAlert;
	}

	private boolean checkStringValues(String comparison, String type,
			String xValue, String yValue) {
		// used for processWarningIfAny
		boolean fireAlert = false;
		if (comparison.equals("<=")) {
			return (xValue.compareTo(yValue) <= 0);
		} else if (comparison.equals(">=")) {
			return (xValue.compareTo(yValue) >= 0);
		} else if (comparison.equals(">")) {
			return (xValue.compareTo(yValue) > 0);
		} else if (comparison.equals("<")) {
			return (xValue.compareTo(yValue) < 0);
		}

		return fireAlert;
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void addWarningIfMessage(int msgColIndex, int row, String xValue,
			String warningIfMsg, String warningIfDisplayType)
			throws VermilionNonFatalException {
		// used for all warningIf processes
		// All, None, Log, Report

		String msgColValue = "";
		if (msgColIndex != -1)
			msgColValue = " ("
					+ outputRecordSet.getRow(row).getString(msgColIndex) + "="
					+ xValue + ")";
		else
			msgColValue = " (" + xValue + ")";

		String cid = outputRecordSet.getDataSourceCode();
		String message = cid + ":: " + warningIfMsg + msgColValue;

		if (warningIfDisplayType.equalsIgnoreCase("None")) {
			outputRecordSet.addWarningMessage(new WarningMessage(0,
					"Rendering", WarningMessage.DONT_OUTPUT, message));
		} else if (warningIfDisplayType.equalsIgnoreCase("Log")) {
			outputRecordSet.addWarningMessage(new WarningMessage(0,
					"Rendering", WarningMessage.OUTPUT_TO_LOG_ONLY, message));
		} else if (warningIfDisplayType.equalsIgnoreCase("Report")) {
			outputRecordSet
					.addWarningMessage(new WarningMessage(0, "Rendering",
							WarningMessage.OUTPUT_TO_REPORT_ONLY, message));
		} else {
			outputRecordSet.addWarningMessage(message);
		}

	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processWarningIfSumDifference()
			throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processWarningIfSumDifference starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processWarningIfSumDifference starting";
		// warningIfSumDifference=A.B=X.Y (used to define which input columns to
		// compare - A & X are input recordset/datasource codes, B and Y are
		// system column codes; default = '"+dsDefault+"'')");
		// warningIfSumDifferenceMethod=X/Y (used to if the two sides should sum
		// all records (SumAll) and/or use to only use one (last) record
		// (LastRecord); default = '"+dsDefault+"'')");

		String warningIfSumDifference = getAndRemoveInstruction("warningIfSumDifference");
		String warningIfSumDifferenceMethod = getAndRemoveInstruction("warningIfSumDifferenceMethod");
		String warningIfSumMsg = getAndRemoveInstruction("warningIfSumMsg");

		String[] SDparts = warningIfSumDifference.split("=");
		String[] SDMparts = warningIfSumDifferenceMethod.split("/");
		if (SDparts.length != 2) {
			outputRecordSet
					.addWarningMessage("Missing parts to [warningIfSumDifference] - only see "
							+ SDparts.length
							+ " parts ("
							+ warningIfSumDifference + ")");
			return;
		}

		if (SDMparts.length != 2) {
			outputRecordSet
					.addWarningMessage("Missing parts to [warningIfSumDifferenceMethod] - only see "
							+ SDMparts.length
							+ " parts ("
							+ warningIfSumDifferenceMethod + ")");
			return;
		}

		String SDA = SDparts[0]; // warningIfSumDifference=A.B=X.Y
		String SDA1 = getSplit(SDA, "\\.", 0);
		String SDA2 = getSplit(SDA, "\\.", 1);

		String SDB = SDparts[1]; // warningIfSumDifference=A.B=X.Y
		String SDB1 = getSplit(SDB, "\\.", 0);
		String SDB2 = getSplit(SDB, "\\.", 1);

		String SDMA = SDMparts[0];
		String SDMB = SDMparts[1];

		RecordSet rsA = getRecordSet(SDA1);
		RecordSet rsB = getRecordSet(SDB1);

		String sumA = getAmount(rsA, SDA2, SDMA);
		String sumB = getAmount(rsB, SDB2, SDMB);

		if (!sumA.equals(sumB)) {
			String cid = outputRecordSet.getDataSourceCode();
			outputRecordSet.addWarningMessage(cid + ":: " + warningIfSumMsg
					+ " (" + sumA + " vs " + sumB + ")");
		} else {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "sumA (" + sumA + ") = sumB ("
							+ sumB + ")");
			outputRecordSet.addWarningMessage(msg);
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processWarningIfSumDiffer()
			throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processWarningIfSumDiffer starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processWarningIfSumDiffer starting";

		// warningIfSumDiffer=A=X where A is the col to sum and X is the amount
		// it should be (typically used to check a set of percentages add up to
		// 100)

		String warningIfSumDiffer = getAndRemoveInstruction("warningIfSumDiffer");
		String warningIfSumMsg = getAndRemoveInstruction("warningIfSumMsg");

		String[] parts = warningIfSumDiffer.split("=");
		if (parts.length < 2) {
			outputRecordSet
					.addWarningMessage("Missing parts to [warningIfSumDiffer]. Expected 'warningIfSumDiffer=xxx=yyy' - only see "
							+ parts.length
							+ " parts ("
							+ warningIfSumDiffer
							+ ")");
			return;
		}

		String col = parts[0].trim();
		String[] expectationAmts = parts[1].split(",");
		BigDecimal expectedAmt = new BigDecimal(expectationAmts[0].trim());
		BigDecimal tolerance = BigDecimal.ZERO;
		if(expectationAmts.length >= 2){
			tolerance = new BigDecimal(expectationAmts[1].trim());
		}
		BigDecimal sumAmt = new BigDecimal(getAmount(outputRecordSet, col, "SumAll"));
		if (expectedAmt.subtract(sumAmt).abs().compareTo(tolerance) > 0) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.OUTPUT_TO_LOG_ONLY, warningIfSumMsg + " ("
							+ sumAmt + " vs " + expectedAmt + " differ by more than "+tolerance+")");
			outputRecordSet.addWarningMessage(msg);
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected String getAmount(RecordSet rs, String col, String method,
			boolean percentage) throws VermilionNonFatalException {
		// used for processWarningIfSumDiffer and processWarningIfSumDifference
		int rowCount = rs.getRowCount();
		BigDecimal sumAmount = new BigDecimal(0);
		RecordSetMetaData rsmd = rs.getMetaData();
		int keyColN = getColumnIndexBySystemNm(rsmd, col);

		String format = (String) rsmd.getField(keyColN).getFieldDisplayFormat();
		if (keyColN < 0) 
			return new DecimalFormat(format).format(0);

		if (method.equalsIgnoreCase("SumAll")) {
			for (int row = 0; row < rowCount; row++) { 
				if(! rs.getRow(row).isEmpty(keyColN))
					sumAmount = sumAmount.add(rs.getRow(row).getDecimal(keyColN));
			}
		}
		if (method.equalsIgnoreCase("LastRecord")) {
			if(! rs.getRow(rowCount - 1).isEmpty(keyColN)){
				sumAmount = rs.getRow(rowCount - 1).getDecimal(keyColN);
			}
		}

		if (percentage)
			return new DecimalFormat(format).format(sumAmount
					.multiply(BigDecimal.valueOf(100)));
		else
			return new DecimalFormat(format).format(sumAmount);
	}

	protected String getAmount(RecordSet rs, String col, String method)
			throws VermilionNonFatalException {
		return getAmount(rs, col, method, false);
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processUnionDS() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processUnionDS starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processUnionDS starting";

		String unionDatasources = getAndRemoveInstruction("unionDatasources");

		String[] parts = unionDatasources.split(",");

		if (parts.length < 2) {
			outputRecordSet
					.addWarningMessage("Missing parts to [unionDS] - only see "
							+ parts.length + " parts but expect 2 ("
							+ unionDatasources + ")");
			return;
		}
		String dragByColumnNameStr = getAndRemoveInstruction("dragByColumnName");
		boolean dragByColumnNameBool = dragByColumnNameStr != null
				&& dragByColumnNameStr.equalsIgnoreCase("Y");
		for (int partNum = 0; partNum < parts.length; partNum++) {
			if (dragByColumnNameBool) {
				writeOutputRecordsetUsingColName(parts[partNum].trim());
			} else {
				writeOutputRecordset(parts[partNum].trim());
			}

		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processSumColumns() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processSumColumns starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processSumColumns starting";

		// sumColumns=col1,col2,resultCol

		String sumColumns = getAndRemoveInstruction("sumColumns");

		String[] colParts = sumColumns.split(",");

		if (colParts.length != 3) {
			outputRecordSet
					.addWarningMessage("Missing parts to [sumColumns] - only see "
							+ colParts.length + " parts (" + sumColumns + ")");
			return;
		}

		String col1Name = colParts[0];
		String col2Name = colParts[1];
		String resultColName = colParts[2];

		RecordSetMetaData rsmdIn = outputRecordSet.getMetaData();
		RecordSetMetaData rsmdOut = outputRecordSet.getMetaData();
		int col1 = getColumnIndexBySystemNm(rsmdIn, col1Name);
		int col2 = getColumnIndexBySystemNm(rsmdIn, col2Name);
		int resultCol = getColumnIndexBySystemNm(rsmdOut, resultColName);

		if (col1 < 0 || col2 < 0 || resultCol < 0)
			return;

		for (int row = 0; row < outputRecordSet.getRowCount(); row++) {
			RecordRow record = outputRecordSet.getRow(row);
			if (record.getString(col1) == null
					|| record.getString(col1).equals("")) {
				continue;
			}
			if (record.getString(col2) == null
					|| record.getString(col2).equals("")) {
				continue;
			}
			BigDecimal col1Value = record.getDecimal(col1);
			BigDecimal col2Value = record.getDecimal(col2);
			BigDecimal sumValue = col1Value.add(col2Value);
			outputRecordSet.setValue(row, resultCol, sumValue.toString());
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processSubtractColumns() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processSubtractColumns starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processSubtractColumns starting";
		// subtractColumns=col1,col2,resultCol

		String subtractColumns = getAndRemoveInstruction("subtractColumns");

		String[] colParts = subtractColumns.split(",");

		if (colParts.length != 3) {
			outputRecordSet
					.addWarningMessage("Missing parts to [subtractColumns] - only see "
							+ colParts.length
							+ " parts ("
							+ subtractColumns
							+ ")");
			return;
		}

		String col1Name = colParts[0];
		String col2Name = colParts[1];
		String resultColName = colParts[2];

		RecordSetMetaData rsmdIn = outputRecordSet.getMetaData();
		RecordSetMetaData rsmdOut = outputRecordSet.getMetaData();
		int col1 = getColumnIndexBySystemNm(rsmdIn, col1Name);
		int col2 = getColumnIndexBySystemNm(rsmdIn, col2Name);
		int resultCol = getColumnIndexBySystemNm(rsmdOut, resultColName);

		if (col1 < 0 || col2 < 0 || resultCol < 0)
			return;

		for (int row = 0; row < outputRecordSet.getRowCount(); row++) {
			RecordRow record = outputRecordSet.getRow(row);
			BigDecimal resultValue = new BigDecimal(0);
			if (!(record.getString(col1) == null || record.getString(col1)
					.equals(""))) {
				resultValue = resultValue.add(record.getDecimal(col1));
			}
			if (!(record.getString(col2) == null || record.getString(col2)
					.equals(""))) {
				resultValue = resultValue.subtract(record.getDecimal(col2));
			}
			outputRecordSet.setValue(row, resultCol, resultValue.toString());
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processSortColumn() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processSortColumn starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processSortColumn starting";

		String sortColumn = getAndRemoveInstruction("sortColumn");
		String sortOrder = getAndRemoveInstruction("sortOrder");

		//Because the input will be auto-dragged through to the output if no dragThrough is specified, by the parent,
		//the old code here was redundant in cases it was built for and broken when the output and inputs didn't match
		//in columns.

		RecordSetMetaData rsmd = outputRecordSet.getMetaData();
		int sortColNum = getColumnIndexBySystemNm(rsmd, sortColumn);
		if (sortColNum < 0)
			return;
		
		if (sortOrder.equalsIgnoreCase("asc")) {
			outputRecordSet.sort(sortColNum, true);
		} else {
			outputRecordSet.sort(sortColNum, false);
		}
		

	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processCheckOpeningBalance()
			throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processCheckOpeningBalance starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processCheckOpeningBalance starting";
		// subtractColumns=col1,col2,resultCol

		String subtractColumns = getAndRemoveInstruction("checkOpeningBalance");

		String[] colParts = subtractColumns.split(",");

		if (colParts.length != 2) {
			outputRecordSet
					.addWarningMessage("processCheckOpeningBalance expected 2 columns - it was passed "
							+ colParts.length);
			return;
		}

		String openBalanceCol = colParts[0];
		String closeBalanceCol = colParts[1];

		RecordSet rsIn = inputRecordSetList.getRecordSet(0); // this can only be
																// run against 1
																// record set

		RecordSetMetaData rsmdIn = rsIn.getMetaData();
		int openBalanceColIndex = getColumnIndexBySystemNm(rsmdIn,
				openBalanceCol);
		int closeBalanceColIndex = getColumnIndexBySystemNm(rsmdIn,
				closeBalanceCol);
		if (openBalanceColIndex < 0 || closeBalanceColIndex < 0)
			return;

		// this can only run against a record set with 1 row
		if (rsIn.getRowCount() != 1) {
			outputRecordSet
					.addWarningMessage("checkOpeningBalance only works when passed a RS with only 1 record - this RS has "
							+ rsIn.getRowCount() + " records");
			return;
		}

		RecordRow record = rsIn.getRow(0);
		BigDecimal openingBalance = new BigDecimal(0.0);
		BigDecimal closingBalance = new BigDecimal(0.0);

		// for the minute check i fthe opening balance = the closing balance on
		// the same RS
		if (!(record.getString(openBalanceColIndex) == null || record
				.getString(openBalanceColIndex).equals(""))) {
			openingBalance = record.getDecimal(openBalanceColIndex);
		}

		String term = null;
		String entityDate = null;

		for (int i = 0; i < parameterList.getParameterCount(); i++) {
			ParameterRecord rec = null;
			try {
				rec = parameterList.getParameterRecord(i);
				if (rec.getParameterName().toUpperCase().equals("TERM")) {
					term = rec.getFieldValue(0);
				}
				if (rec.getParameterName().toUpperCase().equals("DATE")) {
					entityDate = rec.getFieldValue(0);
				}
			} catch (VermilionFatalException e) {
				e.printStackTrace();
			}
		}

		if (term == null || entityDate == null) {
			outputRecordSet
					.addWarningMessage("Couldn't pickup either the date or the term entity value - the DS must be run with these as parameters");
			return;
		}

		String previousDate = getPreviousDate(term, entityDate);

		RecordSet prevDS;
		try {
			prevDS = getPreviousDatasource(rsIn.getDataSourceCode(), term,
					previousDate);
		} catch (VermilionFatalException e) {
			e.printStackTrace();
			outputRecordSet
					.addWarningMessage("Couldn't pickup the DS for the previous term. "
							+ e.getMessage());
			return;
		}

		if (prevDS == null || prevDS.getRowCount() == 0) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processCheckOpeningBalance no previous datasource found for term "
							+ term + " and date " + previousDate);
			outputRecordSet.addWarningMessage(msg);
			return;
		}

		RecordRow prevRecord = prevDS.getRow(0);
		if (!(prevRecord.getString(closeBalanceColIndex) == null || prevRecord
				.getString(closeBalanceColIndex).equals(""))) {
			closingBalance = prevRecord.getDecimal(closeBalanceColIndex);
		}

		if (!(openingBalance.compareTo(closingBalance) == 0)) { // compareTo
																// returns
																// -1/0/1 for
																// </=/>
			outputRecordSet.addWarningMessage("The opening balance ("
					+ openingBalance
					+ ") is not equal to the previous terms closing balance ("
					+ closingBalance + ")");
		}
	}

	private String getPreviousDate(String term, String entityDate)
			throws VermilionNonFatalException {
		//used in processComparePreviousTerm
		//used in processCheckOpeningBalance
		// "31 December 2009 00:00:00"
		SimpleDateFormat dateFormat = new SimpleDateFormat("dd-MMM-yyyy");
		SimpleDateFormat sysdateFormat = new SimpleDateFormat(
				"dd MMMM yyyy 00:00:00");
		Date endDate = null;
		try {
			endDate = dateFormat.parse(entityDate);
		} catch (ParseException e) {
			outputRecordSet.addWarningMessage("Error parsing date : "
					+ entityDate);
			return null;
		}

		// get the end date of the previous term
		Date endDateOfPreviousTerm = DateHelper.getStartDateOfTerm(endDate,
				term, null, true);

		return sysdateFormat.format(endDateOfPreviousTerm);
	}
	
	//This method is overloaded. Are both used. Can they be merged? There is redundant code that needs to be cleaned up here. - Ata
	private RecordSet getPreviousDatasource(String dsCode, String term,
			String endDate, String fund) throws VermilionNonFatalException,
			VermilionFatalException {
		//used in processCheckOpeningBalance
		
		// Now get the datasource for the previous term (if it exists)
		int wfeUserID = CoreBD.getUserIdByUserType(outputRecordSet.getTid(),
				UserConstants.USER_TYPE_WFENGINE);

		ParameterList parameterList = new ParameterList(
				outputRecordSet.getTid(), wfeUserID);
		ParameterRecord param = null;
		try {
			param = new ParameterRecord(outputRecordSet.getTid(),
					"SYSTEM_DATE", endDate);
		} catch (VermilionNonFatalException e) {
			// this will happen if it can't find that entity - meaning that this
			// may be the 1st dataset
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "VermilionNonFatalException "
							+ e.getMessage());
			outputRecordSet.addWarningMessage(msg);
			return null;
		}
		parameterList.addParameterRecord(param);
		if (term != null) {
			param = new ParameterRecord(outputRecordSet.getTid(), "TERM", term);
			parameterList.addParameterRecord(param);
		}
		if (fund != null) {
			param = new ParameterRecord(outputRecordSet.getTid(), "FUND", fund);
			parameterList.addParameterRecord(param);
		}

		DynamicDGSCall dgsHelper = new DynamicDGSCall(outputRecordSet.getTid());
		RecordSet recordSet = dgsHelper.getDGSDataByDstCode(dsCode,
				parameterList);
		return recordSet;
	}

	private RecordSet getPreviousDatasource(String dsCode, String term,
			String endDate) throws VermilionNonFatalException,
			VermilionFatalException {

		// Now get the datasource for the previous term (if it exists)
		int wfeUserID = CoreBD.getUserIdByUserType(outputRecordSet.getTid(),
				UserConstants.USER_TYPE_WFENGINE);

		ParameterList parameterList = new ParameterList(
				outputRecordSet.getTid(), wfeUserID);
		ParameterRecord param = null;
		try {
			param = new ParameterRecord(outputRecordSet.getTid(),
					"SYSTEM_DATE", endDate);
		} catch (VermilionNonFatalException e) {
			// this will happen if it can't find that entity - meaning that this
			// may be the 1st dataset
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "VermilionNonFatalException "
							+ e.getMessage());
			outputRecordSet.addWarningMessage(msg);
			return null;
		}
		parameterList.addParameterRecord(param);
		if (term != null) {
			param = new ParameterRecord(outputRecordSet.getTid(), "TERM", term);
			parameterList.addParameterRecord(param);
		}

		DynamicDGSCall dgsHelper = new DynamicDGSCall(outputRecordSet.getTid());
		RecordSet recordSet = dgsHelper.getDGSDataByDstCode(dsCode,
				parameterList);

		return recordSet;
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processCountBy() throws VermilionNonFatalException,
			VermilionFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processCountBy starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processCountBy starting";

		String countByField = getAndRemoveInstruction("countByField");
		String countByCount = getAndRemoveInstruction("countByCount");
		String countByTotal = getAndRemoveInstruction("countByTotal");
		String countOnOutput = getAndRemoveInstruction("countOnOutput");


		runPos="processCountBy: checking countByField and countByCount";
		if (countByField.equals("-") || countByCount.equals("-")) {
			outputRecordSet
					.addWarningMessage("Missing instructions. If you are using the count command you need to provide both countByField and countByCount instructions");
			return;
		}
		runPos="processCountBy: checking countOnOutput";
		
		RecordSet primaryRS = inputRecordSetList.getRecordSet(0);
		if (countOnOutput.equalsIgnoreCase("Y")) {
			primaryRS = outputRecordSet;
		}

		runPos="processCountBy: getting datasources";
		RecordSetMetaData rsmdPrimary = primaryRS.getMetaData();
		RecordSetMetaData rsmdOut = outputRecordSet.getMetaData();
		int primaryKeyCol = getColumnIndexBySystemNm(rsmdPrimary, countByField);
		int resultCol = getColumnIndexBySystemNm(rsmdOut, countByCount);
		int totalCol = -1;
		if (!countByTotal.equals("-")) {
			totalCol = getColumnIndexBySystemNm(rsmdOut, countByTotal);
		}
		if (primaryKeyCol < 0 || resultCol < 0) {
			return;
		}

		runPos="processCountBy: counting rows";
		int iCounter = 0;
		for (int row = 0; row < primaryRS.getRowCount(); row++) {
			if (row > 0
					&& !primaryRS.getString(row, primaryKeyCol)
							.equalsIgnoreCase(
									primaryRS.getString(row - 1, primaryKeyCol)
											.toString())) {
				iCounter = 0;
			}
			iCounter++;
			outputRecordSet.setValue(row, resultCol, iCounter);
			for (int rowToFixTotal = row; totalCol > -1
					&& rowToFixTotal > row - iCounter; rowToFixTotal--) {
				outputRecordSet.setValue(row, totalCol, iCounter);
			}
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processJoinOnKey() throws VermilionNonFatalException,
			VermilionFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processJoin starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processJoin starting";

		String joinPrimaryDS = getAndRemoveInstruction("joinPrimaryDS");
		String joinSecondaryDS = getAndRemoveInstruction("joinSecondaryDS");
		String joinPrimaryKeyColumn = getAndRemoveInstruction("joinPrimaryKeyColumn");
		String joinSecondaryKeyColumn = getAndRemoveInstruction("joinSecondaryKeyColumn");
		String joinSecondaryEntityName = getAndRemoveInstruction("joinSecondaryEntityName");
		String[] joinSecondaryValueColumns = getAndRemoveInstruction(
				"joinSecondaryValueColumn").split(",");
		String[] joinResultingColumns = getAndRemoveInstruction(
				"joinResultingColumn").split(",");
		String joinUnfoundKeysValue = getAndRemoveInstruction("joinUnfoundKeysValue");
		String joinDuplicates = getAndRemoveInstruction("joinDuplicates");
		boolean isJoinOnEntity = false;
		String entityKeyValue = null;

		boolean missingInstructions = false;

		if (joinPrimaryDS.equals("-") || joinSecondaryDS.equals("-")
				|| joinPrimaryKeyColumn.equals("-")
				|| joinSecondaryValueColumns[0].equals("-")
				|| joinResultingColumns[0].equals("-")) {
			outputRecordSet
					.addWarningMessage("Missing instructions. If you are using the join command you need to provide all 5 "
							+ "mandatory instructions: joinPrimaryDS, joinSecondaryDS, joinPrimaryKeyColumn, joinSecondaryValueColumn, joinResultingColumn");
			missingInstructions = true;
		}

		if (joinSecondaryKeyColumn.equals("-")
				&& joinSecondaryEntityName.equals("-")) {
			outputRecordSet
					.addWarningMessage("Missing instructions. You must provide either the joinSecondaryEntityName or the joinSecondaryKeyColumn");
			missingInstructions = true;
		}

		if (missingInstructions) {
			return;
		}

		// we are joining on an entity List Name
		if (joinSecondaryKeyColumn.equals("-")) {
			ArrayList<String> pl = parameterList.getParameterNames();
			if (!pl.contains(joinSecondaryEntityName)) {
				outputRecordSet
						.addWarningMessage("When joining on an entity the entity must be on the datasource. Please add entity "
								+ joinSecondaryEntityName + " to this DS");
				return;
			} else {
				isJoinOnEntity = true;
				entityKeyValue = parameterList.getParameterRecordByName(
						joinSecondaryEntityName).getKeyValue();
			}
		}

		RecordSet primaryRS = null;
		RecordSet secondaryRS = null;

		// outputRecordSet.addWarningMessage("Looking for primary DS " +
		// joinPrimaryDS);
		// outputRecordSet.addWarningMessage("Looking for secondary DS " +
		// joinSecondaryDS);

		//if instructions has PrimaryDS=Output, then use the output record set up to now, instead of pulling from one of the inputs.
		if(joinPrimaryDS.equalsIgnoreCase("<Output>")){
			joinPrimaryDS="";
			primaryRS=outputRecordSet;
		}
		
		for (int r = 0; r < inputRecordSetList.getRecordSetCount(); r++) {
			// outputRecordSet.addWarningMessage("Found " +
			// inputRecordSetList.getDataSourceCode(r));
			if (inputRecordSetList.getDataSourceCode(r) != null
					&& inputRecordSetList.getDataSourceCode(r).equals(
							joinPrimaryDS)) {
				primaryRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getRecordSetCode(r) != null
					&& inputRecordSetList.getRecordSetCode(r).equals(
							joinPrimaryDS)) {
				primaryRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getDataSourceName(r) != null
					&& inputRecordSetList.getDataSourceName(r).equals(
							joinPrimaryDS)) {
				primaryRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getRecordSetName(r) != null
					&& inputRecordSetList.getRecordSetName(r).equals(
							joinPrimaryDS)) {
				primaryRS = inputRecordSetList.getRecordSet(r);
			}

			if (inputRecordSetList.getDataSourceCode(r) != null
					&& inputRecordSetList.getDataSourceCode(r).equals(
							joinSecondaryDS)) {
				secondaryRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getRecordSetCode(r) != null
					&& inputRecordSetList.getRecordSetCode(r).equals(
							joinSecondaryDS)) {
				secondaryRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getDataSourceName(r) != null
					&& inputRecordSetList.getDataSourceName(r).equals(
							joinSecondaryDS)) {
				secondaryRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getRecordSetName(r) != null
					&& inputRecordSetList.getRecordSetName(r).equals(
							joinSecondaryDS)) {
				secondaryRS = inputRecordSetList.getRecordSet(r);
			}
		}

		boolean missingRecordset = false;
		if (primaryRS == null) {
			outputRecordSet.addWarningMessage("Unable to find the primary RS");
			missingRecordset = true;
		}

		if (secondaryRS == null) {
			outputRecordSet
					.addWarningMessage("Unable to find the secondary RS");
			missingRecordset = true;
		}

		if (missingRecordset) {
			return;
		}

        if(outputRecordSet.getRowCount()==0) {
        	writeOutputRecordset(inputRecordSetList, primaryRS.getDataSourceCode());
        }

		RecordSetMetaData rsmdPrimary = primaryRS.getMetaData();
		RecordSetMetaData rsmdSecondary = secondaryRS.getMetaData();
		RecordSetMetaData rsmdOut = outputRecordSet.getMetaData();
		int primaryKeyCol = getColumnIndexBySystemNm(rsmdPrimary,
				joinPrimaryKeyColumn);

		if (primaryKeyCol < 0) {
			outputRecordSet
					.addWarningMessage("Unable to find the primary key column "
							+ joinPrimaryKeyColumn + " in " + joinPrimaryDS);
			return;
		}

		int secondaryKeyCol = -1;

		if (!isJoinOnEntity) {

			secondaryKeyCol = getColumnIndexBySystemNm(rsmdSecondary,
					joinSecondaryKeyColumn);

			if (secondaryKeyCol < 0) {
				outputRecordSet
						.addWarningMessage("Unable to find the secondary key column "
								+ joinSecondaryKeyColumn
								+ " in "
								+ joinSecondaryDS);
				return;
			}

		}

		for (int i = 0; i < joinSecondaryValueColumns.length; i++) {
			int secondaryValueCol = -1;
			if (joinSecondaryValueColumns[i].startsWith("/monthsSince(")) {
				// if we see the formula /monthsSince that means a date
				// substitution variable is passed that will indicate the column
				// to find

				String strStartDate = getSplit(joinSecondaryValueColumns[i],
						"\\$", 1);
				strStartDate = StringHelper.getResolveSubVars("$"
						+ strStartDate + "$", parameterList, true);

				String strSysDate = parameterList.getParameterRecordByCode(
						"SYSTEM_DATE").getFieldValue("SysDate");

				int monthsBetween = -1;
				try {
					monthsBetween = Math.max(1, DateHelper
							.monthsBetween(strStartDate, "yyyyMMdd",
									strSysDate, "dd-MMM-yyyy"));
				} catch (ParseException e) {
					outputRecordSet
							.addWarningMessage("Error parsing argument date or reporting date: "
									+ e);
					return;
				}

				secondaryValueCol = rsmdSecondary.getColumnCount() - 1;
				for (int col = 0; col < rsmdSecondary.getColumnCount(); col++) {
					if (StringHelper.isNumeric(rsmdSecondary.getField(col)
							.getFieldName().toString())) {
						if (monthsBetween < Integer.parseInt(rsmdSecondary
								.getField(col).getFieldName().toString())) {
							// outputRecordSet.addWarningMessage(monthsBetween +
							// "
							// is less than " +
							// Integer.parseInt(rsmdSecondary.getField(col).getFieldName().toString()));
							secondaryValueCol = col - 1;
							break;
						}// else outputRecordSet.addWarningMessage(monthsBetween
							// + " is not less than " +
							// Integer.parseInt(rsmdSecondary.getField(col).getFieldName().toString()));
					}
				}
			} else {
				secondaryValueCol = getColumnIndexBySystemNm(rsmdSecondary,
						joinSecondaryValueColumns[i]);
			}

			if (secondaryValueCol < 0) {
				outputRecordSet.addWarningMessage("Could not find column "
						+ joinSecondaryValueColumns[i] + " in "
						+ joinSecondaryDS);
				continue;
			}
			int resultCol = getColumnIndexBySystemNm(rsmdOut,
					joinResultingColumns[i]);

			if (resultCol < 0) {
				outputRecordSet.addWarningMessage("Could not find column "
						+ joinResultingColumns[i] + " in output recordset");
				continue;
			}

			HashMap<String, String> secondaryRSMap = new HashMap<String, String>();
			HashMap<String, HashSet<String>> secondaryRSMultiMap = new HashMap<String, HashSet<String>>();

			if (isJoinOnEntity) { // we are joining on an entity List Name
				// Load the secondary recordset into a hashmap
				for (int row = 0; row < secondaryRS.getRowCount(); row++) {
					RecordRow record = secondaryRS.getRow(row);
					secondaryRSMap.put(entityKeyValue,
							record.getString(secondaryValueCol));
				}
			} else { // match records on the key fields
				for (int row = 0; row < secondaryRS.getRowCount(); row++) {
					RecordRow record = secondaryRS.getRow(row);
					if (record.getString(secondaryKeyCol) == null
							|| record.getString(secondaryValueCol).equals("")) {
						continue;
					}

					String key = record.getString(secondaryKeyCol)
							.toUpperCase();
					String value = record.getString(secondaryValueCol);
					if (joinDuplicates.equalsIgnoreCase("Y")) {
						if (secondaryRSMultiMap.containsKey(key)) {
							secondaryRSMultiMap.get(key).add(value);
							// outputRecordSet.addWarningMessage("added value '"+value+"' to existing key '"+key+"'");
						} else {
							HashSet<String> newSet = new HashSet<String>();
							newSet.add(value);
							secondaryRSMultiMap.put(key, newSet);
							// outputRecordSet.addWarningMessage("added new key '"+key+"' with value '"+value+"'");
						}
					} else {
						secondaryRSMap.put(key, value);
					}
				}
			}

			// Output the results
			for (int row = 0; row < outputRecordSet.getRowCount(); row++) {
				runPos = "Updating row " + row;
				RecordRow record = outputRecordSet.getRow(row);
				String key = record.getString(primaryKeyCol).toUpperCase();
				if (joinDuplicates.equalsIgnoreCase("Y")) {
					HashSet<String> values = secondaryRSMultiMap.get(key);
					if (values == null) {
						outputRecordSet.setValue(row, resultCol,
								joinUnfoundKeysValue);
					} else if (values.size() <= 1) {
						// outputRecordSet.addWarningMessage("key '"+key+"' had "+values.size()+" values");
						String value = (String) values.toArray()[0];
						if (value == null) {
							value = joinUnfoundKeysValue;
						}
						outputRecordSet.setValue(row, resultCol, value);
					} else {
						// outputRecordSet.addWarningMessage("key '"+key+"' had "+values.size()+" values");
						Iterator<String> valuesIterator = values.iterator();
						while (valuesIterator.hasNext()) {
							String value = valuesIterator.next();
							// outputRecordSet.addWarningMessage("iterator provided value "+value);
							if (value == null) {
								value = joinUnfoundKeysValue;
							}
							outputRecordSet.setValue(row, resultCol, value);
							if (valuesIterator.hasNext()) {
								outputRecordSet.addRow();
								for (int rowToMove = outputRecordSet
										.getRowCount() - 2; rowToMove >= row; rowToMove--) {
									outputRecordSet.setRow(rowToMove + 1,
											outputRecordSet.getRow(rowToMove));
								}
								row++;
							}
						}
					}
				} else {
					String value = secondaryRSMap.get(key);
					if (value == null)
						value = joinUnfoundKeysValue;
					outputRecordSet.setValue(row, resultCol, value);
				}
			}
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processJoinWithoutKey() throws VermilionNonFatalException,
			VermilionFatalException {
		// this is one of the few commands that must be run on the input RST,
		// otherwise you don't know the number of columns in each record
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processJoinWithoutKey starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processJoinWithoutKey starting";

		String joinLeftDS = getAndRemoveInstruction("joinLeftDS");
		String joinRightDS = getAndRemoveInstruction("joinRightDS");

		if (joinLeftDS.equals("-") || joinRightDS.equals("-")) {
			outputRecordSet
					.addWarningMessage("Missing instructions. If you are using the joinLeftDS command you need to provide the joinRightDS instructions");
			return;
		}

		RecordSet leftRS = null;
		RecordSet rightRS = null;

		for (int r = 0; r < inputRecordSetList.getRecordSetCount(); r++) {
			if (inputRecordSetList.getDataSourceCode(r) != null
					&& inputRecordSetList.getDataSourceCode(r).equals(
							joinLeftDS)) {
				leftRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getRecordSetCode(r) != null
					&& inputRecordSetList.getRecordSetCode(r)
							.equals(joinLeftDS)) {
				leftRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getDataSourceName(r) != null
					&& inputRecordSetList.getDataSourceName(r)
					.equals(joinLeftDS)) {
				leftRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getRecordSetName(r) != null
					&& inputRecordSetList.getRecordSetName(r)
					.equals(joinLeftDS)) {
				leftRS = inputRecordSetList.getRecordSet(r);
			}
			if (inputRecordSetList.getDataSourceCode(r) != null
					&& inputRecordSetList.getDataSourceCode(r).equals(
							joinRightDS)) {
				rightRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getRecordSetCode(r) != null
					&& inputRecordSetList.getRecordSetCode(r).equals(
							joinRightDS)) {
				rightRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getDataSourceName(r) != null
					&& inputRecordSetList.getDataSourceName(r).equals(
							joinRightDS)) {
				rightRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getRecordSetName(r) != null
					&& inputRecordSetList.getRecordSetName(r).equals(
							joinRightDS)) {
				rightRS = inputRecordSetList.getRecordSet(r);
			}
		}

		boolean missingRecordset=false;
		if (leftRS == null) {
			outputRecordSet.addWarningMessage("Unable to find the left RS");
			missingRecordset=true;
		}

		if (rightRS == null) {
			outputRecordSet.addWarningMessage("Unable to find the right RS");
			missingRecordset=true;
		}
		if (missingRecordset){
			return;
		}

		writeOutputRecordset(joinLeftDS);

		if (leftRS.getRowCount() < rightRS.getRowCount()) {
			for (int i = 0; i < (rightRS.getRowCount() - leftRS.getRowCount()); i++) {
				outputRecordSet.addRow();
			}
		}

		for (int row = 0; row < rightRS.getRowCount(); row++) {
			RecordRow record = rightRS.getRow(row);
			for (int col = 0; col < record.getColumnCount(); col++) {
				outputRecordSet.setValue(row, col + leftRS.getColumnCount(),
						record.getString(col));
			}
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processReplaceNamesDS() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processReplaceNamesDS starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processReplaceNamesDS starting";

		String replaceNamesDS = getAndRemoveInstruction("replaceNamesDS");
		String replaceNamesColumn = getAndRemoveInstruction("replaceNamesColumn");

		if (replaceNamesDS.equals("-") || replaceNamesColumn.equals("-")) {
			outputRecordSet
					.addWarningMessage("Missing instructions. If you are using the replaceNames command you need to provide both instructions ");
			return;
		}

		String[] dscode = replaceNamesDS.split(",");

		if (dscode.length != 2) {
			outputRecordSet
					.addWarningMessage("You need to specify 2 datasources in the replaceNamesDS instruction");
			return;
		}

		RecordSet primaryRS = null;
		RecordSet secondaryRS = null;

		for (int r = 0; r < inputRecordSetList.getRecordSetCount(); r++) {
			if (inputRecordSetList.getDataSourceCode(r) != null
					&& inputRecordSetList.getDataSourceCode(r)
							.equals(dscode[0])) {
				primaryRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getRecordSetCode(r) != null
					&& inputRecordSetList.getRecordSetCode(r)
					.equals(dscode[0])) {
				primaryRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getDataSourceName(r) != null
					&& inputRecordSetList.getDataSourceName(r)
					.equals(dscode[0])) {
				primaryRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getRecordSetName(r) != null
					&& inputRecordSetList.getRecordSetName(r)
					.equals(dscode[0])) {
				primaryRS = inputRecordSetList.getRecordSet(r);
			}
			if (inputRecordSetList.getDataSourceCode(r) != null
					&& inputRecordSetList.getDataSourceCode(r)
							.equals(dscode[1])) {
				secondaryRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getRecordSetCode(r) != null
					&& inputRecordSetList.getRecordSetCode(r)
					.equals(dscode[1])) {
				secondaryRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getDataSourceName(r) != null
					&& inputRecordSetList.getDataSourceName(r)
					.equals(dscode[1])) {
				secondaryRS = inputRecordSetList.getRecordSet(r);
			} else if (inputRecordSetList.getRecordSetName(r) != null
					&& inputRecordSetList.getRecordSetName(r)
					.equals(dscode[1])) {
				secondaryRS = inputRecordSetList.getRecordSet(r);
			}
		}

		boolean missingRecordSet=false;
		if (primaryRS == null) {
			outputRecordSet.addWarningMessage("Unable to find the primary RS");
			missingRecordSet=true;
		}

		if (secondaryRS == null) {
			outputRecordSet
					.addWarningMessage("Unable to find the secondary RS");
			missingRecordSet=true;
		}
		
		if(missingRecordSet){
			return;
		}

		// Load the secondary recordset into a hashmap
		HashMap<String, String> nameMap = new HashMap<String, String>();
		for (int row = 0; row < secondaryRS.getRowCount(); row++) {
			RecordRow record = secondaryRS.getRow(row);
			nameMap.put(record.getString(0), record.getString(1));
		}

		// Get the number of the 'name' column to replace
		RecordSetMetaData rsmdPrimary = primaryRS.getMetaData();
		int nameCol = getColumnIndexBySystemNm(rsmdPrimary, replaceNamesColumn);
		if (nameCol < 0) {
			return;
		}

		// Write out the records and replace any names that match with the new
		// name
		RecordSet irs = inputRecordSetList
				.getRecordSetByDataSourceCode(dscode[0]);
		RecordRow rw = null;
		int rowCount = irs.getRowCount();
		String key = null;
		int newRow;

		for (int rec = 0; rec < rowCount; rec++) {
			rw = irs.getRow(rec);
			newRow = outputRecordSet.addRow(rw);
			key = rw.getString(nameCol);

			if (nameMap.containsKey(key)) {
				outputRecordSet.setValue(newRow, nameCol, nameMap.get(key));
			}
		}
	}
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	// processGroupColumns0 is the original grouping - expecting a specific output recordset
	protected void processGroupColumns0() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processGroupColumns starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processGroupColumns starting";

		String groupColumns = getAndRemoveInstruction("groupColumns");
		String groupTotals = getAndRemoveInstruction("groupTotals");

		if (groupColumns.equals("-") || groupTotals.equals("-")) {
			outputRecordSet
					.addWarningMessage("Missing instructions. If you are using the group command you need to provide both groupColumns and groupTotal instructions ");
			return;
		}

		runPos="Substituting variables on instructions";
		String resolveSubVarOutputLang = getAndRemoveInstruction("resolveSubVarOutputLang");
		groupColumns = ComponentUtility.parseText(outputRecordSet.getTid(),
				StringHelper.getResolveSubVars(groupColumns,
						parameterList, true), resolveSubVarOutputLang);
		

		groupTotals = ComponentUtility.parseText(outputRecordSet.getTid(),
				StringHelper.getResolveSubVars(groupTotals,
						parameterList, true), resolveSubVarOutputLang);
		
		
		runPos="Getting recordset";
		RecordSet inputRS = inputRecordSetList.getRecordSet(0);
		RecordSetMetaData rsmdIn = inputRS.getMetaData();

		// Get the columns to group by
		runPos="Itemizing columns";
		String[] groups = groupColumns.split(",");
		int numOfGroups = groups.length;
		Integer[] groupColNum = new Integer[numOfGroups];

		for (int i = 0; i < numOfGroups; i++) {
			groupColNum[i] = getColumnIndexBySystemNm(rsmdIn, groups[i]);
			if (groupColNum[i] < 0) {
				return;
			}
		}

		// Get the columns to total
		runPos="Itemizing groups";
		String[] totals = groupTotals.split(",");
		int numOfTotals = totals.length;
		Integer[] totalColNum = new Integer[numOfTotals];

		for (int i = 0; i < numOfTotals; i++) {
			totalColNum[i] = getColumnIndexBySystemNm(rsmdIn, totals[i]);
			if (totalColNum[i] < 0) {
				return;
			}
		}

		// Need to sort the input recordset in the order of the last column (we
		// can only do 1 column)
		runPos="Sorting recorset";
		inputRS.sort(groupColNum[numOfGroups - 1]);

		// now read the input RS and sum the totals
		int rowCount = inputRS.getRowCount();
		int c;
		RecordRow rw = null;
		boolean recordsEqual = true;
		boolean firstOutputRecord = true;
		BigDecimal[] totalAmt = new BigDecimal[numOfTotals];
		String[] previousRow = new String[numOfGroups];
		String[] currentRow = new String[numOfGroups];

		for (int rec = 0; rec < rowCount; rec++) {
			runPos="Looping ("+rec+")";
			rw = inputRS.getRow(rec);

			for (int col = 0; col < numOfGroups; col++) {
				c=totalColNum[col];
				currentRow[col] = rw.getString(groupColNum[col]);
			}

			recordsEqual = true;

			for (int i = 0; i < currentRow.length; i++) {
				if (!currentRow[i].equals(previousRow[i]))
					recordsEqual = false;
			}

			// if the group columns haven't changed then carry on totalling
			if (recordsEqual) {
				runPos="processing recordEqual";

				for (int col = 0; col < numOfTotals; col++) {
					c=totalColNum[col];
					runPos="processing recordEqual3006 ("+col+")["+c+"]";
					if (totalAmt[col] == null) {
						totalAmt[col] = new BigDecimal(0);
					}
					if ( rw.isNull(c)==false && rw.isEmpty(c)==false ) {
						totalAmt[col] = totalAmt[col].add(rw.getDecimal(c));
					}
				}
			} 
			else {
				runPos="processing recordNotEqual(a)";
				// if they have write out the record, unless it's the 1st one
				if (firstOutputRecord) {
					firstOutputRecord = false;
				} else {
					writeGroupRecord0(numOfGroups, numOfTotals, previousRow, totalAmt);
				}
				
				runPos="processing recordNotEqual(b)";
				for (int col = 0; col < numOfGroups; col++) {
					previousRow[col] = currentRow[col];
				}
				runPos="processing recordNotEqual(c)";
				String ft;
				for (int col = 0; col < numOfTotals; col++) {
					c=totalColNum[col];
					ft=rsmdIn.getField(c).getFieldType();
					runPos="processing recordNotEqual(d)["+col+"]["+rsmdIn.getField(c).getFieldName()+"]["+ft+"]";
					if ( ft.equals("decimal") || ft.equals("integer") ) {
						if ( rw.isNull(c) || rw.isEmpty(c)) {
							totalAmt[col]=new BigDecimal(0);
						}
						else {
							runPos="processing recordNotEqual(e)["+col+"]["+rsmdIn.getField(c).getFieldName()+"]="+rw.getDecimal(c);
							totalAmt[col] = rw.getDecimal(c);
						}
					}
				}
			}
		}
		// Now write out the last record
		runPos="Writing out last record";
		writeGroupRecord0(numOfGroups, numOfTotals, previousRow, totalAmt);
	}
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	private void writeGroupRecord0(int numOfGroups, int numOfTotals,
			String[] previousRow, BigDecimal[] totalAmt)
			throws VermilionNonFatalException {
		int outputRowNum = outputRecordSet.addRow();
		//used in processGroupColumns
		
		// Write out the group fields
		int outputColNum = 0;
		for (int col = 0; col < numOfGroups; col++) {
			runPos="writeGroupRecord3064["+col+"]= "+previousRow[col];
			outputRecordSet.setValue(outputRowNum, outputColNum++,
					previousRow[col]);
		}

		// now write out the summed values
		for (int col = 0; col < numOfTotals; col++) {
			if (totalAmt[col] != null)
				runPos="writeGroupRecord3056["+col+"]= "+totalAmt[col];
				outputRecordSet.setValue(outputRowNum, outputColNum++,
						totalAmt[col].toString());
		}
		runPos="Done writeGroupRecord";
	}
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processGroupColumns() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering", WarningMessage.DONT_OUTPUT, "processGroupColumns starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processGroupColumns starting";

		String groupColumns  = getAndRemoveInstruction("groupColumns");  // grouping columns
		String groupTotals   = getAndRemoveInstruction("groupTotals");   // columns to total up
		String groupColumns2 = getAndRemoveInstruction("groupColumns2"); // columns to be included/dragged through in the result, but not grouped by
		if (groupColumns.equals("-") || groupTotals.equals("-")) {
			outputRecordSet.addWarningMessage("Missing instructions. If you are using the group command you need to provide both groupColumns and groupTotal instructions ");
			return;
		}

		runPos="Substituting variables on instructions";
		String resolveSubVarOutputLang = getAndRemoveInstruction("resolveSubVarOutputLang");
		groupColumns = ComponentUtility.parseText(outputRecordSet.getTid(),
				StringHelper.getResolveSubVars(groupColumns,
						parameterList, true), resolveSubVarOutputLang);
		
		if ( !groupColumns2.equals("") ) {
			groupColumns2 = ComponentUtility.parseText(outputRecordSet.getTid(),
				StringHelper.getResolveSubVars(groupColumns2,
						parameterList, true), resolveSubVarOutputLang);
		}

		groupTotals = ComponentUtility.parseText(outputRecordSet.getTid(),
				StringHelper.getResolveSubVars(groupTotals,
						parameterList, true), resolveSubVarOutputLang);
		
		runPos="Getting recordset";
		RecordSet inputRS = inputRecordSetList.getRecordSet(0);
		RecordSetMetaData rsmdIn = inputRS.getMetaData();
		int colCount = rsmdIn.getColumnCount();

		// Get the columns to group by
		runPos="Itemizing columns";
		String[] groups = groupColumns.split(",");
		int numOfGroups = groups.length;
		Integer[] groupColNum = new Integer[numOfGroups];

		String[] groups2 = new String[0];
		int numOfGroups2 = 0;
		if ( !groupColumns2.equals("") ) {
			groups2 = groupColumns2.split(",");
			numOfGroups2 = groups2.length;
		}
		Integer[] groupColNum2 = new Integer[numOfGroups2];

		for (int i = 0; i < numOfGroups; i++) {
			groupColNum[i] = getColumnIndexBySystemNm(rsmdIn, groups[i]);
			if (groupColNum[i] < 0) {
				return;
			}
		}
		
		for (int i = 0; i < numOfGroups2; i++) {
			groupColNum2[i] = getColumnIndexBySystemNm(rsmdIn, groups2[i]);
			if (groupColNum2[i] < 0) {
				return;
			}
		}

		// Get the columns to total
		runPos="Itemizing groups";
		String[] totals = groupTotals.split(",");
		int numOfTotals = totals.length;
		Integer[] totalColNum = new Integer[numOfTotals];

		for (int i = 0; i < numOfTotals; i++) {
			totalColNum[i] = getColumnIndexBySystemNm(rsmdIn, totals[i]);
			if (totalColNum[i] < 0) {
				return;
			}
		}

		// Need to sort the input recordset in the order of the last column (we
		// can only do 1 column)
		runPos="Sorting recorset";
		inputRS.sort(groupColNum[numOfGroups - 1]);

		// now read the input RS and sum the totals
		boolean recordsEqual      = true;
		boolean firstOutputRecord = false;
		int rowCount              = inputRS.getRowCount();
		int c;
		RecordRow rw              = null;
		BigDecimal[] totalAmt     = new BigDecimal[colCount];
		String[] previousRow      = new String[colCount];
		String[] currentRow       = new String[colCount];

		for (int rec = 0; rec < rowCount; rec++) {
			runPos="Looping3155 ("+rec+")";
			rw = inputRS.getRow(rec);

			for (int col = 0; col < numOfGroups; col++) {
				c=groupColNum[col];
				currentRow[c] = rw.getString(c);
			}
			for (int col = 0; col < numOfGroups2; col++) {
				c=groupColNum2[col];
				currentRow[c] = rw.getString(c);
			}
			
			recordsEqual = true;

			runPos="Looping3165 ("+rec+")";
			String a,b;
			for (int col = 0; col < numOfGroups; col++) {
				c=groupColNum[col];
				runPos="Looping3165 ("+col+")("+c+")";
				a=currentRow[c];
				b=previousRow[c];
				if (a==null ) a="";
				if (b==null ) b="";
			
				if (!a.equals(b)) {
					recordsEqual = false;
				}
			}

			// if the group columns haven't changed then carry on totalling
			if (recordsEqual) {
				runPos="processing recordEqual";
				for (int col = 0; col < numOfTotals; col++) {
					c=totalColNum[col];
					runPos="processing recordEqual3006 ("+col+")["+c+"]";
					if (totalAmt[c] == null) {
						totalAmt[c] = new BigDecimal(0);
					}
					if ( rw.isNull(c)==false && rw.isEmpty(c)==false ) {
						totalAmt[c] = totalAmt[c].add(rw.getDecimal(c));
					}
				}
			} 
			else {
				runPos="processing recordNotEqual(a)";
				// if they have write out the record, unless it's the 1st one
				if (firstOutputRecord) {
					firstOutputRecord = false;
				} 
				else {
					writeGroupRecord(rec, numOfGroups, numOfTotals, previousRow, totalAmt, groupColNum, totalColNum, colCount);
				}
				
				runPos="processing recordNotEqual(b)";
				//for (int col = 0; col < numOfGroups; col++) {
				//	c=groupColNum[col];
				//	previousRow[c] = currentRow[c];
				//}
				for (int col = 0; col < colCount; col++) {
					previousRow[col] = currentRow[col];
				}
				runPos="processing recordNotEqual(c)";
				String ft;
				for (int col = 0; col < numOfTotals; col++) {
					c=totalColNum[col];
					ft=rsmdIn.getField(c).getFieldType();
					runPos="processing recordNotEqual(d)["+col+"]["+rsmdIn.getField(c).getFieldName()+"]["+ft+"]";
					if ( ft.equals("decimal") || ft.equals("integer") ) {
						if ( rw.isNull(c) || rw.isEmpty(c)) {
							totalAmt[c]=new BigDecimal(0);
						}
						else {
							runPos="processing recordNotEqual(e)["+col+"]["+rsmdIn.getField(c).getFieldName()+"]="+rw.getDecimal(c);
							totalAmt[c] = rw.getDecimal(c);
						}
					}
				}
			}
		}
		// Now write out the last record
		runPos="Writing out last record";
		writeGroupRecord(9999, numOfGroups, numOfTotals, previousRow, totalAmt, groupColNum, totalColNum, colCount);
	}
	
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	private void writeGroupRecord(int rec, int numOfGroups, int numOfTotals,
			String[] previousRow, BigDecimal[] totalAmt, Integer[] groupColNum, Integer[] totalColNum, int colCount)
			throws VermilionNonFatalException {
		
		int outputRowNum = outputRecordSet.addRow();
		//used in processGroupColumns
		
		// Write out the group fields
		int outputColNum = 0;
		int c;
		//for (int col = 0; col < numOfGroups; col++) {
		//	c=groupColNum[col];
		//	runPos="writeGroupRecord3048."+rec+"/"+col+" row "+outputRowNum+" ["+c+"] == "+previousRow[c];
		//	outputRecordSet.setValue(outputRowNum, c, previousRow[c]);
		//}
		for (int col = 0; col < colCount; col++) {
			runPos="writeGroupRecord3048."+rec+"/"+col+" row "+outputRowNum+" ["+col+"] == "+previousRow[col];
			outputRecordSet.setValue(outputRowNum, col, previousRow[col]);
		}

		// now write out the summed values
		for (int col = 0; col < numOfTotals; col++) {
			c=totalColNum[col];
			if (totalAmt[c] != null) {
				runPos="writeGroupRecord3056."+rec+"/"+col+" row "+outputRowNum+" ["+c+"]= "+c+" == "+totalAmt[c].toString();
				outputRecordSet.setValue(outputRowNum, c, totalAmt[c].toString());
			}
		}
		runPos="Done writeGroupRecord";
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processThresholdCheck() throws VermilionNonFatalException,
			VermilionFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processThresholdCheck starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processThresholdCheck starting";
		
		String instr = getAndRemoveInstruction("thresholdCheck");

		String[] instrParts = instr.split(",");

		if (instrParts.length < 2) {
			outputRecordSet
					.addWarningMessage("Missing parts to [thresholdCheck] - only see "
							+ instrParts.length + " parts (" + instrParts + ")");
			return;
		}

		String col1Name = instrParts[0];
		String entityFieldname = instrParts[1];
		String logLevel = "all";
		String comparisonOperator = "<";
		String customMessage = "";

		String entityValue = StringHelper.getResolveSubVars(entityFieldname,
				parameterList, true);

		BigDecimal thresholdValue = new BigDecimal(entityValue);

		if (instrParts.length >= 3) {
			logLevel = instrParts[2];
		}

		if (instrParts.length >= 4) {
			if (instrParts[3].equals(">") || instrParts[3].equals("<")
					|| instrParts[3].equals(">=") || instrParts[3].equals("<=")) {
				comparisonOperator = instrParts[3];
			}
		}
		if (instrParts.length >= 5) {
			customMessage = instrParts[4];
			for (int csvIndex = 5; csvIndex < instrParts.length; csvIndex++) {
				customMessage = customMessage + instrParts[csvIndex];
			}
		}
		RecordSetMetaData rsmdIn = outputRecordSet.getMetaData();
		int col1 = getColumnIndexBySystemNm(rsmdIn, col1Name);

		if (col1 < 0) {
			return;

		}

		for (int row = 0; row < outputRecordSet.getRowCount(); row++) {
			RecordRow record = outputRecordSet.getRow(row);
			if (record.getString(col1) == null
					|| record.getString(col1).equals("")) {
				continue;
			}
			BigDecimal col1Value = record.getDecimal(col1);

			String message = "";

			if (comparisonOperator.equals("<")
					&& col1Value.compareTo(thresholdValue) < 0) {
				if (customMessage.equals(""))
					message = col1Value + " is below threshold of "
							+ thresholdValue;
				else
					message = StringHelper.getStringReplacement(StringHelper
							.getStringReplacement(customMessage,
									"$ThresholdValue$",
									String.valueOf(thresholdValue)),
							"$Col1Value$", String.valueOf(col1Value));
			}
			if (comparisonOperator.equals(">")
					&& col1Value.compareTo(thresholdValue) > 0) {
				if (customMessage.equals(""))
					message = col1Value + " is above threshold of "
							+ thresholdValue;
				else
					message = StringHelper.getStringReplacement(StringHelper
							.getStringReplacement(customMessage,
									"$ThresholdValue$",
									String.valueOf(thresholdValue)),
							"$Col1Value$", String.valueOf(col1Value));
			}
			if (comparisonOperator.equals("<=")
					&& col1Value.compareTo(thresholdValue) <= 0) {
				if (customMessage.equals(""))
					message = col1Value + " is below or equal to threshold of "
							+ thresholdValue;
				else
					message = StringHelper.getStringReplacement(StringHelper
							.getStringReplacement(customMessage,
									"$ThresholdValue$",
									String.valueOf(thresholdValue)),
							"$Col1Value$", String.valueOf(col1Value));
			}
			if (comparisonOperator.equals(">=")
					&& col1Value.compareTo(thresholdValue) >= 0) {
				if (customMessage.equals(""))
					message = col1Value + " is above or equal to threshold of "
							+ thresholdValue;
				else
					message = StringHelper.getStringReplacement(StringHelper
							.getStringReplacement(customMessage,
									"$ThresholdValue$",
									String.valueOf(thresholdValue)),
							"$Col1Value$", String.valueOf(col1Value));
			}
			if (!message.equals("")) {
				WarningMessage warningMessage;
				if (logLevel.equalsIgnoreCase("LogOnly")) {
					warningMessage = new WarningMessage(0, "Validation",
							WarningMessage.OUTPUT_TO_LOG_ONLY, message);
				} else if (logLevel.equalsIgnoreCase("ReportOnly")) {
					warningMessage = new WarningMessage(0, "Validation",
							WarningMessage.OUTPUT_TO_REPORT_ONLY, message);
				} else {
					warningMessage = new WarningMessage(0, "Validation",
							WarningMessage.OUTPUT_TO_ALL, message);
				}

				outputRecordSet.addWarningMessage(warningMessage);
			}
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processCompareTwoValues() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "compareTwoValues starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "compareTwoValues starting";

		String instr = getAndRemoveInstruction("compareTwoValues");
		String customMessage = getAndRemoveInstruction("compareTwoValuesMessage");

		instr += " ";
		String[] parts = instr.split("\\,");
		if (parts.length < 5) {
			outputRecordSet
					.addWarningMessage("Missing parts to [compareTwoValues]. Expected 'compareTwoValues=col1,val1,val2,col2,threshold' - only see "
							+ parts.length + " parts (" + instr + ")");
			return;
		}

		String xCol = parts[0].trim();
		String val1 = parts[1].trim();
		String val2 = parts[2].trim();
		String yCol = parts[3].trim();
		String thresholdString = parts[4].trim();
		thresholdString = StringHelper.getResolveSubVars(thresholdString,
				parameterList, true);
		BigDecimal thresholdValue = new BigDecimal(thresholdString);

		RecordSetMetaData rsmd = outputRecordSet.getMetaData();

		int xColIndex = getColumnIndexBySystemNm(rsmd, xCol);
		int yColIndex = getColumnIndexBySystemNm(rsmd, yCol);

		if (xColIndex < 0 || yColIndex < 0)
			return;

		BigDecimal valToCompare1 = new BigDecimal("0");
		BigDecimal valToCompare2 = new BigDecimal("0");

		for (int row = 0; row < outputRecordSet.getRowCount(); row++) {
			String xValue;
			String yValue;

			xValue = outputRecordSet.getRow(row).getString(xColIndex);
			yValue = outputRecordSet.getRow(row).getString(yColIndex);

			if (xValue.equals(val1)) {
				if (yValue != null && !yValue.equals("")) {
					try {
						valToCompare1 = new BigDecimal(yValue);
					} catch (NumberFormatException E) {
						outputRecordSet.addWarningMessage("'" + yValue
								+ "' is not a valid number to compare.");
					}
				}
			}
			if (xValue.equals(val2)) {
				if (yValue != null && !yValue.equals("")) {
					try {
						valToCompare2 = new BigDecimal(yValue);
					} catch (NumberFormatException E) {
						outputRecordSet.addWarningMessage("'" + yValue
								+ "' is not a valid number to compare.");
					}
				}
			}
		}
		BigDecimal diff = valToCompare1.subtract(valToCompare2).abs();

		if (diff.compareTo(thresholdValue) > 0) {
			if (customMessage == null || customMessage.equals(""))
				customMessage = "Difference between $CompareValue1$ ($RowTag1$) and  "
						+ "$CompareValue2$ ($RowTag2$) is $Difference$, which exceeds defined "
						+ "threshold of $ThresholdValue$.";
			customMessage = StringHelper.getStringReplacement(customMessage,
					"$CompareValue1$", String.valueOf(valToCompare1.setScale(2,
							BigDecimal.ROUND_DOWN)));
			customMessage = StringHelper.getStringReplacement(customMessage,
					"$CompareValue2$", String.valueOf(valToCompare2.setScale(2,
							BigDecimal.ROUND_DOWN)));
			customMessage = StringHelper.getStringReplacement(customMessage,
					"$RowTag1$", val1);
			customMessage = StringHelper.getStringReplacement(customMessage,
					"$RowTag2$", val2);
			customMessage = StringHelper.getStringReplacement(customMessage,
					"$Difference$",
					String.valueOf(diff.setScale(2, BigDecimal.ROUND_DOWN)));
			customMessage = StringHelper.getStringReplacement(customMessage,
					"$ThresholdValue$", String.valueOf(thresholdValue.setScale(
							2, BigDecimal.ROUND_DOWN)));

			// originally coded with message:
			// "Current Month return has moved by " + diff.setScale(2,
			// BigDecimal.ROUND_DOWN) +
			// "0% since previous (warning threshold for the fund is movements above "
			// + threshold + "%)");
			WarningMessage warningMessage = new WarningMessage(0, "Validation",
					WarningMessage.OUTPUT_TO_LOG_ONLY, customMessage);

			outputRecordSet.addWarningMessage(warningMessage);
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processSmallPercentage() throws VermilionNonFatalException,
			VermilionFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processSmallPercentage starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processSmallPercentage starting";
		
		String instr = getAndRemoveInstruction("smallPercentage");
		String smallPercentageException = getAndRemoveInstruction("smallPercentageException");

		String[] instrParts = instr.split(",");

		if (instrParts.length < 3) {
			outputRecordSet
					.addWarningMessage("Missing parts to [smallPercentage] - only see "
							+ instrParts.length + " parts (" + instrParts + ")");
			return;
		}

		String col1Name = instrParts[0];
		String col2Name = instrParts[1];
		String entityFieldname = instrParts[2];

		String entityValue = StringHelper.getResolveSubVars(entityFieldname,
				parameterList, true);

		BigDecimal thresholdValue = new BigDecimal(entityValue);

		RecordSetMetaData rsmdIn = outputRecordSet.getMetaData();
		int col1 = getColumnIndexBySystemNm(rsmdIn, col1Name);
		int col2 = getColumnIndexBySystemNm(rsmdIn, col2Name);

		if (col1 < 0 || col2 < 0) {
			return;
		}

		// Move the records to an array
		int rowCount = outputRecordSet.getRowCount();

		RecordRow[] row = new RecordRow[rowCount];
		RecordRow otherRow = null;
		BigDecimal otherAmount = new BigDecimal(0.0);

		for (int i = 0; i < rowCount; i++) {
			row[i] = outputRecordSet.getRow(i);
		}

		this.deleteOutputRecordset();

		for (int i = 0; i < rowCount; i++) {
			if (row[i].getString(col2) == null
					|| row[i].getString(col2).equals("")) {
				continue;
			}
			String col11Value = row[i].getString(col1);
			BigDecimal col12Value = row[i].getDecimal(col2);

			if (col12Value.compareTo(thresholdValue) == -1) {
				if (otherRow == null) {
					otherRow = row[i];
				}
				if (smallPercentageException.equals(col11Value)) {
					outputRecordSet.addRow(row[i]);
				} else {
					otherAmount = otherAmount.add(col12Value);
				}
			} else {
				outputRecordSet.addRow(row[i]);
			}
		}

		// add in the 'other' row
		if (otherRow != null) {
			int newRow = outputRecordSet.addRow(otherRow);
			outputRecordSet.setValue(newRow, col2, otherAmount.toString());
			outputRecordSet.setValue(newRow, col1, "Other");
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processHideRow() throws VermilionNonFatalException,
			VermilionFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processHideRow starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processHideRow starting";

		String instr = getAndRemoveInstruction("hideRow");

		String[] instrParts = instr.split(",");

		if (instrParts.length < 3) {
			outputRecordSet
					.addWarningMessage("Missing parts to [hideRow] - only see "
							+ instrParts.length + " parts (" + instrParts + ")");
			return;
		}

		String[] colParts = instrParts[0].split("=");

		if (colParts.length < 2) {
			outputRecordSet
					.addWarningMessage("Missing parts to col1 details [hideRow] - only see "
							+ colParts.length + " parts (" + colParts + ")");
			return;
		}

		String col1Name = colParts[0];
		String col1Value = colParts[1];
		String col2Name = instrParts[1];
		String spreadType = instrParts[2];

		if (!(spreadType.equals("even") || spreadType.equals("percent"))) {
			outputRecordSet
					.addWarningMessage("spreadType must be either 'even' or 'percent', but was "
							+ spreadType);
			return;
		}

		RecordSetMetaData rsmdIn = outputRecordSet.getMetaData();
		int col1 = getColumnIndexBySystemNm(rsmdIn, col1Name);
		int col2 = getColumnIndexBySystemNm(rsmdIn, col2Name);

		if (col1 < 0 || col2 < 0) {
			return;
		}

		// Move the records to an array
		int rowCount = outputRecordSet.getRowCount();

		RecordRow[] row = new RecordRow[rowCount];

		for (int i = 0; i < rowCount; i++) {
			row[i] = outputRecordSet.getRow(i);
		}

		this.deleteOutputRecordset();

		BigDecimal amountRemoved = new BigDecimal(0.0);
		BigDecimal totalAmount = new BigDecimal(0.0);
		BigDecimal newRowCount = new BigDecimal(0.0);
		BigDecimal one = new BigDecimal(1.0);

		// take out any records matching the value
		for (int i = 0; i < rowCount; i++) {
			// if it matches the value passed then store the amounts
			if (row[i].getString(col1) == null
					|| row[i].getString(col1).equals(col1Value)) {
				amountRemoved = amountRemoved.add(row[i].getDecimal(col2));
			} else {
				totalAmount = totalAmount.add(row[i].getDecimal(col2));
				newRowCount = newRowCount.add(one);
			}
		}

		BigDecimal percentIncrease = amountRemoved.divide(totalAmount, 10,
				RoundingMode.HALF_UP);

		// now spread the amount removed across the other records
		for (int i = 0; i < rowCount; i++) {
			// if it matches the value passed then store the amounts
			if (row[i].getString(col1) == null
					|| row[i].getString(col1).equals(col1Value)) {
				continue;
			} else {
				BigDecimal newAmount = new BigDecimal(0.0);
				BigDecimal oldAmount = row[i].getDecimal(col2);

				if (spreadType.equals("percent")) {
					newAmount = oldAmount.add(oldAmount
							.multiply(percentIncrease));
				} else {
					newAmount = oldAmount.add(amountRemoved.divide(newRowCount,
							10, RoundingMode.HALF_UP));
				}
				int rowNum = outputRecordSet.addRow(row[i]);
				outputRecordSet.setValue(rowNum, col2, newAmount.toString());
			}
		}

	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processSubstituteValues() throws NumberFormatException,
			VermilionNonFatalException, VermilionFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processSubstituteValues starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processSubstituteValues starting";
		
		
		// If resolveSubVarColNm has been set in the DST notes then this will
		// trigger the DragThrough to look at substituting anything it finds in
		// the field for each record...
		String resolveSubVarOutputLang = getAndRemoveInstruction("resolveSubVarOutputLang");
	    String resolveSubVarEscapeQuotes = getAndRemoveInstruction("resolveSubVarEscapeQuotes");
	    boolean escapeQuotesSelection = false;
	    if (resolveSubVarEscapeQuotes.equalsIgnoreCase("Y")) {
	    	escapeQuotesSelection = true;
	    }
		RecordSetMetaData rsmd = outputRecordSet.getMetaData();
		int subVarCol = getColumnIndexBySystemNm(rsmd,
				getAndRemoveInstruction("resolveSubVarColNm"));
		if (subVarCol < 0)
			return;

		ParameterRecord global = null;
		if (isInstructionValuePassed("globalVariableCd")
				&& isInstructionValuePassed("globalVariableKey")) {
			global = new ParameterRecord(outputRecordSet.getTid(),
					getAndRemoveInstruction("globalVariableCd"),
					getAndRemoveInstruction("globalVariableKey"));
			outputRecordSet.addWarningMessage(global.toString());
		}

		for (int rec = 0; rec < outputRecordSet.getRowCount(); rec++) {
			try {

				String unsubstituted = "";
		        String substituted = this.outputRecordSet.getString(rec, subVarCol);
		        while (!unsubstituted.equals(substituted))
		        {
		          unsubstituted = substituted;
		          substituted = getResolveGlobalVariables(global, 
		            StringHelper.getResolveSubVars(unsubstituted, 
		            this.parameterList, true), true);
		        }
		        String unparsed = "";
		        //replaces any single or double quotes that are not next to a comma or directly inside parentheses w/ escaped versions
		        String parsed = substituted;
		        if(escapeQuotesSelection){
		        	parsed = parsed.replaceAll("(?<![\\(,])\"(?![\\),])","&#34;").replaceAll("(?<![\\(,])\'(?![\\),])","&#39;");
		        	//logManager.warn("[DTP] escaped quotes, about to parse: "+parsed);
		        }
		        while (!unparsed.equals(parsed))
		        {
		          //logManager.error("[DTP] About to parse formulas in: "+parsed);
		          unparsed = parsed;
		          parsed = StringHelper.parseFunctions(parsed);
		          parsed = ComponentUtility.parseText(
		            this.outputRecordSet.getTid(), parsed, 
		            resolveSubVarOutputLang);
		        }
		          //logManager.error("[DTP] Done parsing: "+parsed);

		        if(escapeQuotesSelection){
		        	//logManager.warn("[DTP]parsed, about to unescape quotes: "+parsed);
		        	parsed= parsed.replaceAll("&#34;", "\"").replaceAll("&#39;","\'");
		        }
		        String translated = getTranslatedText(this.outputRecordSet.getTid(), 
		          parsed, resolveSubVarOutputLang);
		        outputRecordSet.setValue(rec, subVarCol, translated);

				if (outputRecordSet.getString(rec, subVarCol).indexOf(
						"$_") > -1) {
					String s = getSVWarning(
							outputRecordSet.getString(rec, subVarCol),
							parameterList).trim();
					if (!s.equals("")) {
						outputRecordSet.addWarningMessage(s
								+ " (err1002)");
					}
				}
			} catch (Exception e) {
				outputRecordSet
						.addWarningMessage("Could not resolve substitution variable from within record "
								+ rec
								+ " please check syntax (entities available are "
								+ parameterList.getParameterNames());
			}
		}
	}

	private String getResolveGlobalVariables(ParameterRecord globalEntity,
			String text, boolean ignoreEntityNulls) {
		//used in processSubstituteValues
		String result = text;
		if (globalEntity != null) {
			if (result.indexOf("$_") > -1) {
				String entityNm = "";
				try {
					entityNm = (String) globalEntity.getParameterName();
					if (result.indexOf("$_") > -1) {
						result = StringHelper.getParameterReplacement(result,
								globalEntity, ignoreEntityNulls);
					}
				} catch (Exception e) {
					if (e.toString().indexOf("Unable to create") > -1
							&& e.toString().indexOf(
									"No data found for specified parameters") > -1) {
						logManager
								.error("::Problems processing Global Parameter record "
										+ entityNm
										+ "); is this parameter chained from an outer datasource?");
					} else {
						logManager
								.error("::Problems processing Global Parameter record "
										+ entityNm
										+ ".. in the getResolveGlobalVariables method of DragThroughProcessor custom class. ("
										+ e.toString() + ")");
					}
				}
			}
		}
		return result;
	}

	private String getTranslatedText(int tid, String sourceText,
			String outputLang) {
		//used in processSubstituteValues
		
		String translated = null;
		String sourceLanguage = "en-gb";

		// Translate the entity field value.
		TranslationData transData = new TranslationData();
		transData.setOutputLanguage(outputLang);
		transData.setInputLanguage(sourceLanguage);
		transData.setInputText(sourceText);
		transData.setModifiedFg(0);

		try {
			translated = new TranslateXML(tid).translateText(transData);
		} catch (Exception ex) {
			outputRecordSet.addWarningMessage("Unable to translate text ["
					+ sourceText + "] from " + sourceLanguage + " to "
					+ outputLang + ".");
			translated = sourceText;
		}

		return translated;
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processComparePreviousTerm()
			throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processComparePreviousTerm starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processComparePreviousTerm starting";

		String comparePreviousTerm = getAndRemoveInstruction("comparePreviousTerm");

		RecordSet currDS = inputRecordSetList.getRecordSet(0); // this can only
																// be run
																// against 1
																// record set

		RecordSetMetaData rsmdIn = currDS.getMetaData();
		int keyFieldColIndex = getColumnIndexBySystemNm(rsmdIn,
				comparePreviousTerm);

		if (keyFieldColIndex < 0)
			outputRecordSet.addWarningMessage("Couldn't find column "
					+ comparePreviousTerm
					+ ". This should contain the column name");

		String term = null;
		String entityDate = null;
		String fund = null;

		for (int i = 0; i < parameterList.getParameterCount(); i++) {
			ParameterRecord rec = null;
			try {
				rec = parameterList.getParameterRecord(i);
				if (rec.getParameterName().toUpperCase().equals("TERM")) {
					term = rec.getFieldValue(0);
				}
				if (rec.getParameterName().toUpperCase().equals("FUND")) {
					fund = rec.getKeyValue();
				}
				if (rec.getParameterName().toUpperCase().equals("DATE")) {
					entityDate = rec.getFieldValue(0);
				}
			} catch (VermilionFatalException e) {
				e.printStackTrace();
			}
		}

		if (entityDate == null) {
			outputRecordSet
					.addWarningMessage("Couldn't pickup the date entity value - the DS must be run with this as a parameter");
			return;
		}

		if (term == null) {
			outputRecordSet.addWarningMessage(new WarningMessage(0,
					"Rendering", WarningMessage.DONT_OUTPUT,
					"The term entity value was missing, so defaulting to QTD"));
			term="QTD";
		}

		String previousDate = getPreviousDate(term, entityDate);
		SimpleDateFormat inputDateFormat = new SimpleDateFormat("dd MMMM yyyy");
		SimpleDateFormat outputDateFormat = new SimpleDateFormat("dd-MMM-yyyy");

		Date endCalDate;
		try {
			endCalDate = inputDateFormat.parse(previousDate);
		} catch (ParseException e1) {
			outputRecordSet
					.addWarningMessage("Exception caught when parsing the date "
							+ previousDate);
			return;
		}
		String prevDateFormatted = outputDateFormat.format(endCalDate);

		RecordSet prevDS;
		try {
			prevDS = getPreviousDatasource(currDS.getDataSourceCode(), term,
					previousDate, fund);
		} catch (VermilionFatalException e) {
			e.printStackTrace();
			outputRecordSet
					.addWarningMessage("Couldn't pickup the DS for the previous term. "
							+ e.getMessage());
			return;
		}

		if (prevDS == null || prevDS.getRowCount() == 0) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processComparePreviousTerm no previous datasource found for term "
							+ term + " fund " + fund + " and date "
							+ previousDate);
			outputRecordSet.addWarningMessage(msg);
			return;
		}

		// load the keys into a set
		HashSet<String> prev = loadSet(prevDS, keyFieldColIndex);
		HashSet<String> curr = loadSet(currDS, keyFieldColIndex);

		prev.removeAll(curr); // f1 now contains only the lines which are not in
								// f2

		for (String key : prev) {
			outputRecordSet.addWarningMessage("In previous term ("
					+ prevDateFormatted + "), but not current term ("
					+ entityDate + ") : " + key);
		}

		// reload prev
		prev = loadSet(prevDS, keyFieldColIndex);

		curr.removeAll(prev); // f2 now contains only the lines which are not in
								// f1

		for (String key : curr) {
			outputRecordSet.addWarningMessage("In current term (" + entityDate
					+ "), but not previous term (" + prevDateFormatted + ") : "
					+ key);
		}

	}

	private HashSet<String> loadSet(RecordSet ds, int column)
			throws VermilionNonFatalException {
		HashSet<String> retValues = new HashSet<String>();
		//used in processComparePreviousTerm
		for (int i = 0; i < ds.getRowCount(); i++) {
			retValues.add(ds.getRow(i).getString(column).trim());
		}

		return retValues;
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processResolveDataSubVar() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processResolveDataSubVar starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processResolveDataSubVar starting";
		
		String runThisInstruction = getAndRemoveInstruction("resolveDataVariables");
		if (! runThisInstruction.equalsIgnoreCase("Y")) {
			return;
		}
	    String useDefaultRowInstruction = getAndRemoveInstruction("dataVarUseDefault");
	    boolean useDefaultRowSelection = false;
	    if (useDefaultRowInstruction.equalsIgnoreCase("Y")) {
	      useDefaultRowSelection = true;
	    }
	    String resolveInsideInstruction = getAndRemoveInstruction("dataVarResolveInside");
	    boolean resolveInsideSelection = false;
	    if (resolveInsideInstruction.equalsIgnoreCase("Y")) {
	      resolveInsideSelection = true;
	    }
	    String makeDynamicCallsInstruction = getAndRemoveInstruction("dataVarDynamicCalls");
	    boolean makeDynamicCallsSelection = false;
	    if (makeDynamicCallsInstruction.equalsIgnoreCase("Y")) {
	    	makeDynamicCallsSelection = true;
	    }	    
	    String valueCaseSensitivityInstruction = getAndRemoveInstruction("dataVarValueCaseSensitive");
	    boolean valueCaseSensitivitySelection = false;
	    if (valueCaseSensitivityInstruction.equalsIgnoreCase("Y")) {
	    	valueCaseSensitivitySelection = true;
	    }	    
	    
	    
		// If the user failed to specify a drag through RS/DS with the
		// instructions field, pick the first
		// manual datasource among input datasource. If no input record sets are
		// manual, just drag through
		// the first one.
		if (this.outputRecordSet == null) {
			// In case we do not find a manual data source among our availing
			// input RS/DS, drag through
			// the first input item as a failsafe.
			try {
				this.outputRecordSet = inputRecordSetList.getRecordSet(0);
			} catch (VermilionNonFatalException e) {
				e.printStackTrace();
			}

			// Look for the first manual DS among the input datasources; input DS "order" is arbitrary at runtime though
			for (int recordSetNum = 0; recordSetNum < inputRecordSetList
					.getRecordSetCount(); recordSetNum++) {
				try {
					String DSinfo = inputRecordSetList
							.getRecordSet(recordSetNum).getDataSourceInfo()
							.toString();
					String DStype = DSinfo.substring(
							DSinfo.indexOf("Data Source Type = ") + 19, 1);
					if (DStype.equals("3")) {
						this.outputRecordSet = inputRecordSetList
								.getRecordSet(recordSetNum);
						break;
					}
				} catch (Exception E) {
					// Record set was not from a data source
				}
			}
		}

		// Ready a hashmap to cache any dynamic datasource calls (if need be)
		HashMap<String,RecordSet> cachedRecordSets = new HashMap<String,RecordSet>();
		// Now that we're sure we have something in our output record set, let's
		// go through it and replace
		// its data substitution variables

		try {
			for (int r = 0; r < this.outputRecordSet.getRowCount(); r++) {
    			//Date date= new Date();
    			//Timestamp timestamp = new Timestamp(date.getTime());
    			//logManager.debug("about to resolve row "+r+" at "+timestamp);
				for (int c = 0; c < this.outputRecordSet.getColumnCount(); c++) {
	    			//logManager.debug("about to resolve col "+c+" at "+timestamp);
					String dataText=this.outputRecordSet.getString(r,c);
					if(dataText==null){
						//the input data was null and contains no variables. and serach on it will error.
						continue;
					}
					boolean needsHTMLWrapper= false;
					int searchFrom = 0;
					while (dataText.indexOf("$", searchFrom) > -1) {
		    			//logManager.debug("about to search from "+searchFrom+" in ("+r+","+c+") at "+timestamp);
						if (resolveInsideSelection){
							String unsubstituted = "";
						    String substituted = dataText;
						    while (!unsubstituted.equals(substituted)){
						    	unsubstituted = substituted;
						    	substituted = StringHelper.getResolveSubVars(unsubstituted, 
						    			this.parameterList, true);
						    }
				    		if (!substituted.equals(dataText)){
				    			dataText = substituted;
				    		}

						}
						// Looking for either $DSName.DSColumn.DSRow$ or
						// $DSName.DSColumn$ or even $DSName.DsColumn.DSKeyColumn=DSValue$

						int startVariable = dataText.indexOf("$", searchFrom);
						int endVariableDS = dataText.indexOf(".",
								startVariable + 1);
						int endVariableColumn = dataText.indexOf(".",
								endVariableDS + 1);
						int endVariable = dataText.indexOf("$",
								endVariableDS + 1);
						searchFrom = startVariable + 1;	
						// used to be searchFrom = max(endVar, startVa+1)
						// why was that a problem? I can't remember -Ata
						if (endVariableColumn > -1) {
							endVariableColumn = Math.min(endVariableColumn,
									endVariable);
						}
						if (endVariableColumn < 0) {
							endVariableColumn = endVariable;
						}
						if (endVariableDS > 0 && endVariableColumn > 0
								&& endVariable > 0) {
							// We have a complete substitution variable
							// (probably)!
							String variableDSName = dataText.substring(
									startVariable + 1, endVariableDS);
							if(variableDSName == null || variableDSName.length()==0 || variableDSName.substring(0,1).equals("_")){
								continue;
							}
							String variableColumnName = dataText.substring(
									endVariableDS + 1, endVariableColumn);
							String variableRowArgument = "";
							String variableText = "$$";
							if (endVariable == endVariableColumn) {
								variableText = "$" + variableDSName + "."
										+ variableColumnName + "$";
							} else {
								variableRowArgument = dataText.substring(
										endVariableColumn + 1, endVariable);
								variableText = "$" + variableDSName + "."
										+ variableColumnName + "."
										+ variableRowArgument + "$";
							}
							//logManager.debug("Trying to resolve potential data variable "+variableText);
							// Find the input RS needed for the data reference
							RecordSet referencedRS = null;
							if(variableDSName != null & variableDSName.equalsIgnoreCase("Output")){
				    			try{
				    				referencedRS = outputRecordSet;
				    			}
				    			catch(Exception E){
				    				// No data source input exists with the desired name.
				    			}
			    			}
							if (referencedRS == null) {
								try {
									referencedRS = inputRecordSetList
											.getRecordSetByDataSourceName(variableDSName);
								} catch (Exception E) {
									// No data source input exists with the
									// desired name.
								}
							}
							if (referencedRS == null) {
								try {
									referencedRS = inputRecordSetList
											.getRecordSetByDataSourceCode(variableDSName);
								} catch (Exception E) {
									// No data source input exists with the
									// desired name.
								}
							}
							if (referencedRS == null) {
								try {
									referencedRS = inputRecordSetList
											.getRecordSetByRecordSetName(variableDSName);
								} catch (Exception E) {
									// No record set input exists with the
									// desired name.
								}
							}
							if (referencedRS == null) {
								try {
									referencedRS = inputRecordSetList
											.getRecordSetByRecordSetCode(variableDSName);
								} catch (Exception E) {
									// No record set input exists with the
									// desired code.
								}
							}
//		I'm preeettty sure I intended for this to be removed and replaced with the below block, which allows retrievd datasources to be reused.
//							
//							if (referencedRS == null && makeDynamicCallsSelection) {
//								//try a dynamic DGS call then
//								try {
//									//below lines were borrowed from DynamicDGSCall. We make sure the code exists before making
//									// a dynamic call, or else dynamicDGSCall will fill logs w/ error messages
//									 ViewDSTListData dstLD = null;
//									 dstLD = DgsBD.viewDST(parameterList.getTid(), String.valueOf(parameterList.getUserId()));
//								     ViewDSTData[] aVDLD = (ViewDSTData[])dstLD.getAllRows().toArray(new ViewDSTData[0]);
//								     for (int index = 0; index < aVDLD.length; index++) {
//								    	 if (aVDLD[index].getCompCode().equals(variableDSName)) {
//								        	DynamicDGSCall dgsHelper = new DynamicDGSCall(inputRecordSetList.getTid());  
//											referencedRS = dgsHelper.getDGSDataByDstCode(variableDSName, parameterList);
//											break;
//								    	 }
//								     }
//								} catch (Exception E) {
//									// No datasource in VRS exists with the
//									// desired code.
//								}
//							}
							if (referencedRS == null && makeDynamicCallsSelection) {
								//have we already cached this one?
								if(cachedRecordSets.containsKey(variableDSName)){
									referencedRS = cachedRecordSets.get(variableDSName);
								} else{
									//try a dynamic DGS call then
									try {
									     //Check to see if variableDSName contains ">ENTITY_CODE:EntityValue" in it; if so, we will need
									     //to add this to the parameter list (and exclude it from the
										String variableDSNameSplitByParams[] = variableDSName.split(">");
										ParameterList newParameterList;
										if(variableDSNameSplitByParams.length == 1){
											newParameterList = parameterList;
										}else{
											//clone the existing parameterList so we can add parameters just for this 
											newParameterList = new ParameterList(parameterList.getTid(), parameterList.getUserId());
											for(int parameterNumber=0;parameterNumber<parameterList.getParameterCount();parameterNumber++){
												newParameterList.addParameterRecord(parameterList.getParameterRecord(parameterNumber));
											}
											//iterate through the additional parameters and try to add them
											for(int substringNumber=1;substringNumber<variableDSNameSplitByParams.length;substringNumber++){
												String entityCodeAndKey[] = variableDSNameSplitByParams[substringNumber].split("[:=]",2);
												ParameterRecord newParameterRecord;
												try{
													newParameterRecord = new ParameterRecord(parameterList.getTid(),entityCodeAndKey[0],entityCodeAndKey[1]);
												}catch(Exception E){
													throw new Exception(
															variableText+ " has an invalid (nonmatching) entity parameter: "+ variableDSNameSplitByParams[substringNumber]);
												}
												try{
													newParameterList.addParameterRecord(newParameterRecord);
												}catch(Exception E){
													throw new Exception(
														variableText+ " has an invalid (likely redundant) entity parameter: "+ variableDSNameSplitByParams[substringNumber]);
												}
											}
										}
										
										//below lines were borrowed from DynamicDGSCall. We make sure the cocde exists before making
										// a dynamic call, or else dynamicDGSCall will fill logs w/ error messages
										 ViewDSTListData dstLD = null;
										 dstLD = DgsBD.viewDST(parameterList.getTid(), String.valueOf(parameterList.getUserId()));
									     ViewDSTData[] aVDLD = (ViewDSTData[])dstLD.getAllRows().toArray(new ViewDSTData[0]);

									     for (int index = 0; index < aVDLD.length; index++) {
									    	 if (aVDLD[index].getCompName().equals(variableDSNameSplitByParams[0])) {
													DynamicDGSCall dgsHelper = new DynamicDGSCall(inputRecordSetList.getTid());  
													referencedRS = dgsHelper.getDGSDataByDstName(aVDLD[index].getCompName(), newParameterList);
													cachedRecordSets.put(variableDSName, referencedRS);
												break;
									    	 }else if (aVDLD[index].getCompCode().equals(variableDSNameSplitByParams[0])) {
									    			DynamicDGSCall dgsHelper = new DynamicDGSCall(inputRecordSetList.getTid());  
													referencedRS = dgsHelper.getDGSDataByDstName(aVDLD[index].getCompName(), newParameterList);
													cachedRecordSets.put(variableDSName, referencedRS);
												break;
									    	 }
									     }
	
									} catch (Exception e) {
										outputRecordSet.addWarningMessage(new WarningMessage(
					                            5, 
					                            "Invalid Substitution Variable", 
					                            1, 
					                            e.getMessage()));
									}
								}
							}
							if (referencedRS == null) {
								// Couldn't find the desired referenced record
								// set. Jump to the next
								// present variable.
								continue;
							}
							runPos="found the recordset for "+variableText;
							// Find the appropriate column of the record set
							int refColumn = -1;
							for (int colNum = 0; colNum < referencedRS
									.getColumnCount(); colNum++) {
								String refColumnSysName = referencedRS
										.getMetaData(colNum).getFieldName();
								String refColumnPrefName = referencedRS
										.getMetaData(colNum).getPreferredName();
								if (refColumnSysName.equals(variableColumnName)) {
									refColumn = colNum;
									break;
								} else if (refColumnPrefName != null
										&& refColumnPrefName
												.equals(variableColumnName)) {
									refColumn = colNum;
									break;
								}
							}
							if (refColumn < 0) {
								// Couldn't find the desired column name. Jump
								// to the next
								// present variable.
								continue;
							}
							runPos="found the column for "+variableText;
							if (referencedRS.getRowCount() > 0) {
								// Find the appropriate row of data desired
								int refRow = -1;
								if (variableRowArgument.equals("")) {
									// user did not give third argument to
									// substitution variable
									refRow = r % referencedRS.getRowCount();
									runPos="chose row "+refRow +" automatically for "+variableText;
								} else {
									if (resolveInsideSelection){
										runPos="resolving row argument of "+variableText;
										String resolveSubVarOutputLang = getAndRemoveInstruction("resolveSubVarOutputLang");
										String unparsed = "";
										String parsed = variableRowArgument;
										while (!unparsed.equals(parsed)){
											unparsed = parsed;
											parsed = StringHelper.parseFunctions(parsed);
											parsed = ComponentUtility.parseText(outputRecordSet.getTid(), parsed, resolveSubVarOutputLang);
										}
										variableRowArgument = parsed;
				                    }
									try {
										runPos="testing if row argument of "+variableText+" is a row number";
										refRow = (Integer
												.parseInt(variableRowArgument))
												% referencedRS.getRowCount();
									} catch (NumberFormatException E) {
										runPos="treating row argument of "+variableText+" as search condition";
										// variableRowNum is a string. Try to
										// parse it for a row-finding process.
										// use the default row selection
										// process as a backup, in case the
										// argument is mal-formed
										if (useDefaultRowSelection) {
					                        refRow = r % referencedRS.getRowCount();
					                      }
										try {
											int equalsLocation = variableRowArgument
													.indexOf("=");
											String fieldToSearch = variableRowArgument
													.substring(0,
															equalsLocation);
											String valueToSearchFor = variableRowArgument
													.substring(equalsLocation + 1);
											int columnToSearch = -1;
											for (int colNum = 0; colNum < referencedRS
													.getColumnCount(); colNum++) {
												String refColumnSysName = referencedRS
														.getMetaData(colNum)
														.getFieldName();
												String refColumnPrefName = referencedRS
														.getMetaData(colNum)
														.getPreferredName();
												if (refColumnSysName
														.equals(fieldToSearch)) {
													columnToSearch = colNum;
													break;
												} else if (refColumnPrefName != null
														&& refColumnPrefName
																.equals(fieldToSearch)) {
													columnToSearch = colNum;
													break;
												}
											}
											if (columnToSearch < 0) {
												// Couldn't find the desired
												// column name.
												throw new Exception(
														variableText
																+ " has an unfound field name in row argument.");
											}
											int rowFound = -1;
											if(valueCaseSensitivitySelection){
												rowFound = referencedRS.find(
													columnToSearch,
													valueToSearchFor);
											}else{
											    //manually search through, b/c RecordSet.find is case sensitive
											    for (int currentRow = 0; currentRow < referencedRS.getRowCount(); currentRow++) {
											      if (referencedRS.getString(currentRow, columnToSearch).equalsIgnoreCase(valueToSearchFor)) {
											    	  rowFound = currentRow;
											    	  break;
											      }
											    }
											}
											if (rowFound > -1) {
												refRow = rowFound;
												if (rowFound < referencedRS.getRowCount()){
													int additionalRowFound = -1;
													if(valueCaseSensitivitySelection){
														rowFound = referencedRS.find(
															columnToSearch,
															valueToSearchFor, rowFound+1);
													}else{
													    //manually search through, b/c RecordSet.find is case sensitive
													    for (int currentRow = rowFound+1; currentRow < referencedRS.getRowCount(); currentRow++) {
													      if (referencedRS.getString(currentRow, columnToSearch).equalsIgnoreCase(valueToSearchFor)) {
													    	  rowFound = currentRow;
													    	  break;
													      }
													    }
													}
													//warn if more rows match!!
													if(additionalRowFound > -1){
														WarningMessage multipleRowsWarning = new WarningMessage(
																5,
																"Ambiguous Substitution Variable",
																WarningMessage.OUTPUT_TO_LOG_ONLY,
																variableText
																		+ " matches multiple rows. Used first row found.");
														outputRecordSet
																.addWarningMessage(multipleRowsWarning);
													}
												}
											} else {
												throw new Exception(
														variableText
																+ " has an unfound value in row argument.");
											}
										} catch (IndexOutOfBoundsException e) {
											WarningMessage badRowArgumentWarning;
											if (useDefaultRowSelection) {
						                          badRowArgumentWarning = new WarningMessage(
						                            5, 
						                            "Invalid Substitution Variable", 
						                            1, 
						                            variableText + 
						                            " has an improper row argument. Used default row selection instead.");
						                        } else {
						                          badRowArgumentWarning = new WarningMessage(
						                            5, 
						                            "Invalid Substitution Variable", 
						                            1, 
						                            variableText + 
						                            " has an improper row argument. Did not resolve data variable.");
						                        }
											outputRecordSet
													.addWarningMessage(badRowArgumentWarning);
										} catch (Exception e) {
											WarningMessage badRowArgumentWarning;
											 if (useDefaultRowSelection) {
						                          badRowArgumentWarning = new WarningMessage(
						                            5, 
						                            "Invalid Substitution Variable", 
						                            1, 
						                            e.getMessage() + 
						                            " Used default row selection instead.");
						                        } else {
						                          badRowArgumentWarning = new WarningMessage(
						                            5, 
						                            "Invalid Substitution Variable", 
						                            1, 
						                            e.getMessage() + 
						                            " Did not resolve data variable.");
						                        }
											outputRecordSet
													.addWarningMessage(badRowArgumentWarning);
										}
									}
								}
								if(refRow < 0 ){
									// the row argument was malformed and default row selection was turned off. We won't substitute this variable.
									continue;
								}
								runPos="finding target data for "+variableText+"";
								String replacementText = referencedRS
										.getString(refRow, refColumn);
								if(replacementText==null){
									//avoid some nasty nullpointer exceptions that can come from truly-null data. For VRS, nulls are
									//effectively blanks.
									replacementText="";
								}else{
									try{
										runPos="checking if target data for "+variableText+" is numeric and formatting it accordingly";
										BigDecimal replacementDecimal = referencedRS.getDecimal(refRow, refColumn);
										
										
										if(replacementDecimal == null){
											throw new NumberFormatException("Number is null");
										}
										DecimalFormat refFormat = null;
										runPos="checking the field type of target data for "+variableText+"";
										if(referencedRS.getMetaData(refColumn).getFieldType().equalsIgnoreCase("decimal")
												&& (outputRecordSet.getMetaData(c).getFieldType().equalsIgnoreCase("text")
												|| outputRecordSet.getMetaData(c).getFieldType().equalsIgnoreCase("html"))){
											try{
												refFormat = new DecimalFormat(referencedRS.getMetaData(refColumn).getFieldDisplayFormat());
											}catch(Exception E){
												//the record set didn't have a format defined
												logManager.warn("[DTP] unable to find formatting for decimal data variable.");
											}
										}
										if(refFormat==null){
											refFormat = new DecimalFormat("#.####");
										}
										replacementText= refFormat.format(replacementDecimal.doubleValue());
									}catch(NumberFormatException e){
										//data value was not a number
									}
								}
								runPos="replacing("+variableText+")in("+dataText+")with("+replacementText+")";
								if(trace){
									WarningMessage msg = new WarningMessage(0, "Rendering",
											WarningMessage.DONT_OUTPUT,
											runPos);
									outputRecordSet.addWarningMessage(msg);
								}
								/*
								 * WarningMessage warningMessage2 = new
								 * WarningMessage(5, "Invalid Settings",
								 * WarningMessage.OUTPUT_TO_LOG_ONLY,
								 * "replacing("
								 * +variableText+")in("+dataText+")with("
								 * +replacementText+")");
								 * outputRS.addWarningMessage(warningMessage2);
								 */
								if(replacementText.contains("<HTML>") || replacementText.contains("</HTML>")){
									needsHTMLWrapper=true;
									if(dataText.contains("<HTML>") || dataText.contains("</HTML>")){
										replacementText = StringHelper.getStringReplacement(replacementText, "<HTML><P>","");
										replacementText = StringHelper.getStringReplacement(replacementText, "</P></HTML>","");										
									}
									replacementText = StringHelper.getStringReplacement(replacementText, "<HTML>","");
									replacementText = StringHelper.getStringReplacement(replacementText, "</HTML>","");
								}
								dataText = StringHelper
										.getStringReplacement(dataText,
												variableText, replacementText);
								searchFrom += replacementText.length();
							} else {
								// referenced RS didn't have any rows of data!
								// We can't substitute!
								WarningMessage badRowArgumentWarning = new WarningMessage(
										5,
										"Invalid Substitution Variable",
										WarningMessage.OUTPUT_TO_LOG_ONLY,
										variableText
												+ " Referenced record set contained no data.");
								outputRecordSet
										.addWarningMessage(badRowArgumentWarning);
								searchFrom += variableText.length();
							}

						}

						this.outputRecordSet.setValue(r, c, dataText);
					}
					if(needsHTMLWrapper){
						dataText = this.outputRecordSet.getString(r, c);
						dataText = StringHelper.getStringReplacement(dataText, "<HTML>","");
						dataText = StringHelper.getStringReplacement(dataText, "</HTML>","");
						dataText = "<HTML>"+dataText+"</HTML>";
						this.outputRecordSet.setValue(r, c, dataText);
					}
				}
			}
		} catch (VermilionNonFatalException e) {
			outputRecordSet.addWarningMessage(e.toString());
			throw e;
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processAddBlankLine() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processAddBlankLine started");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processAddBlankLine starting";
		
		String addBlankLine = getAndRemoveInstruction("addBlankLine");
		RecordSet outRs = null;

		if (addBlankLine.equalsIgnoreCase("top")) {
			outRs = new RecordSet(outputRecordSet.getMetaData());
			outRs.addRow();

			for (int i = 0; i < outputRecordSet.getRowCount(); i++) {
				outRs.addRow(outputRecordSet.getRow(i));
			}

			setRecordSet(outRs);

		} else if (addBlankLine.equalsIgnoreCase("bottom")) {
			outputRecordSet.addRow();
		} else {
			outputRecordSet
					.addWarningMessage("addBlankLine is only valid with 'top' or 'bottom'. Was passed "
							+ addBlankLine);
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processAddTotalRow() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processAddTotalRow started");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processAddTotalRow starting";
		
		String addTotalRow = getAndRemoveInstruction("addTotalRow");
		String[] sumColumns = addTotalRow.split(",");
		if (sumColumns.length < 1) {
			outputRecordSet
					.addWarningMessage("addTotalRow requires a list of columns to sum.");
			return;
		}

		Integer colIndex[] = new Integer[sumColumns.length];
		Double colTotal[] = new Double[sumColumns.length];
		RecordSetMetaData rsmdIn = outputRecordSet.getMetaData();

		for (int i = 0; i < sumColumns.length; i++) {
			colIndex[i] = getColumnIndexBySystemNm(rsmdIn, sumColumns[i]);
			if (colIndex[i] < 0) {
				outputRecordSet
						.addWarningMessage("addTotalRow could not find column "
								+ sumColumns[i]);
				return;
			}
			colTotal[i] = outputRecordSet.getDoubleSum(colIndex[i]);
		}

		int newRow = outputRecordSet.addRow();
		for (int i = 0; i < colTotal.length; i++) {
			outputRecordSet.setValue(newRow, colIndex[i], colTotal[i]);
		}
    	String putTotalAt = getAndRemoveInstruction("putTotalAt");
    	if(putTotalAt.equalsIgnoreCase("-")){
    		//putTotalAt instruction was not used. Default to putting the word total in column 0.
    		outputRecordSet.setValue(newRow, 0, "Total");
    	}else{
    		int totalColumnIndex = getColumnIndexBySystemNm(rsmdIn, putTotalAt);
    		if(totalColumnIndex<0 && ! putTotalAt.equalsIgnoreCase("N/A")){
    			outputRecordSet.addWarningMessage("addTotalRow could not find column " + 
    					putTotalAt + "for 'Total' label");
    		}else if(totalColumnIndex>=0){
    			outputRecordSet.setValue(newRow, totalColumnIndex, "Total");
    		}
    	}		


	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processSubstituteColumn() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processSubstituteColumn started");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processSubstituteColumn starting";

		String substituteColumn = getAndRemoveInstruction("substituteColumn");
		String[] parts = substituteColumn.split("(?<!\\\\),");
		String substituteColumnNulls = getAndRemoveInstruction("substituteColumnNulls");

		if (parts.length < 2) {
			outputRecordSet
					.addWarningMessage("Missing parts to [substituteColumn]. Expected 'substituteColumn=col,value' - only see "
							+ parts.length
							+ " parts ("
							+ substituteColumn
							+ ")");
			return;
		}

		String newValue="";
		for(int instrParts=1; instrParts<parts.length; instrParts++){
			newValue+=parts[instrParts].replace("\\,",",");
			if(instrParts<parts.length-1){
				newValue+=",";
			}
		}
		String colName = parts[0];

		RecordSetMetaData rsmd = outputRecordSet.getMetaData();
		int col = getColumnIndexBySystemNm(rsmd, colName);

		if (col < 0) {
			outputRecordSet
					.addWarningMessage("Couldn't find column " + colName);
			return;
		}

		for (int row = 0; row < outputRecordSet.getRowCount(); row++) {
			RecordRow record = outputRecordSet.getRow(row);
			if(record.getString(col)==null && substituteColumnNulls.equalsIgnoreCase("N")) {
				continue;
			}
			outputRecordSet.setValue(row, col, newValue);
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	public void processBulletBreakFlagged() throws VermilionFatalException, VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processBulletBreakFlagged started");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processBulletBreakFlagged starting";

		if (! getAndRemoveInstruction("bulletBreakFlagged").equalsIgnoreCase("Y")) {
			return;
		}
		
		if (outputRecordSet.getRowCount() == 0) {
			writeOutputRecordset();
		}

		if (outputRecordSet.getColumnCount() != 2) {
			outputRecordSet
					.addWarningMessage("BulletBreaksFlagged is only valid where you have 2 and only 2 output columns");
			return;
		}

		if (!outputRecordSet.getMetaData().getField(0).getFieldType()
				.equals("html")) {
			outputRecordSet
					.addWarningMessage("BulletBreaksFlagged is only valid where the first output column is of type html");
			return;
		}

		RecordSet outRs = new RecordSet(outputRecordSet.getMetaData());

		try {
			String htmlTx = null;
			int newRow;
			String newTx = null;

			for (int row = 0; row < outputRecordSet.getRowCount(); row++) {
				htmlTx = outputRecordSet.getRow(row).getString(0)
						.replaceAll("\\<BR>", "\\{BR}").replaceAll("", "")
						.replaceAll("<LI>", "\\<LI><P>")
						.replaceAll("</LI>", "\\</LI></P>");

				// only process if we have a </P>
				if (htmlTx.toUpperCase().indexOf("</P>") == -1) {
					outRs.addRow(outputRecordSet.getRow(row));
					continue;
				}
				String[] parts = htmlTx.split("</P>");

				if (parts.length < 2) {
					continue;
				}

				int isBullet = 0;

				for (int i = 0; i < parts.length; i++, isBullet = 0) {
					if (parts[i].equals("</HTML>")) {
						continue;
					}

					// found list/bullet
					if (parts[i].contains("<LI>"))
						isBullet = 1;
					else
						isBullet = 0;

					// remove all list tags
					parts[i] = parts[i].replaceAll("\\<LI>", "").replaceAll(
							"\\<LI>", "");

					newRow = outRs.addRow();
					newTx = removeFormatting(parts[i].replace("\\{BR}",
							"\\<BR>")) + "</P>";

					outRs.setValue(newRow, 1, isBullet);
					outRs.setValue(newRow, 0, newTx);
				}
			}
		} catch (Exception e) {
			logManager.error("Error occured in custom class procedure at '"
					+ runPos + "': " + e.toString());
			outRs.addWarningMessage("Error occured in custom class procedure at '"
					+ runPos + "': " + e.toString());
		}

		setRecordSet(outRs);

	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	public void processParagraphBreaker() throws VermilionFatalException,
			VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processParagraphBreaker starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processParagraphBreaker starting";
		
		String para1 = getAndRemoveInstruction("paragraphBreaks");
		String para2 = getAndRemoveInstruction("paragraphBreak");
		if (!(para1.equalsIgnoreCase("Y") || para2.equalsIgnoreCase("Y"))) {
			return;
		}
    	String breakHeaders = getAndRemoveInstruction("breakHeaders");
    	String removeFormatting = getAndRemoveInstruction("removeFormatting");
    	
		if (outputRecordSet.getColumnCount() != 1) {
			outputRecordSet
					.addWarningMessage("ParagraphBreaks is only valid where you have 1 and only 1 output column");
			return;
		}

		if (!outputRecordSet.getMetaData().getField(0).getFieldType()
				.equals("html")) {
			outputRecordSet
					.addWarningMessage("ParagraphBreaks is only valid where you have an html output column");
			return;
		}

		RecordSet outRs = new RecordSet(outputRecordSet.getMetaData());

		try {
			String htmlTx = null;
			int newRow;
			String newTx = null;
			String widowTx = "";

			for (int row = 0; row < outputRecordSet.getRowCount(); row++) {
				htmlTx = outputRecordSet.getRow(row).getString(0)
						.replaceAll("\\<BR>", "\\{BR}").replaceAll("", "")
						.replaceAll("<LI>", "\\<P>")
						.replaceAll("</LI>", "\\</P>");

				// only process if we have a </P>
				if (htmlTx.toUpperCase().indexOf("</P>") == -1) {
					outRs.addRow(outputRecordSet.getRow(row));
					continue;
				}
				String[] parts = htmlTx.split("</P>");

				for (int i = 0; i < parts.length; i++) {
					if (parts[i].equals("</HTML>")) {
						continue;
					}
					newTx = widowTx+parts[i].replace("\\{BR}","\\<BR>") + "</P>";
					if(removeFormatting.equalsIgnoreCase("Y")){
						newTx = removeFormatting(newTx).trim();
					} else{
						newTx = newTx.trim();
					}
					if(!breakHeaders.equalsIgnoreCase("Y") && newTx.matches("\\<P>\\<STRONG>.*?\\</STRONG>\\</P>")){
						widowTx = newTx;
					} else if ( ! newTx.equals("<P></P>") ) {
						newRow = outRs.addRow();
						outRs.setValue(newRow, 0, "<HTML>"+newTx+"</HTML>");
						widowTx="";
					}
				}
			}
		} catch (Exception e) {
			logManager.error("Error occured in custom class procedure at '"
					+ runPos + "': " + e.toString());
			outRs.addWarningMessage("Error occured in custom class procedure at '"
					+ runPos + "': " + e.toString());
		}

		setRecordSet(outRs);

	}

	protected String removeFormatting(String html) {
		// Strip away everything except BOLD and ITALICS
		// used by processParagraphBreaker
		return "<HTML>"
				+ (html.replaceAll("\\<BR>", "@@BR@@")
						.replaceAll("\\<STRONG>", "@@BOLD@@")
						.replaceAll("\\</STRONG>", "@@-BOLD@@")
						.replaceAll("\\<EM>", "@@ITALICS@@")
						.replaceAll("\\</EM>", "@@-ITALICS@@")
						.replaceAll("\\<.*?>", "")
						.replaceAll("@@BOLD@@", "\\<STRONG>")
						.replaceAll("@@-BOLD@@", "\\</STRONG>")
						.replaceAll("@@BR@@", "\\<BR>")
						.replaceAll("@@ITALICS@@", "\\<EM>").replaceAll(
						"@@-ITALICS@@", "\\</EM>")).trim() + "</HTML>";
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processDistinctValues() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processDistinctValues starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processDistinctValues starting";

		String column = getAndRemoveInstruction("distinctValues");

		if (column.equals("-")) {
			outputRecordSet
					.addWarningMessage("Missing instructions. If you are using the group command you need to provide both groupColumns and groupTotal instructions ");
			return;
		}

		RecordSet inputRS = inputRecordSetList.getRecordSet(0);
		RecordSetMetaData rsmdIn = inputRS.getMetaData();

		// Get the columns to group by
		int columnNum = getColumnIndexBySystemNm(rsmdIn, column);
		if (columnNum < 0) {
			return;
		}

		// Need to sort the input recordset in the order of the last column (we
		// can only do 1 column)
		inputRS.sort(columnNum);

		// now read the input RS and get the distinct values
		int rowCount = inputRS.getRowCount();
		RecordRow rw = null;
		boolean firstOutputRecord = true;
		String previousRow = "";
		String currentRow = "";
		int outputRowNum = 0;

		for (int rec = 0; rec < rowCount; rec++) {
			rw = inputRS.getRow(rec);

			currentRow = rw.getString(columnNum);

			if (!currentRow.equals(previousRow)) {
				if (firstOutputRecord) {
					firstOutputRecord = false;
				} else {
					outputRowNum = outputRecordSet.addRow();
					outputRecordSet.setValue(outputRowNum, 1, previousRow);
				}

				previousRow = currentRow;
			}
		}
		outputRowNum = outputRecordSet.addRow();
		outputRecordSet.setValue(outputRowNum, 1, previousRow);
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processAddUniqueId() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processAddUniqueId starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processAddUniqueId starting";

		getAndRemoveInstruction("addUniqueId");

		for (int row = 0; row < outputRecordSet.getRowCount(); row++) {
			outputRecordSet.setValue(row, 0, row + 1);
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processMergeRecords() throws VermilionNonFatalException {
		// this is one of the few commands that must be run on the input RST,
		// otherwise you don't know the number of columns in each record
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processMergeRecords starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processMergeRecords starting";

		getAndRemoveInstruction("mergeRecords");

		RecordSet inputRS = inputRecordSetList.getRecordSet(0);

		RecordSetMetaData rsmdIn = inputRS.getMetaData();

		int outputCol = 0;
		int outputRow = 0;

		for (int row = 0; row < inputRS.getRowCount(); row++) {
			outputRow = outputRecordSet.addRow();
			// write out the 1st record
			RecordRow record1 = inputRS.getRow(row);

			for (int col = 0; col < rsmdIn.getColumnCount(); col++) {
				outputRecordSet.setValue(outputRow, outputCol++,
						record1.getString(col));
			}
			// now write out the 2nd record
			RecordRow record2 = inputRS.getRow(++row);

			for (int col = 1; col < rsmdIn.getColumnCount(); col++) {
				outputRecordSet.setValue(outputRow, outputCol++,
						record2.getString(col));
			}

			outputCol = 0;
		}
	}
	
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processConcatenateColumn() throws VermilionNonFatalException {
		// read all the input commentary records, and merge them all into one - with a single space between.
		// output is placed in the first row / first column of the output recordset (which obviously requires a Text or HTML format to receive it)
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processConcatenateColumn starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processConcatenateColumn starting";

		String concatenateColumnNm = getAndRemoveInstruction("concatenateColumn");
		String removeLinebreaks = getAndRemoveInstruction("concatenateColumnRemoveLinebreaks");

		RecordSet inputRS = inputRecordSetList.getRecordSet(0);
		RecordSetMetaData rsmdIn = inputRS.getMetaData();

		int conCol=0;
		for (int col = 0; col < rsmdIn.getColumnCount(); col++) {
			if ( rsmdIn.getField(col).getFieldName().equals(concatenateColumnNm) ) {
				conCol=col;
			}
		}
		
		int outputCol = 0;
		int outputRow = 0;

		String result = "";
		String text = "";
		boolean hasHTML=false;
		for (int row = 0; row < inputRS.getRowCount(); row++) {
			// write out the 1st record
			text = inputRS.getRow(row).getString(conCol);
			System.out.println("Collected row text : "+text);
			if ( text.contains("<HTML") ) {
				text = text.replace("<HTML>","");
				text = text.replace("</HTML>","");
				hasHTML=true;
			}
			if ( removeLinebreaks.equals("Y") ) {
				text = text.replace("<BR>", "");
				text = text.replace("</BR>", "");
				text = text.replace("<P>", "");
				text = text.replace("</P>", "");
			}
			result=result+text+" ";
		}
		if ( hasHTML ) {
			result="<HTML>"+result+"</HTML>";
		}
		int newRow = outputRecordSet.addRow();
		outputRecordSet.setValue(newRow, 0,  result);
		System.out.println("processConcatenateColumn4788 result = ["+result+"] in the output");
		
	}
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processSetData() throws VermilionNonFatalException,
			VermilionFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "setData starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processSetData starting";

		String setDataString = getAndRemoveInstruction("setData");
		String setDataParts[] = setDataString.split(",");

		// warning if insufficient instructions passed
		if (setDataParts.length < 3) {
			WarningMessage msgUserError = new WarningMessage(0, "DTP Error",
					WarningMessage.OUTPUT_TO_LOG_ONLY,
					"setData instructions require three arguments. You provided: "
							+ setDataString);
			outputRecordSet.addWarningMessage(msgUserError);
			return;
		}

		// confirm that the instructions are valid arguments
		int targetRow = -1;
		int targetCol = -1;
		boolean validArguments = true;
		try {
			targetRow = Integer.parseInt(setDataParts[0]);
		} catch (NumberFormatException E) {
			WarningMessage msgUserError = new WarningMessage(0, "DTP Error",
					WarningMessage.OUTPUT_TO_LOG_ONLY,
					"first argument of setData must be an integer. You provided: "
							+ setDataParts[0]);
			outputRecordSet.addWarningMessage(msgUserError);
			validArguments = false;
		}
		try{
			targetCol = Integer.parseInt(setDataParts[1]);
		}
		catch(NumberFormatException E){
			for (int colNum = 0; colNum < outputRecordSet.getColumnCount(); colNum++) {
				String outputColumnSysName = outputRecordSet.getMetaData(colNum).getFieldName();
				String outputColumnPrefName = outputRecordSet.getMetaData(colNum).getPreferredName();
				if (outputColumnSysName.equals(setDataParts[1])) {
					targetCol = colNum;
					break;
				} else if (outputColumnPrefName != null&& outputColumnPrefName.equals(setDataParts[1])) {
					targetCol = colNum;
					break;
				}
			}
			if(targetCol<0){
				WarningMessage msgUserError = 
						new WarningMessage(0, "DTP Error",  WarningMessage.OUTPUT_TO_LOG_ONLY,
								"second argument of setData must be an integer or valid column name. You provided: "+setDataParts[1]);
					outputRecordSet.addWarningMessage(msgUserError);
					validArguments=false;
			}
			
		}
		if(outputRecordSet.getColumnCount()<=targetCol){
			WarningMessage msgUserError = 
				new WarningMessage(0, "DTP Error",  WarningMessage.OUTPUT_TO_LOG_ONLY,
						"second argument must be less than the number of columns in the output record " +
						"set ("+outputRecordSet.getColumnCount()+"). You provided: "+setDataParts[1]);
			outputRecordSet.addWarningMessage(msgUserError);
			validArguments=false;				
		}
		
		if (outputRecordSet.getColumnCount() <= targetCol) {
			WarningMessage msgUserError = new WarningMessage(0, "DTP Error",
					WarningMessage.OUTPUT_TO_LOG_ONLY,
					"second argument must be less than the number of columns in the output record "
							+ "set (" + outputRecordSet.getColumnCount()
							+ "). You provided: " + setDataParts[1]);
			outputRecordSet.addWarningMessage(msgUserError);
			validArguments = false;
		}
		if (!validArguments){
			return;
		}
		
		while (outputRecordSet.getRowCount() <= targetRow) {
			outputRecordSet.addRow();
		}

		// setValue could contain commas (our instruction delimiter)
		String setValue = setDataString.substring(2 + setDataParts[0].length() + setDataParts[1].length());
		
		outputRecordSet.setValue(targetRow, targetCol, setValue);
		return;
	}
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processSetValue() throws VermilionNonFatalException, VermilionFatalException {
		// WARNING
		// setValue was added as an experiment but did not have time to do full testing. 
		// WARNING.
		
		// setValue = <columnName> = <expression, including entity and datasource substitution values),
		// e.g.
		//  setValue=columnNm=123
		//  setValue=columnNm=/if($columnNm2$=CASH,CASH,$columnNm$)
		//  setValue=columnNm=/if($_Fund.Show Me$=1,1,0)
		//
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "setData starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processSetValue starting";

		RecordSet inputRS = inputRecordSetList.getRecordSet(0);
		RecordSetMetaData rsmdIn = inputRS.getMetaData();
		int colCount=rsmdIn.getColumnCount();
		String[] columns = new String[colCount];
		for(int c=0; c<colCount; c++) {
			columns[c]=rsmdIn.getField(c).getFieldName();
		}

		String resolveSubVarOutputLang = getAndRemoveInstruction("resolveSubVarOutputLang");		
		String setInstruction = getAndRemoveInstruction("setValue");
		String setInstructionParts[] = new String[2];
		

		//setInstruction = ComponentUtility.parseText(outputRecordSet.getTid(),
		//		StringHelper.getResolveSubVars(setInstruction, parameterList, true), resolveSubVarOutputLang);
		
		setInstruction = StringHelper.getResolveSubVars(setInstruction, parameterList, true);
		
		runPos="Resolving instructions "+setInstruction;
		outputRecordSet.addWarningMessage(runPos);

		int x=setInstruction.indexOf("=");
		setInstructionParts[0]=setInstruction.substring(0, x);
		setInstructionParts[1]=setInstruction.substring(x+1);
		
		

		
		// warning if insufficient instructions passed
		if (setInstructionParts.length < 2) {
			WarningMessage msgUserError = new WarningMessage(0, "DTP Error",
					WarningMessage.OUTPUT_TO_LOG_ONLY,
					"setData instructions require three arguments. You provided: "
							+ setInstructionParts);
			outputRecordSet.addWarningMessage(msgUserError);
			return;
		}

		// confirm that the instructions are valid arguments
		int targetCol = -1;
		boolean validArguments = true;
		
		try{
			targetCol = Integer.parseInt(setInstructionParts[0]);
		}
		catch(NumberFormatException E){
			for (int colNum = 0; colNum < outputRecordSet.getColumnCount(); colNum++) {
				String outputColumnSysName = outputRecordSet.getMetaData(colNum).getFieldName();
				String outputColumnPrefName = outputRecordSet.getMetaData(colNum).getPreferredName();
				if (outputColumnSysName.equals(setInstructionParts[0])) {
					targetCol = colNum;
					break;
				} else if (outputColumnPrefName != null && outputColumnPrefName.equals(setInstructionParts[0])) {
					targetCol = colNum;
					break;
				}
			}
			if(targetCol<0){
				WarningMessage msgUserError = 
						new WarningMessage(0, "DTP Error",  WarningMessage.OUTPUT_TO_LOG_ONLY,
								"first argument of setValue must be an integer or valid column name. You provided: "+setInstructionParts[0]);
					outputRecordSet.addWarningMessage(msgUserError);
					validArguments=false;
			}
			
		}
		if(outputRecordSet.getColumnCount()<=targetCol){
			WarningMessage msgUserError = 
				new WarningMessage(0, "DTP Error",  WarningMessage.OUTPUT_TO_LOG_ONLY,
						"first argument must be less than the number of columns in the output record " +
						"setValue ("+outputRecordSet.getColumnCount()+"). You provided: "+setInstructionParts[0]);
			outputRecordSet.addWarningMessage(msgUserError);
			validArguments=false;				
		}
		
		if (outputRecordSet.getColumnCount() <= targetCol) {
			WarningMessage msgUserError = new WarningMessage(0, "DTP Error",
					WarningMessage.OUTPUT_TO_LOG_ONLY,
					"first argument must be less than the number of columns in the output record "
							+ "setValue (" + outputRecordSet.getColumnCount()
							+ "). You provided: " + setInstructionParts[0]);
			outputRecordSet.addWarningMessage(msgUserError);
			validArguments = false;
		}
		if (!validArguments){
			return;
		}
		

		runPos="Working through rows";
		String setValue, setValue0, var;
		RecordRow rw;
		int p;
		for(int row=0; row<inputRS.getRowCount(); row++) {
			runPos="Working through row "+row;
			//int newRow=outputRecordSet.addRow();
			int newRow=row;
			rw=inputRS.getRow(row);
			//outputRecordSet.setRow(newRow, rw);
		
			setValue = setInstructionParts[1];
			setValue0=setValue;
			if ( setValue.indexOf("$")> 0 ) {
				for(int c=0; c<colCount; c++) {
					var="$"+columns[c]+"$";
					p=setValue.indexOf(var);
					if ( p>0 ) {
						setValue=setValue.replace(var, rw.getString(c));
						outputRecordSet.addWarningMessage("row ("+row+") Seen "+var+" at "+p+" in "+setValue0+" ... now = "+setValue);
					}
				}
			}
			
			setValue = ComponentUtility.parseText(outputRecordSet.getTid(),
					StringHelper.getResolveSubVars(setValue, parameterList, true), resolveSubVarOutputLang);
			
			outputRecordSet.setValue(newRow, targetCol, setValue);
		}
		runPos="Done setValue";
	}
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processloopDSOverEntityRS() throws VermilionNonFatalException,
			VermilionFatalException {
			if (trace) {
				WarningMessage msg = new WarningMessage(0, "Rendering",
						WarningMessage.DONT_OUTPUT,
						"processloopDSOverEntityRS starting");
				outputRecordSet.addWarningMessage(msg);
			}
			runPos = "processloopDSOverEntityRS starting";
	
			loopDSOverEntityRS(getAndRemoveInstruction("loopDSOverEntityRS"),
				0, parameterList, "", false);
	}
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void loopDSOverEntityRS(String targetDS, int recordSetsUsed,
			ParameterList parametersSoFar, String parametersAsString,
			boolean appendParameters) throws VermilionNonFatalException,
			VermilionFatalException {

		if (recordSetsUsed == inputRecordSetList.getRecordSetCount()) {
			int outputRowCount = outputRecordSet.getRowCount();
			int outputColumnCount = outputRecordSet.getColumnCount();
			DynamicDGSCall dgsHelper = new DynamicDGSCall(
					outputRecordSet.getTid());
			writeOutputRecordset(dgsHelper.getDGSDataByDstCode(targetDS,
					parametersSoFar));
			if (appendParameters) {
				for (int rowToAppend = outputRowCount; rowToAppend < outputRecordSet
						.getRowCount(); rowToAppend++) {
					outputRecordSet.setValue(rowToAppend,
							outputColumnCount - 1, parametersAsString);
				}
			}
		} else {
			if (recordSetsUsed == 0) {
				
				deleteOutputRecordset();
				if (isInstructionValuePassed("appendParameters")
						&& getAndRemoveInstruction("appendParameters")
								.equalsIgnoreCase("y")) {
					appendParameters = true;
				}
			}
			RecordSet thisParameterRS = inputRecordSetList
					.getRecordSet(recordSetsUsed);
			int rowIDColumnNum = -1;
			int entityCodeColumnNum = -1;
			for (int c = 0; c < thisParameterRS.getColumnCount(); c++) {
				MetaDataField columnMetaData = thisParameterRS.getMetaData(c);
				if (columnMetaData.getFieldName().equals("ROW_ID")) {
					rowIDColumnNum = c;
				}
				if (columnMetaData.getFieldName().equals("ENTITY_CODE")) {
					entityCodeColumnNum = c;
				}
			}
			if (rowIDColumnNum < 0 || entityCodeColumnNum < 0) {
				if (rowIDColumnNum < 0) {
					WarningMessage msg = new WarningMessage(0, "DTP Error",
							WarningMessage.DONT_OUTPUT,
							thisParameterRS.getDataSourceName()
									+ "lacks ROW_ID column.");
					outputRecordSet.addWarningMessage(msg);
				}
				if (entityCodeColumnNum < 0) {
					WarningMessage msg = new WarningMessage(0, "DTP Error",
							WarningMessage.DONT_OUTPUT,
							thisParameterRS.getDataSourceName()
									+ "lacks ENTITY_CODE column.");
					outputRecordSet.addWarningMessage(msg);
				}
				return;
			}
			for (int recRowNum = 0; recRowNum < thisParameterRS.getRowCount(); recRowNum++) {
				ParameterList parametersIncludingThisRow = new ParameterList(
						parametersSoFar.getTid(), parametersSoFar.getUserId());
				for (int parameterNum = 0; parameterNum < parametersSoFar
						.getParameterCount(); parameterNum++) {
					parametersIncludingThisRow
							.addParameterRecord(parametersSoFar
									.getParameterRecord(parameterNum));
				}
				int parameterRowId = thisParameterRS.getInteger(recRowNum,
						rowIDColumnNum);
				String parameterEntityCode = thisParameterRS.getString(
						recRowNum, entityCodeColumnNum);

				ParameterRecord nextParameterRecord = new ParameterRecord(
						parametersSoFar.getTid(), parameterEntityCode,
						parameterRowId);
				parametersIncludingThisRow
						.addParameterRecord(nextParameterRecord);
				String nextParametersAsString = nextParameterRecord
						.getParameterCode()
						+ ":"
						+ nextParameterRecord.getKeyValue();
				if (!parametersAsString.equals("")) {
					nextParametersAsString = nextParametersAsString + "|"
							+ parametersAsString;
				}
				loopDSOverEntityRS(targetDS, recordSetsUsed + 1,
						parametersIncludingThisRow, nextParametersAsString,
						appendParameters);
			}
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------

	private void processBalanceColumns(String balanceType) throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processSplitIntoBalancedColumns started");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processSplitIntoBalancedColumns starting";
		
		try {

			String splitFactorString = getAndRemoveInstruction(balanceType+"IntoBalancedColumns");
			int splitFactor;
			try {
				splitFactor = Integer.parseInt(splitFactorString);
			} catch (Exception NumberFormatException) {
				WarningMessage argumentNotInteger = new WarningMessage(0,
						"DTP Error", WarningMessage.OUTPUT_TO_LOG_ONLY,
						balanceType+"IntoBalancedColumns argument must be an integer. '"
								+ splitFactorString + "' is not an integer.");
				outputRecordSet.addWarningMessage(argumentNotInteger);
				return;
			}
			runPos = "confirming column count";
			// confirm that the output is # x input in column formation
			RecordSet inputRecordSet = inputRecordSetList.getRecordSet(0);
			RecordSetMetaData inputRSMD = inputRecordSet.getMetaData();
			RecordSetMetaData outputRSMD = outputRecordSet.getMetaData();
			int inputColumnCount = inputRSMD.getColumnCount();
			int outputColumnCount = outputRSMD.getColumnCount();
			if (inputColumnCount * splitFactor != outputColumnCount) {
				WarningMessage wrongNumberOfColumns = new WarningMessage(0,
						"DTP Error", WarningMessage.OUTPUT_TO_LOG_ONLY,
						"Input has " + inputColumnCount
								+ " columns and instruction splits by "
								+ splitFactor + ", so output needs "
								+ (inputColumnCount * splitFactor)
								+ " columns but output only has "
								+ outputColumnCount + "columns.");
				outputRecordSet.addWarningMessage(wrongNumberOfColumns);
				return;
			}
			runPos = "confirming column types";
			boolean validColumns = true;
			for (int outputColumnNumber = 0; outputColumnNumber < inputColumnCount; outputColumnNumber++) {
				String outputFieldType = outputRSMD
						.getField(outputColumnNumber).getFieldType();
				String inputFieldType = inputRSMD.getField(
						outputColumnNumber % inputColumnCount).getFieldType();
				String outputFieldName = outputRSMD
						.getField(outputColumnNumber).getFieldName();
				String inputFieldName = inputRSMD.getField(
						outputColumnNumber % inputColumnCount).getFieldName();
				if (!outputFieldType.equals(inputFieldType)) {
					WarningMessage wrongColumnType = new WarningMessage(0,
							"DTP Error", WarningMessage.OUTPUT_TO_LOG_ONLY,
							"Ouput column " + outputColumnNumber + " "
									+ outputFieldName + "is of type "
									+ outputFieldType
									+ ",which does not match input column "
									+ outputColumnNumber % splitFactor + " "
									+ inputFieldName + " of type "
									+ inputFieldType + ".");
					outputRecordSet.addWarningMessage(wrongColumnType);
					validColumns = false;
				}
			}
			if (!validColumns) {
				return;
			}
			runPos = "moving data";
			int inputRowCount = inputRecordSet.getRowCount();
			int outputRowCount = (int) Math.ceil((double) inputRowCount
					/ splitFactor);
			if(balanceType.equalsIgnoreCase("split")){
				for (int inputRowNumber = outputRowCount; inputRowNumber < inputRowCount; inputRowNumber++) {
					int destinationRow = inputRowNumber % outputRowCount;
					int destinationColumn = inputRowNumber / outputRowCount;
					for (int inputColumnNumber = 0; inputColumnNumber < inputColumnCount; inputColumnNumber++) {
						outputRecordSet.setValue(destinationRow, inputColumnCount
								* destinationColumn + inputColumnNumber,
								inputRecordSet.getString(inputRowNumber,
										inputColumnNumber));
					}
				}
			}
			if(balanceType.equalsIgnoreCase("wrap")){
				for (int inputRowNumber = 1; inputRowNumber < inputRowCount; inputRowNumber++) {
					int destinationRow = inputRowNumber / splitFactor;
					int destinationColumn = inputRowNumber % splitFactor;
					for (int inputColumnNumber = 0; inputColumnNumber < inputColumnCount; inputColumnNumber++) {
						outputRecordSet.setValue(destinationRow, inputColumnCount
								* destinationColumn + inputColumnNumber,
								inputRecordSet.getString(inputRowNumber,
										inputColumnNumber));
					}
				}
			}
			runPos = "trimming rows";
			while (outputRecordSet.getRowCount() > outputRowCount) {
				outputRecordSet.removeRow(outputRowCount);
			}
		} catch (Exception E) {
			WarningMessage runPosError = new WarningMessage(0, "DTP Error",
					WarningMessage.OUTPUT_TO_LOG_ONLY, "error at " + runPos
							+ ":" + E.toString());
			outputRecordSet.addWarningMessage(runPosError);
			return;
		}
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void processRowPadding() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processRowPadding started");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processRowPadding starting";
		try {
	
//			allParams.put("padRowsToMultiple",				new String[] { "Y", "1",
//			"# (Adds extra blank rows to recordset until total number of rows is a multiple of # (including 1 * # = #)" });
//	allParams.put("padRowsByGroup",					new String[] { "Y", "-",
//			"{column name} (If using padRowsToMultiple, this will insert them with the same content in the grouped column)" });
//	allParams.put("padRowsLocation",				new String[] { "Y", "After",
//			"Before|After (If using padRowsToMultiple, this will instruct DTP whether to prepend or append additional rows)" });
//	
			String rowFactorString = getAndRemoveInstruction("padRowsToMultiple");
			int rowFactor;
			try {
				rowFactor = Integer.parseInt(rowFactorString);
			} catch (Exception NumberFormatException) {
				WarningMessage argumentNotInteger = new WarningMessage(0,
						"DTP Error", WarningMessage.OUTPUT_TO_LOG_ONLY,
						"padRowsToMultiple argument must be an integer. '"
								+ rowFactorString + "' is not an integer.");
				outputRecordSet.addWarningMessage(argumentNotInteger);
				return;
			}
			String paddingLocation = getAndRemoveInstruction("padRowsLocation");
			if(!paddingLocation.equalsIgnoreCase("before") && !paddingLocation.equalsIgnoreCase("after") ){
				WarningMessage argumentNotInteger = new WarningMessage(0,
						"DTP Error", WarningMessage.OUTPUT_TO_LOG_ONLY,
						"paddingLocation argument must be a valid option. '"
								+ paddingLocation + "' is neither 'before' nor 'after'.");
				outputRecordSet.addWarningMessage(argumentNotInteger);
				return;
			}
			String groupColString = getAndRemoveInstruction("padRowsByGroup");

			RecordSetMetaData outputRSMetadata = outputRecordSet.getMetaData();
			if(groupColString.equalsIgnoreCase("-")){
				//No individual groups. Just add extra rows to top or bottom.
				for(int i=outputRecordSet.getRowCount();i%rowFactor>0;i++){
					if(paddingLocation.equalsIgnoreCase("before")){
						RecordSet newRecSet = new RecordSet(outputRecordSet.getMetaData());
						newRecSet.addRow();
						for (int j = 0; j < outputRecordSet.getRowCount(); j++) {
							newRecSet.addRow(outputRecordSet.getRow(j));
						}
						setRecordSet(newRecSet);
				    }else if(paddingLocation.equalsIgnoreCase("after")){
				    	outputRecordSet.addRow();
				    }
				}
			}else{
				//We *are* padding individual groups
				
				int groupColNum=-1;
				groupColNum= getColumnIndexBySystemNm(outputRSMetadata,groupColString);
				if(groupColNum<0){
					WarningMessage argumentNotValidColumn = new WarningMessage(0,
							"DTP Error", WarningMessage.OUTPUT_TO_LOG_ONLY,
							"padRowsByGroup argument is not a valid column name. '"
									+ groupColString + "' is not a system columnm name of the output record set.");
					outputRecordSet.addWarningMessage(argumentNotValidColumn);
					return;
				}
				
				//collect a running count of members of each group
				HashMap<String,Integer> groupMemberCount = new HashMap<String,Integer>();
				for(int i=0;i<outputRecordSet.getRowCount();i++){
					String groupOfRow = outputRecordSet.getString(i, groupColNum);
					if(groupMemberCount.containsKey(groupOfRow)){
						groupMemberCount.put(groupOfRow,groupMemberCount.get(groupOfRow)+1);
					}else{
						groupMemberCount.put(groupOfRow,1);
					}
				}
				
				//for each member of groupMemberCount, add a row to the datasort
				for (Map.Entry<String, Integer> countEntry : groupMemberCount.entrySet()) {
					String groupOfRows = countEntry.getKey();
					int numberOfRows = countEntry.getValue();
					for(int i=numberOfRows;i%rowFactor>0;i++){
						if(paddingLocation.equalsIgnoreCase("before")){
							RecordSet newRecSet = new RecordSet(outputRecordSet.getMetaData());
							newRecSet.addRow();

							for (int j = 0; j < outputRecordSet.getRowCount(); j++) {
								newRecSet.addRow(outputRecordSet.getRow(j));
							}

							setRecordSet(newRecSet);
							outputRecordSet.setValue(0, groupColNum, groupOfRows);
					    }else if(paddingLocation.equalsIgnoreCase("after")){
					    	outputRecordSet.addRow();
							outputRecordSet.setValue(outputRecordSet.getRowCount()-1, groupColNum, groupOfRows);
					    }
					}
				}
			}



			
		} catch (Exception E) {
			WarningMessage runPosError = new WarningMessage(0, "DTP Error",
					WarningMessage.OUTPUT_TO_LOG_ONLY, "error at " + runPos
							+ ":" + E.toString());
			outputRecordSet.addWarningMessage(runPosError);
			return;
		}
	}
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------

	protected void processStripTags() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processStripTags started");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processStripTags starting";
		
		String stripTags = getAndRemoveInstruction("stripTags");
		String stripTagsExceptions = getAndRemoveInstruction("stripTagsException") + getAndRemoveInstruction("stripTagsExceptions");

		int colNumToStrip = -1;
		if (stripTags.equalsIgnoreCase("N")) {
			return;
		}else if(!stripTags.equalsIgnoreCase("Y")){
			colNumToStrip = getColumnIndexBySystemNm(outputRecordSet.getMetaData(),stripTags);
		}

		for (int row = 0; row < outputRecordSet.getRowCount(); row++) {
			for (int col = Math.max(0, colNumToStrip);
					(colNumToStrip < 0 && col < outputRecordSet.getColumnCount()) || (col == colNumToStrip);
					col++) {
				RecordRow record = outputRecordSet.getRow(row);
				if (record.getString(col) == null) {
					continue;
				}
				outputRecordSet.setValue(
						row,
						col,
						removeAllFormatting(record.getString(col),
								stripTagsExceptions));
			}
		}
	}
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------

	private void processUndoHtmlEscapes() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processUndoHTMLEscapes started");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processUndoHTMLEscapes starting";
		

	    String stripTags = getAndRemoveInstruction("undoHTMLEscapes");
	    int targetColNum = -1;
	    try{
	      targetColNum = Integer.parseInt(stripTags);
	    }
	    catch (NumberFormatException localNumberFormatException) {}
	    for (int row = 0; row < this.outputRecordSet.getRowCount(); row++)
	    {
	      int maxColumn;
	      if (targetColNum < 0) {
	        maxColumn = this.outputRecordSet.getColumnCount()-1;
	      } else {
	        maxColumn = targetColNum;
	      }
	      for (int col = Math.max(targetColNum, 0); col <= maxColumn; col++)
	      {
	        String setValue = this.outputRecordSet.getString(row, col);
	        setValue = setValue.replaceAll("&gt;", ">");
	        setValue = setValue.replaceAll("&lt;", "<");
	        setValue = setValue.replaceAll("&nbsp;", " ");
	        this.outputRecordSet.setValue(row, col, setValue);
	      }
	    }
	  }
	
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	/* Change a vertical list of values into an monthly crosstab (years running vertically, months running horizontally)
	 The output recordset is assumed to contain the following columns
	 Year, Jan, Feb, Mar, Apr, May, Jun, Jul, Aug, Sep, Oct, Nov, Dec, YTD
	 
	 YTD is computed as APY = (1+Jan)*(1+Feb) *(1+Mar) *(1+Apr) *(1+May) *(1+Jun) *(1+Jul) *(1+Aug) *(1+Sep) *(1+Oct) *(1+Nov) *(1+Dec)  1
	 */
	private void processTransformMonthlyMatrix() throws VermilionNonFatalException {
		//System.out.println("processTransformMonthlyMatrix starting #############################################################");
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "transformMonthlyMatrix started");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "transformMonthlyMatrix starting";
		
		runPos = "collecting input to Transform Monthly Matrix";
		int dateColumnNb=-1;
		int returnColumnNb=-1;
		
		String transformMonthlyMatrix = getAndRemoveInstruction("transformMonthlyMatrix"); // This is required to remove the instruction from the list
		String dateColumnNm = getAndRemoveInstruction("transformMonthlyMatrixDateColumn");
		String returnColumnNm = getAndRemoveInstruction("transformMonthlyMatrixReturnColumn");
		
		//outputRecordSet.addWarningMessage("Looking for "+dateColumnNm+" and "+returnColumnNm);

		RecordSet inputRS = inputRecordSetList.getRecordSet(0);
		RecordSetMetaData rsmdIn = inputRS.getMetaData();
		RecordSet outRs = new RecordSet(outputRecordSet.getMetaData());
		
		//System.out.println("Looking for "+dateColumnNm+" and "+returnColumnNm);
		if ( isNumeric(dateColumnNm) ) {
			//outputRecordSet.addWarningMessage("Date column is numeric:"+dateColumnNm);
			dateColumnNb=new Integer(dateColumnNm).intValue()-1;
		}
		else {
			//outputRecordSet.addWarningMessage("Date column is not numeric:"+dateColumnNm);
			dateColumnNb=getColumnIndexBySystemNm(rsmdIn, dateColumnNm);
		}
		if (dateColumnNb < 0) {
			outputRecordSet.addWarningMessage("Did not recognize transformMonthlyMatrixDateColumn:"+dateColumnNm);
			return;
		}

		
		if ( isNumeric(returnColumnNm) ) {
			//outputRecordSet.addWarningMessage("Return column is numeric:"+returnColumnNm);
			returnColumnNb=new Integer(returnColumnNm).intValue()-1;
		}
		else {
			//outputRecordSet.addWarningMessage("Return column is not numeric:"+returnColumnNm);
			returnColumnNb=getColumnIndexBySystemNm(rsmdIn, returnColumnNm);
		}
		if (returnColumnNb < 0) {
			outputRecordSet.addWarningMessage("Did not recognize transformMonthlyMatrixReturnColumn:"+returnColumnNm);
			return;
		}
		
		
		if ( outputRecordSet.getColumnCount()< 1+12 ) {
			outputRecordSet.addWarningMessage("Output recordset needs to contain 1+12 or 1+12+1 (year/months x12/ytd) columns to receive tranformation.");
		}
		//System.out.println("Looking for columns ["+dateColumnNb+"] and ["+returnColumnNb+"]");
		
		
		HashMap<Integer, BigDecimal[]> yearsHmp = new HashMap<Integer, BigDecimal[]>();
		BigDecimal                     monthValue=new BigDecimal(-1);
		String                         dtColumnValue="na";
		int                            year0=99999;
		int                            yearX=0;
		
		try {
			//System.out.println("Looking at "+inputRS.toString());
			SimpleDateFormat inputDateFormat;
			SimpleDateFormat outputDateFormat;
			Date endCalDate;
			String sMonth;
			String sYear;
			BigDecimal[] yearArray;
			for(int row=0; row<inputRS.getRowCount(); row++) {
				dtColumnValue="na";
				monthValue=new BigDecimal(-1);
				//System.out.println("Looking at "+row+" column "+dateColumnNb+"/"+returnColumnNb);
				
				runPos="5303: Looking at Date column "+dateColumnNb+" on inputRecordset row ("+row+" of "+inputRS.getRowCount()+")";
				dtColumnValue=inputRS.getString(row, dateColumnNb);
				if ( !dtColumnValue.equals("") ) {
					runPos="5309: Looking at Value column "+returnColumnNb+" on inputRecordset row ("+row+" of "+inputRS.getRowCount()+")";
					if ( !inputRS.isEmpty(row,  returnColumnNb) ) {
						monthValue = inputRS.getDecimal(row, returnColumnNb); // returns MUST be in 0-1 range = 0%-100%
	
						runPos="5310: converting date value "+dtColumnValue+" on inputRecordset ("+row+") ("+monthValue+")";
						//System.out.println("Looking at "+row+" column "+dateColumnNb+"/"+returnColumnNb+"="+inputRS.getString(row, dateColumnNb)+"/"+inputRS.getDecimal(row, returnColumnNb));
						
						inputDateFormat =new SimpleDateFormat("dd MMMM yyyy 00:00:00");
						outputDateFormat=new SimpleDateFormat();
						endCalDate=inputDateFormat.parse(dtColumnValue);
						sMonth=new SimpleDateFormat("M").format(endCalDate);
						sYear=new SimpleDateFormat("yyyy").format(endCalDate);
	
						runPos="5319: Looking through inputRecordset ("+row+") ("+monthValue+")";
						int yearNb =new Integer(sYear);
						int monthNb=new Integer(sMonth).intValue()-1;

						//System.out.println("5322: Storing "+dtColumnValue+" ["+yearNb+"]["+monthNb+"]="+monthValue);					
						if ( yearNb>yearX ) {
							yearX=yearNb;
						}
						if ( yearNb<year0 ) {
							year0=yearNb;
						}
	
						runPos="5330: Looking through inputRecordset ("+row+") ("+monthValue+")";
						
						if ( yearsHmp.containsKey(yearNb) ) {
							yearArray = yearsHmp.get(yearNb);
							yearArray[monthNb]=monthValue;
							yearsHmp.put(yearNb, yearArray);
						}
						else {
							yearArray=new BigDecimal[13];
							yearArray[monthNb]=monthValue;
							yearsHmp.put(yearNb, yearArray);
						}
						//System.out.println(row+": Seen transformMonthlyMatrixReturn column value ("+dateColumnNb+")="+dtColumnValue+" = ("+returnColumnNb+")="+monthValue);
					}
				}
			}		
			//monthValue=inputRS.getFloat(row, returnColumnNb);
			
			runPos="Adding collected array to outputRecordSet (from "+year0+" to "+yearX+")";
			
			BigDecimal one=new BigDecimal(1);
			BigDecimal ytdRtn;
			for(int y=year0; y<=yearX; y++) {
				yearArray = (BigDecimal[])yearsHmp.get(y);
				int newRow=outputRecordSet.addRow();
				outputRecordSet.setValue(newRow, 0, y);
				ytdRtn=new BigDecimal(1);
				for(int m=0; m<=12; m++) {
					if ( yearArray[m] != null ) {
						outputRecordSet.setValue(newRow, m+1, yearArray[m].doubleValue());
						ytdRtn=ytdRtn.multiply(one.add(yearArray[m])); // YTD = (1+Vm1)*(1+Vm2)*(1+Vm3)....*(1+Vm12)
					}
				}
				if ( outputRecordSet.getColumnCount()==1+12+1 ) { // 1=year number, +12=months, +1=YTD
					outputRecordSet.setValue(newRow, 13, ytdRtn.subtract(one).doubleValue());
				}
			}
			//outputRecordSet.addWarningMessage("Done "+dateColumnNb+"/"+returnColumnNb);
			

		}
		catch (Exception e) {
			System.out.println("Could not convert at "+runPos);
			outputRecordSet.addWarningMessage("Could not convert at "+runPos);
		}
	}
	
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	private void processRotateData() throws VermilionNonFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "processRotateData started");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processRotateData starting";
		
		runPos = "collecting input to rotate";
		String rotateDataString = getAndRemoveInstruction("rotateData");
		RecordSet inputRecordSet = this.getInputRecordSet(rotateDataString, false);
		if(inputRecordSet == null){
			if(rotateDataString.equalsIgnoreCase("Y")){
				//just grab the first (presumably only) input record set
				inputRecordSet = this.inputRecordSetList.getRecordSet(0);
				
			} else if(rotateDataString.equalsIgnoreCase("Output")){
				//grab the post-other-instructions working output RS
				inputRecordSet = new RecordSet(outputRecordSet.getMetaData());
				for(int row=0;row<outputRecordSet.getRowCount();row++){
					inputRecordSet.addRow();
					for(int col=0;col<outputRecordSet.getColumnCount();col++){
						inputRecordSet.setValue(row, col, outputRecordSet.getString(row, col));
					}
				}
			} else{
				//user has specified an invalid record set
				inputRecordSet = this.getInputRecordSet(rotateDataString, true);
				return;
			}
		}
		String rotateFormatString = getAndRemoveInstruction("rotateDataFormatIntoText");
		Boolean rotateFormat = rotateFormatString.equalsIgnoreCase("Y");
        runPos = "Collecting metadata on the input";
        RecordSetMetaData inputRSMD = inputRecordSet.getMetaData();
        ArrayList<String> inputColTypes = new ArrayList<String>();
        ArrayList<String> inputColFormats = new ArrayList<String>();

        runPos = "Populating input metadata arrays";
        for (int c = 0; c < inputRSMD.getColumnCount(); c++) {
        	inputColTypes.add((String) inputRSMD.getField(c).getFieldType());
        	inputColFormats.add((String) inputRSMD.getField(c).getFieldDisplayFormat());
            //outputRecordSet.addWarningMessage("input column "+c +" is of type "+inputColTypes.get(c)+" with format "+inputColFormats.get(c));
        }
        
        runPos = "Collecting metadata on the output";
        RecordSetMetaData outputRSMD = outputRecordSet.getMetaData();
        ArrayList<String> outputColTypes = new ArrayList<String>();
        ArrayList<String> outputColFormats = new ArrayList<String>();

        runPos = "Populating output metadata arrays";
        for (int c = 0; c < outputRSMD.getColumnCount(); c++) {
        	outputColTypes.add((String) outputRSMD.getField(c).getFieldType());
            outputColFormats.add((String) outputRSMD.getField(c).getFieldDisplayFormat());
            //outputRecordSet.addWarningMessage("output column "+c +" is of type "+outputColTypes.get(c)+" with format "+outputColFormats.get(c));
        }
		
		runPos = "checking if a column has header-data that needs to be removed";
		String removeColumnName = getAndRemoveInstruction("rotateDataRemoveHeaderColumn");
		int removeColumnIndex = getColumnIndexBySystemNm(inputRSMD, removeColumnName, false);
		if(removeColumnIndex==-1 && ! removeColumnName.equalsIgnoreCase("-")){
			if(removeColumnName.equalsIgnoreCase("Y")){
				removeColumnIndex=0;
			}else{
				getColumnIndexBySystemNm(inputRSMD, removeColumnName, true);
			}
		}
		boolean removeColumn = (removeColumnIndex>-1);

		runPos="checking if the old headers need to be turned into a column";
		String addColumnName = getAndRemoveInstruction("rotateDataAddHeaderColumn");
		int addColumnIndex = getColumnIndexBySystemNm(outputRSMD, addColumnName, false);
		if(addColumnIndex==-1 && ! addColumnName.equalsIgnoreCase("-")){
			if(addColumnName.equalsIgnoreCase("Y")){
				addColumnIndex=0;
			}else{
				getColumnIndexBySystemNm(outputRSMD, addColumnName, true);
			}
		}
		boolean addColumn = (addColumnIndex>-1);
		
		runPos="populating outputRecordSet";
		deleteOutputRecordset();
		for(int newRow=0;newRow<inputRecordSet.getColumnCount() - (removeColumn?1:0) ; newRow++){
			outputRecordSet.addRow();
			for(int newCol=0;newCol<outputRecordSet.getColumnCount() && newCol<inputRecordSet.getRowCount() + (addColumn?1:0); newCol++){
				String newValue="";
				if(newCol==addColumnIndex){
					newValue=inputRecordSet.getMetaData().getField(newRow 
							+ (removeColumn?(newRow>=removeColumnIndex?1:0):0)).getPreferredName();
					if(newValue==null || newValue.equals("")){
						//the field didn't have a preferred name. Use the system name.
						newValue=inputRecordSet.getMetaData().getField(newRow 
								+ (removeColumn?(newRow>=removeColumnIndex?1:0):0)).getFieldName();
					}
				}else{
					newValue=inputRecordSet.getString(newCol - (addColumn?(newCol>=addColumnIndex?1:0):0),
							newRow + (removeColumn?(newRow>=removeColumnIndex?1:0):0));
					//if the source column is a numerical data type and the destination column is not the same, format it:
					String inputColType = inputColTypes.get(newRow + (removeColumn?(newRow>=removeColumnIndex?1:0):0));
					String inputColFormat = inputColFormats.get(newRow + (removeColumn?(newRow>=removeColumnIndex?1:0):0));
					String outputColType = outputColTypes.get(newCol);
					
					if(rotateFormat && (inputColType.equals("decimal") || inputColType.equals("integer"))
						&& ! (outputColType.equals("decimal") || outputColType.equals("integer"))){
			            //outputRecordSet.addWarningMessage("value " + newValue + "needs to be formatted as " + inputColType + ":" + inputColFormat );
						if(inputColFormat!=null && !inputColFormat.equals("") && newValue!=null && !newValue.equals("")){
							try{
								DecimalFormat inputFormat = new DecimalFormat(inputColFormat);
								BigDecimal newDecimal = new BigDecimal(newValue);
								newValue = inputFormat.format(newDecimal);
							}catch(IllegalArgumentException E){
								//somehow the output record set has an invalid format
							}
						}
					} else if(rotateFormat && inputColType.equals("datetime") &&  ! outputColType.equals("datetime")){
						if(inputColFormat!=null && !inputColFormat.equals("") && newValue!=null && !newValue.equals("")){
						//outputRecordSet.addWarningMessage("value " + newValue + "needs to be formatted as " + inputColType + ":" + inputColFormat );
							try{
								SimpleDateFormat vrsFormat = new SimpleDateFormat(DateHelper._VRS_DATE_FORMAT);
								Date inputDate = vrsFormat.parse(newValue);
								SimpleDateFormat inputFormat = new SimpleDateFormat(inputColFormat.replaceAll("m","M"));
								newValue = inputFormat.format(inputDate);
							}catch(ParseException E){
								//somehow the output record set has an invalid format
							}
						}
					}
					
				}
				outputRecordSet.setValue(newRow, newCol, newValue);
				
			}
		}
		
	}
		// -----------------------------------------------------------------------------------------------------------------------------------------------------------------

	protected String removeAllFormatting(String html, String exceptions) {
		// Strip away everything except the exceptions
		// return html.replaceAll("\\<.*?>", "")
		// return html.replaceAll("(?i)<(?!img|/img).*?>", "")
		// used by processStripTags
		return html.replaceAll("(?i)<(?!" + exceptions + ").*?>", "")
				.replaceAll("&nbsp;", "").replaceAll("&gt;", ">")
				.replaceAll("&lt;", "<").replaceAll("&quot;", "\"")
				.replaceAll("&amp;", "&").replaceAll("&copy;", "")
				.replaceAll("&reg;", "").replaceAll("&euro;", "").trim();
	}

	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	protected void suppressWarnings() throws VermilionNonFatalException,
			VermilionFatalException {
		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT, "suppressWarnings starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "suppressWarnings starting";
		
		String groupToSuppress = getAndRemoveInstruction("suppressWarnings");
		RecordSet warningHolder = new RecordSet(outputRecordSet.getMetaData());
		int warningsRemoved = 0;
		for (int warningNum = 0; warningNum < outputRecordSet
				.getWarningMessageCount(); warningNum++) {
			WarningMessage warningFromOutput = outputRecordSet
					.getWarningMessage(warningNum);
			if (!warningFromOutput.getMessageGroup().equals(groupToSuppress)) {
				warningHolder.addWarningMessage(warningFromOutput);
			} else
				warningsRemoved++;
		}
		WarningMessage msg2 = new WarningMessage(0, "Rendering",
				WarningMessage.DONT_OUTPUT, "suppressWarnings removed "
						+ warningsRemoved + " warning messages of group "
						+ groupToSuppress);
		outputRecordSet.addWarningMessage(msg2);

		outputRecordSet.clearWarningMessages();
		for (int warningNum = 0; warningNum < warningHolder
				.getWarningMessageCount(); warningNum++) {
			WarningMessage warningFromHolder = warningHolder
					.getWarningMessage(warningNum);
			outputRecordSet.addWarningMessage(warningFromHolder);
		}
	}

	// ----------------------------------------------------------------------------------------------------------------------------------
	protected void processManualOverride() throws VermilionNonFatalException,
			VermilionFatalException {
		runPos = "processManualOverride starting";

		if (trace) {
			WarningMessage msg = new WarningMessage(0, "Rendering",
					WarningMessage.DONT_OUTPUT,
					"processManualOverride starting");
			outputRecordSet.addWarningMessage(msg);
		}
		runPos = "processManualOverride starting";

		// If the first datasource is populated - drag through that - otherwise
		// drag through the second one
		// manualOverride=AUTOMATIC_CD,MANUAL_CD
		
		String setDataString = getAndRemoveInstruction("manualOverride");
		String alert = getAndRemoveInstruction("warningIfOverriden");
		int overrideCol = Integer
				.parseInt(getAndRemoveInstruction("overrideCol"));
		String setDataParts[] = setDataString.split(",");
		runPos = "1";
		// warning if insufficient instructions passed
		if (setDataParts.length < 2) {
			WarningMessage msgUserError = new WarningMessage(0, "DTP Error",
					WarningMessage.OUTPUT_TO_LOG_ONLY,
					"manualOverride instructions require two arguments. You provided: "
							+ setDataString);
			outputRecordSet.addWarningMessage(msgUserError);
			return;
		}
		RecordSet dragThroughRSorDS1 = this.getInputRecordSet(setDataParts[0],
				true);
		if (dragThroughRSorDS1 != null) {
			if (dragThroughRSorDS1.getRowCount() != 0) {
				if (trace) {
					outputRecordSet.addWarningMessage("Dragging through "
							+ setDataParts[0] + " ("
							+ dragThroughRSorDS1.getRowCount() + " rows)");
				}

				if (alert.equalsIgnoreCase("y")) {
					WarningMessage msg = new WarningMessage(0, "Rendering",
							WarningMessage.DONT_OUTPUT,
							"The default data has been overridden");
					outputRecordSet.addWarningMessage(msg);
				}

				// writeOutputRecordset(dragThroughRSorDS1);
				RecordSet irs = dragThroughRSorDS1;
				int rowCount = irs.getRowCount();

				RecordSetMetaData rsmdIn = irs.getMetaData();
				RecordSetMetaData rsmdOut = outputRecordSet.getMetaData();
				String inColName;
				String outColName;

				if (irs.getColumnCount() == outputRecordSet.getColumnCount()) {
					runPos = "x1";
					// you may have the same number of input and output columns,
					// but if one of the input cols contains the word 'null'
					// then this turns the column into a null col and it isn't
					// returned, so need to match on column names
					// if your columns names aren't identical then this won't
					// work
					for (int inRow = 0; inRow < rowCount; inRow++) {
						if (outputRecordSet.getColumnCount() == irs.getRow(
								inRow).getColumnCount()) {
							outputRecordSet.addRow(irs.getRow(inRow));
						} else {
							int outRow = outputRecordSet.addRow();
							for (int outCol = 0; outCol < outputRecordSet
									.getColumnCount(); outCol++) {
								outColName = rsmdOut.getField(outCol)
										.getFieldName();

								for (int inCol = 0; inCol < irs
										.getColumnCount(); inCol++) {
									inColName = rsmdIn.getField(inCol)
											.getFieldName();

									if (inColName.equals(outColName)) {
										outputRecordSet.setValue(outRow,
												outCol,
												irs.getString(inRow, inCol));
									}
								}
							}
						}
					}
					return;
				}

				// JH 03/07/14: originally it was just one line of code:
				// writeOutputRecordSet(dragThroughRSorDS1);

				// if they are different then check if the output RS has more or
				// less cols than the input
				int cols = 0;

				if (overrideCol < 0) {
					// outputRecordSet.addWarningMessage("Hereee???");

					if (irs.getColumnCount() < outputRecordSet.getColumnCount()) {
						cols = irs.getColumnCount();
					} else {
						cols = outputRecordSet.getColumnCount();
					}

					for (int rec = 0; rec < rowCount; rec++) {
						int newRow = outputRecordSet.addRow();
						for (int c = 0; c < cols; c++) {
							outputRecordSet.setValue(newRow, c, irs.getRow(rec)
									.getString(c));
						}
					}

				}
				if (overrideCol > 0) {
					// outputRecordSet.addWarningMessage("or here???");

					if (irs.getColumnCount() < outputRecordSet.getColumnCount()) {
						cols = irs.getColumnCount();
					} else {
						cols = outputRecordSet.getColumnCount();
					}

					for (int rec = 0; rec < rowCount; rec++) {
						int newRow = outputRecordSet.addRow();
						// outputRecordSet.addWarningMessage("Do we ever reach here??");
						for (int c = 0; c < cols; c++) {

							outputRecordSet.setValue(newRow, overrideCol, irs
									.getRow(rec).getString(c));
						}
					}

				}

			}

			else {
				RecordSet dragThroughRSorDS2 = this.getInputRecordSet(
						setDataParts[1], true);
				if (trace) {
					outputRecordSet.addWarningMessage("Dragging through "
							+ setDataParts[1] + " ("
							+ dragThroughRSorDS2.getRowCount() + " rows)");
				}
				if (!(dragThroughRSorDS2 == null || dragThroughRSorDS2
						.getRowCount() == 0)) {
					runPos = "2";
					writeOutputRecordset(dragThroughRSorDS2);
					runPos = "3";
				}
			}
		}

		return;
	}
	
	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
		protected void processJoinWhere() throws VermilionNonFatalException,VermilionFatalException {
			
			// join is a light weight version of an SQL WHERE clause, using AND or OR logic, along with =, !=, <>, > and < operators
			// join will find links between records of the input data/recordsets, and will copy fields from the input to the output recordsets if they
			// have the same system column name and data type.
			// join may work on two or more input recordsets
			//
			// E.g.
			//     joinWhere=DS001.portcode=DS002.portcode
			//   or
			//     joinWhere=DS001.portcode=DS002.portcode AND DS002.sector='TOTAL'
			//   or
			//     joinWhere=DS001.portcode=DS002.portcode AND sector='TOTAL'
			//   or
			//     joinWhere=DS001.portcode=DS002.portcode AND (sector='TOTAL' or sector='CASH')
			//   or
			//     joinWhere=DS001.portcode=DS002.portcode AND total_effect<0
			//   or
			//     joinWhere=                 <=== will cartesian product all combinations of the input data/recordsets
			//
			// Note, as VRS cannot do it, it is up to the implementation to create an outputrecordset with the required columns names and types
			// (It has been requested that VRS custom processors should do this - so then we could introduce another instruction joinResults=portcode, sector, total_effect
			//  which would take the input recordsets definitions of these columns and then build an outputrecordset to matcb).
			
			boolean ltrace=false;
			// "trace" is set to true when quietMode is disabled
			//logManager.debug("\n\n\n\n\n\n\n===========processJoin========================");
			if (ltrace) {
				WarningMessage msg = new WarningMessage(0, "Rendering",
						WarningMessage.DONT_OUTPUT, "processJoin starting");
				outputRecordSet.addWarningMessage(msg);
			}
			runPos = "(2543) processJoin starting";
			try {
				// joinWhere=PrimaryDS/RS, the rest.
				// joinWhere will apply a Cartesian product to the results, unless a key is given via joinKeyColumns
				// joinKeyColumns=a or =a,b or a=b,c=d, or a<>b
				// the results will be dictated by the resulting column names - name matching is used to fill this out; where duplicates are detected then 
				// fill out policy will be on the first come first served
				// If duplicate rows are encountered - then if joinDistinct=Y then they are supressed, else they are allowed
		
				if ( inputRecordSetList.getRecordSetCount() <2 ) {
					outputRecordSet.addWarningMessage("Must provide at least 2 recordsets to join on");
					return;
				}
				
				runPos="(2557) getting instructions";
				String joinRSDS = ""; // getAndRemoveInstruction("join");
				//String joinType = getAndRemoveInstruction("joinType"); // Inner (default) or Outer
				String joinWhere = getAndRemoveInstruction("joinWhere").trim(); // If no keycolumns are given then this is an outer join
				//String joinDistinct = getAndRemoveInstruction("joinDistinct");
				
				String leftOuterJoinDS="";
				String loj="LEFT OUTER JOIN ";
				//System.out.println("Trace5892: ["+joinWhere+"]");
				if ( joinWhere.startsWith(loj) ) {
					String joinWhere0=joinWhere;
					joinWhere=joinWhere.substring( loj.length());
					//System.out.println("Trace5896: ["+joinWhere+"]");
					leftOuterJoinDS=joinWhere.substring(0, joinWhere.indexOf(" "));
					//System.out.println("Trace5898: ["+leftOuterJoinDS+"]");
					joinWhere=joinWhere.substring( leftOuterJoinDS.length() );
					//System.out.println("Trace5900: ["+joinWhere+"]");
					if ( leftOuterJoinDS.indexOf("(")>0 ) {
						throw new VermilionNonFatalException("Must put a space between LEFT OUTER JOIN TABLE and the ( in "+joinWhere0,1);
					}
				}
				
				String resolveSubVarOutputLang = getAndRemoveInstruction("resolveSubVarOutputLang");
				joinWhere = ComponentUtility.parseText(outputRecordSet.getTid(),
						StringHelper.getResolveSubVars(joinWhere,
								parameterList, true), resolveSubVarOutputLang);

				WarningMessage msg = new WarningMessage(0, "Rendering",
						WarningMessage.DONT_OUTPUT, "processJoin starting : "+joinWhere);
				outputRecordSet.addWarningMessage(msg);

				if ( joinWhere.equals("") ) {
					for (int r = 0; r < inputRecordSetList.getRecordSetCount(); r++) {
						if ( inputRecordSetList.getDataSourceCode(r) != null ) joinRSDS=joinRSDS+","+inputRecordSetList.getDataSourceCode(r);
						if ( inputRecordSetList.getRecordSetCode(r) != null ) joinRSDS=joinRSDS +","+inputRecordSetList.getRecordSetCode(r);
					}
					if ( joinRSDS.startsWith(",") ) joinRSDS=joinRSDS.substring(1);
					if ( trace ) outputRecordSet.addWarningMessage("joinRSDS=["+joinRSDS+"]");
				}
				else {
					String seeing="";
					for (int r = 0; r < inputRecordSetList.getRecordSetCount(); r++) {
						if ( inputRecordSetList.getDataSourceCode(r) != null ) seeing=inputRecordSetList.getDataSourceCode(r);
						if ( inputRecordSetList.getRecordSetCode(r) != null ) seeing=inputRecordSetList.getRecordSetCode(r);
						if ( joinWhere.indexOf(seeing)>-1 ) {
							joinRSDS=joinRSDS+","+seeing;
						}
					}
					if ( joinRSDS.startsWith(",") ) joinRSDS=joinRSDS.substring(1);
					if ( trace ) outputRecordSet.addWarningMessage("joinRSDS=["+joinRSDS+"]");
				}
				
				runPos="(2564) getting input recorset/datasource codes";
				String dsCodes="";
				String rsCodes="";
				for (int r = 0; r < inputRecordSetList.getRecordSetCount(); r++) {
					if ( inputRecordSetList.getDataSourceCode(r) != null ) dsCodes=dsCodes+", "+inputRecordSetList.getDataSourceCode(r);
					if ( inputRecordSetList.getRecordSetCode(r) != null ) rsCodes=rsCodes +", "+inputRecordSetList.getRecordSetCode(r);
				}
				if ( ! dsCodes.equals("") ) dsCodes=dsCodes.substring(2);
				if ( ! rsCodes.equals("") ) rsCodes=rsCodes.substring(2);
				
				runPos="(2574) validating given join codes";
				int[] recordsetIndex = new int[10]; // max 10 (for now real reason)
//				recordsetIndex[0]=1;
				String[] codes = joinRSDS.split(","); // TODO : autocompute this
				runPos="(2578) validating given join codes "+codes.length;
				
				for (int c = 0; c < codes.length; c++) {
					boolean seen=false;
					for (int r = 0; r < inputRecordSetList.getRecordSetCount(); r++) {
						runPos="(2588) validating given join codes "+c+"/"+r+" of "+codes.length;
						if (inputRecordSetList.getDataSourceCode(r) != null && inputRecordSetList.getDataSourceCode(r).equals(codes[c])) {
							//logManager.debug("***** Storing "+codes[c]+" as recordsetIndex["+c+"]="+r);
							recordsetIndex[c]=r; 
							// this allows me to map from codes[*] to inputRecordSetList entries 
							// e.g. inputRecordSetList.getRecordSet(recordSetIndex[codes[c]])
							seen=true;
						}
						if (inputRecordSetList.getRecordSetCode(r) != null && inputRecordSetList.getRecordSetCode(r).equals(codes[c])) {
							//logManager.debug("***** Storing "+codes[c]+" as recordsetIndex["+c+"]="+r);
							recordsetIndex[c]=r; 
							// this allows me to map from codes[*] to inputRecordSetList entries 
							// e.g. inputRecordSetList.getRecordSet(recordSetIndex[codes[c]])
							seen=true;
						}
					}
					runPos="(2603) validating given join codes "+codes.length;
					if ( ! seen ) {
						runPos="(2605) validating given join codes "+codes.length;
						String errMsg="Error : Cannot see DS code ("+c+") from "+joinWhere+", ('"+codes[c]+"') (only saw ";
						if ( dsCodes!="" ) {
							errMsg=errMsg+dsCodes;
							if ( rsCodes!="" ) {
								errMsg=errMsg+", "+rsCodes;
							}
						}
						else {
							if ( rsCodes!="" ) {
								errMsg=errMsg+rsCodes;
							}
						}
						errMsg=errMsg+")";
						outputRecordSet.addWarningMessage(errMsg);
						return;
					}
				}

				// recordsetIndex[x] = the recordset for the codes[c]
				if ( 1 == 2 ) {
					for(int c = 0; c < codes.length; c++) {
						System.out.println("datasource codes["+c+"]="+codes[c]+" = inputRecordset("+recordsetIndex[c]+")= RS Code "+inputRecordSetList.getDataSourceCode(recordsetIndex[c]));
					}
				}	
				// Run through the primary recordset
				// if keycolumn is given then find matching rows and merge them in
				// if no keycolumn is given then get all rows and merge them in
				
				// Loop through all the output recordset rows, 
				// if keycolumn is given then find matching rows and merge them in
				// if no keycolumn is given then get all rows and merge them in
				
				// Run through each record of the primary
				// if where is given then 
				//   if all clauses evaluation as true then merge in all records
				//   run through each clause, and bring in each related table if evaluates to true
				int x=codes.length;
				runPos="(dtpi2636) joining matches...";
				
				RecordSet addRecordSet=getJoinMatches(
								outputRecordSet, joinWhere+" ", inputRecordSetList, codes, recordsetIndex, leftOuterJoinDS);

				if ( ltrace ) outputRecordSet.addWarningMessage("Finished1");
				
			}
			catch(VermilionNonFatalException e) {
				logManager.error("VNFE2646:"+e.getMessage());
				throw new VermilionNonFatalException("VNFE:"+e.getMessage(),1);
			}
			catch (VermilionFatalException e) {
				logManager.error("VFE2650:"+e.getMessage());
				throw new VermilionFatalException("processJoin VFE exception: "+e.toString()+" ("+runPos+")",2628);
			}
			catch (Exception e) {
				logManager.error("E2654:"+e.getMessage());
				throw new VermilionFatalException("processJoin exception: "+e.toString()+" ("+runPos+")",2628);
			}
			if ( ltrace ) outputRecordSet.addWarningMessage("Finished2");
		}
		
		// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
		public RecordSet getJoinMatches(RecordSet outputRecordSet, String joinWhere, RecordSetList inputRecordSetList, String[] codes, int[] recordsetIndex, String leftOuterJoinDS)
			throws VermilionNonFatalException,VermilionFatalException {
			boolean ltrace=false;
			
			String runPos="dtpi2656";
			if ( ltrace ) outputRecordSet.addWarningMessage(runPos);
			try {
				logManager.debug("(2705) getJoinMatches");

				int x, rs, asize2, newRow, rowNum, maxLeftOuterRows=0, leftOuterRS=0;
				int asize         = codes.length;
				int[] rowindex    = new int[asize];
				int[] maxrowindex = new int[asize];
				for (int c = 0; c < asize; c++) { 
					rs              = recordsetIndex[c];
					asize2          = inputRecordSetList.getRecordSet(rs).getRowCount();
					rowindex[rs]    = 0; 
					maxrowindex[rs] = asize2-1;
					if ( codes[c].equals(leftOuterJoinDS))  {
						//System.out.println("Trace6050: left outer join rows = "+asize2);
						maxLeftOuterRows=asize2;
						leftOuterRS=rs;
					}
				}
				boolean[] leftOuterRowUsed=new boolean[maxLeftOuterRows];
				//System.out.println("Trace6056: left outer join rows for "+leftOuterJoinDS+" is = "+maxLeftOuterRows);

				
				runPos="dtpi2736";
				logManager.debug("(2717) getJoinMatches");
				String[] orsColumns  = new String[outputRecordSet.getColumnCount()];
				RecordSetMetaData md = outputRecordSet.getMetaData();
				for (int i = 0; i < outputRecordSet.getColumnCount(); i++) {
					orsColumns[i] = md.getField(i).getFieldName() + "/" + md.getField(i).getFieldType();
				}


				runPos="dtpi2604";
				// save the join recordset columns for quick lookup...
				// dstColumns is an array of column names for each codes table
				String[][] dstColumns = new String[codes.length][];
				for (int c = 0; c < codes.length; c++) {
					rs = recordsetIndex[c];
					md = inputRecordSetList.getRecordSet(rs).getMetaData();
					x  = inputRecordSetList.getRecordSet(rs).getColumnCount();
					dstColumns[c] = new String[x];
					for (int i = 0; i < x; i++) {
						dstColumns[c][i] = md.getField(i).getFieldName() + "/" + md.getField(i).getFieldType();
					}
				}


				runPos="dtpi2619";
				// turn "(A.a=B.a) and (A.b=B.b)" into "1 and 1" into "1"
				// or   "((A.a=B.a or A.b=B.b) or (A.c=B.c))" into "(1 or 0) or (1)"
				// and evaluate each bracket to consider each row combination
				// This means that we cannot just loop rowsA and loop rowsB inside that,
				// but we must find every combination of row and decide
				// (execution plans are not going to be optimal - this isn't SQL)
				// loop through every row using rowindex as the multi dimensional counter...

				if ( trace ) outputRecordSet.addWarningMessage("JoinWhere="+joinWhere);		
				
				int safety=9999;
				while (getIsNotLastCombination(rowindex, maxrowindex) && safety-->0) {

					runPos="dtpi2660";
					String sTrace="rows ";
					for(int c=0;c<codes.length;c++) {
						sTrace+=codes[c]+"=row "+rowindex[recordsetIndex[c]]+", ";
					}

					runPos="dtpi2639";
					if ( ltrace ) logManager.error("\n\nRecord "+sTrace);
					if ( ltrace ) logManager.error("getJoinMatches2776: check clause for "+sTrace+" ("+safety+")");
					
					// getJoinWhereClauseResult returns true or false if the row combination (rowindex*) matches the where clause
					// * this is all the input recordsets, 1, 2, 3, or more, and a pointer for each; the joinWhere logic is basically looping
					//   through each rowindex to see if the where clause is true
					//
					if ( getJoinWhereClauseResult(joinWhere, inputRecordSetList, codes, recordsetIndex, rowindex, dstColumns)) {

						//System.out.println("Trace6113: leftOuterRowused["+rowindex[leftOuterRS]+"]=true");
						if ( ! leftOuterJoinDS.equals("") ) {
							leftOuterRowUsed[rowindex[leftOuterRS]]=true;
						}
						runPos="dtpi2643";
						newRow = outputRecordSet.addRow();
						for (int c = 0; c < codes.length; c++) {
							runPos="dtpi2644."+c;
							rs                = recordsetIndex[c];
							rowNum            = rowindex[rs];
							RecordRow joinrow = inputRecordSetList.getRecordSet(rs).getRow(rowNum);
							outputRecordSet   = getJoinMergeRows(outputRecordSet, joinrow, newRow, orsColumns, dstColumns[c]);
						}
					}
					runPos="dtpi2651";
					rowindex = getNextRowIndexCombination(rowindex, maxrowindex); // increment the row counters...
					if ( ltrace ) outputRecordSet.addWarningMessage("");
				}
				
				// -------------------- LEFT OUTER JOIN ---------------------------
				runPos="outerjoin6110";
				//System.out.println("Trace6132: checkpoint");
//				if ( leftOuterJoin ) {
				if ( ! leftOuterJoinDS.equals("") ) {
					int leftRSindex=0;
					for (int c=0; c< codes.length; c++) {
						if ( codes[c].equals(leftOuterJoinDS) ) {
							leftRSindex=recordsetIndex[c];
						}
					}
					RecordSet leftRS=inputRecordSetList.getRecordSet(leftRSindex);
					int rowCount=leftRS.getRowCount();
	
					//System.out.println("Trace6143: checkpoint");
					boolean joined=true;
					for(int r=0; r<rowCount; r++) {
	
						//System.out.println("Trace6147: checkpoint ("+r+")");
						if ( !leftOuterRowUsed[r] ) {
							newRow = outputRecordSet.addRow();
							outputRecordSet=getJoinMergeRows(outputRecordSet, leftRS.getRow(r), newRow, orsColumns, dstColumns[leftRSindex]);
						}
					}
				}
				// -------------------- LEFT OUTER JOIN ---------------------------

				//System.out.println("Trace6155: checkpoint");

				if ( safety==0 ) {
					logManager.error("getJoinMatches2734: safetycatch triggered");
					throw new VermilionFatalException("Gave up looping in getJoinMatches",1);
				}
				
				runPos="dtpi2782";
			}
			catch(VermilionNonFatalException e) {
				String msg="getJoinMatches VNFE Exception caught "+e.getMessage()+"; (at "+runPos+")";
				logManager.error(msg);
				throw new VermilionNonFatalException(e.getMessage(),1);
			}
			catch(Exception e) {
				String msg="getJoinMatches2671 Exception caught "+e.toString()+"; (at "+runPos+")";
				logManager.error(msg);
				throw new VermilionFatalException(msg,1);
			}
			
			return outputRecordSet;
		}
		
		// --------------------------------------------------------------------------------------------------------------
		private int[] getNextRowIndexCombination(int[] rowindex, int[] maxrowindex) 
			throws VermilionNonFatalException,VermilionFatalException {
			boolean ltrace=false;
			// multidimensional loop through rowindex

			if ( ltrace ) {
				for(int c=0; c<rowindex.length; c++) {
					logManager.error("getNextRowIndexCombination2754: ==> ("+c+")"+rowindex[c]);
				}
			}
			
			boolean keepLooping=true;
			for(int rs=0; rs<rowindex.length; rs++) {
				rowindex[rs]=rowindex[rs]+1;
				if ( rowindex[rs]>maxrowindex[rs]) {
					logManager.debug("getNextRowIndexCombination2812x: hit limit on recordset["+rs+"], moving on to next rs and incrementing that");
					if ( rs!=rowindex.length-1 ) {
						rowindex[rs]=0;
					}
				}
				else {
					break;
				}
			}

			if ( ltrace ) {
				for(int c=0; c<rowindex.length; c++) {
					logManager.error("getNextRowIndexCombination2747: ==> ("+c+" = "+rowindex[c]+" of "+maxrowindex[c]+")");
				}
			}
			return rowindex;
		}
		
		// --------------------------------------------------------------------------------------------------------------
		private boolean getIsNotLastCombination(int[] rowindex, int[] maxrowindex)
			throws VermilionNonFatalException,VermilionFatalException {
			boolean ltrace=false;
			boolean isnotlast=true;
			for(int rs=0; rs<rowindex.length; rs++) {
				if ( rowindex[rs]>maxrowindex[rs]) {
					isnotlast=false;
					break;
				}
				logManager.debug("Checking isLastCombination ("+rs+") = "+rowindex[rs]+" (max="+maxrowindex[rs]+"="+isnotlast+")");
			}

			logManager.debug("Checking isLastCombination="+isnotlast);
			return isnotlast;
		}
		
		// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
		public boolean getJoinWhereClauseResult(
				String joinWhere,
				RecordSetList inputRecordSetList,
				String[] codes,  // this is the codes of the recordsets in order (0=primary)
				int[] recordsetIndex, // this is the index of the recordsets in "codes" order
				int[] rowindex, // this is the rowNumber for each recordset
				String[][] dstColumns) // join recordset columns information
			throws VermilionFatalException, VermilionNonFatalException {

			//logManager.error("(dtpi2757) getJoinWhereClauseResult for "+joinWhere);
			
			String result=getJoinWhereClauseResultParts(joinWhere, 1, inputRecordSetList, codes, recordsetIndex, rowindex, dstColumns);
			
			//logManager.error("(dtpi2814) getJoinWhereClauseResult result="+result);
			//outputRecordSet.addWarningMessage("===============================getJoinWhereClauseResult["+joinWhere+"]=["+result+"]");
			if ( result.trim().equals("1") || result.trim().equals("") ) {
				return true;
			}
			else {
				return false;
			}
		}
		
		// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
		public String getJoinWhereClauseResultParts(String joinWhere, int nesting,
							RecordSetList inputRecordSetList, String[] codes, int[] recordsetIndex, int[] rowindex, String[][] dstColumns)
			throws VermilionFatalException, VermilionNonFatalException {

			String nestingIndent    = "==============".substring(0,nesting*2);
			boolean ltrace=false;
			if ( nesting > 10 ) {
				logManager.error("getJoinWhereClauseParts nested too deep; giving up");
				throw new VermilionFatalException("getJoinWhereClauseParts nested too deep; giving up",1);
			}
			if ( ltrace ) logManager.error(nesting+":"+nestingIndent+":DTP-I(2831) getJoinWhereClauseParts called for ["+joinWhere+"]");
			if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":getJoinWhereClauseResultParts["+nesting+"]["+joinWhere+"]");
			joinWhere=joinWhere.trim();
			
			// step through each character in the jobWhere clause,
			// looking out for brackets and quotes
			// if it's a bracket, then nest it
			// if it's a quote then wait for the close quote
			// when returning, substitute the values and evaluate the test
			// return an array of results separated by ORs or ANDs

			String sResult = "";
			String runPos=nestingIndent+":dtpi2840";
			try { 
				ArrayList<String> parts = new ArrayList();
				String ch,part;
				String tracespaces="                                                                                                                 ";
				boolean quotes =false;
				boolean squotes=false;
				int bracketdepthmax=0;
				int bracketdepth=0;
				int end=joinWhere.length();
				int c=0; 
				int c0=c;
				int safety=1000;
				int bracketsDepth=0;
				int findDepth=0;
				boolean brackets=false;
				if ( joinWhere.substring(0,0).equals("(")) {
					logManager.debug(nestingIndent+":Brackets discovered - step forward 1");
					c++;
					c0++;
				}
				while ( c < joinWhere.length() && safety-->0 ) {
					runPos=nestingIndent+":dtpi2859("+c0+" to "+c+" of "+joinWhere.length()+"):from "+joinWhere;
					//logManager.error(runPos);
					ch=joinWhere.substring(c,c+1);
					//logManager.error(runPos+" [char="+ch+"]");
					runPos=nestingIndent+":dtpi2858("+c+")";
					if ( ch.equals("'") ) {
						squotes=!squotes;
					}
					if ( ch.equals("\"") ) {
						quotes=!quotes;
					}

					//logManager.error(nestingIndent+":Looking at char("+c+")="+ch+" (from "+joinWhere+"; nesting="+nesting+")");
					if ( ch.equals(")") ) { // Find brackets
						if ( !squotes && !quotes ) {
							bracketsDepth--;
							if ( bracketsDepth==findDepth) {
								runPos=nestingIndent+":dtpi2873(a="+c0+"/b="+c+"/c="+(c0+1)+"/d="+c+"-"+c0+"+1="+(c-c0+1)+")";
								if( ltrace ) logManager.error(nestingIndent+":getJoinWhereClauseResultParts2735) getJoinWhereClauseParts jumping to "+joinWhere.substring(c+1));
								if( ltrace ) logManager.error(nestingIndent+":getJoinWhereClauseResultParts2736) joinWhere.substring("+(c0+1)+","+(c-c0+1)+")"); 
								//part=joinWhere.substring(c0+1,c-c0+1).trim();
								part=joinWhere.substring(c0+1,c).trim();
								if( ltrace ) logManager.error(nestingIndent+":getJoinWhereClauseResultParts2737) joinWhere.substring("+(c0+1)+","+(c-c0+1)+")="+part); 
								
								// from c0=37 to c=68
								
								runPos=nestingIndent+":dtpi2808("+c0+"/"+c+")";
								if( ltrace ) logManager.error(nestingIndent+":DTP-I(2860."+nesting+") reevaluating "+part+" from "+c);
								
								runPos=nestingIndent+":dtpi2865("+c+")";
								if( ltrace ) logManager.error(nestingIndent+":DTP-I(2875."+" ("+c0+"):"+part+"--------------------------------------------");
								
								String ris="";
								for(int ri=0; ri<rowindex.length; ri++) {
									ris=ris+"["+rowindex[ri]+"]";
								}
								
								if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":ResultParts["+nesting+"]:"+ris+" : starting recursive call ["+part+"]");
								
								runPos="2825";
								String p1=" "+getJoinWhereClauseResultParts(part, nesting+1, inputRecordSetList, codes, recordsetIndex, rowindex, dstColumns)+" ";
								if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":ResultParts["+nesting+"]:"+ris+" : returned from recursive call with ["+p1+"]");
								
								runPos="2829";
								if ( c0>0 ) {
									if ( ltrace ) {
										outputRecordSet.addWarningMessage(nestingIndent+":ResultParts["+nesting+"]:"+ris+" : returned from recursive call with ["+p1+"]==["
											+joinWhere.substring(0,c0-1)
											+"+"+p1
											+"+"+joinWhere.substring(c+1)+"]");
									}
									joinWhere=joinWhere.substring(0,c0-1)+p1+joinWhere.substring(c+1);
								}
								else {
									if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":ResultParts["+nesting+"]:"+ris+" : returned from recursive call with ["+p1+"]==["
											+"+"+p1
											+"+"+joinWhere.substring(c+1)+"]");
									joinWhere=p1+joinWhere.substring(c+1);
								}
								if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":ResultParts["+nesting+"]:"+ris+" : joinWhere is now ["+joinWhere+"]");
								
								runPos=nestingIndent+":dtpi2867("+c+")";
								logManager.debug(nestingIndent+":DTP-I(2863) returned ["+part+"]=="+p1+" .... giving "+joinWhere);

								c=c-part.length()+p1.length();
								c=0;
								
//								if ( ltrace || true ) {
//									logManager.error(nestingIndent+":"+joinWhere);
//									logManager.error(nestingIndent+":"+tracespaces.substring(0,c-1)+"^");
//								}
								
								logManager.debug(nestingIndent+":DTP-I2892: c back a bit from "+c);
								c=c;
								c0=c+1;
								bracketsDepth--;
							}
						}
					}
					if ( ch.equals("(") ) {
						if ( !squotes && !quotes) {
							brackets=true;
							findDepth=bracketsDepth;
							bracketsDepth++;
							c0=c;
							if ( ltrace ) logManager.error(nestingIndent+":DTP-I(2741) getJoinWhereClauseParts detected start-brackets at "+c+" giving: "+joinWhere.substring(0,c)+";");
						}
					}
					c++;
				} // while ( c < joinWhere.length() && safety-->0 ) {
				
				runPos=nestingIndent+":dtpi2885("+c0+" of "+joinWhere.length()+")("+brackets+"): "+joinWhere+"::";
				if ( !brackets ) {
					if ( c0 < joinWhere.length() ) {
						if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":(2936)Taking joinWhere("+joinWhere+") from "+c0);
						part=joinWhere.substring(c0).trim();
						if ( !part.equals(")") && !part.equals("") ) {
							logManager.debug(nestingIndent+":DTP-I(2915) finishing with \""+part.trim()+"\" (c0="+c0+")");
							runPos=nestingIndent+":dtpi2923("+c0+")";
							if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":getJoinWhereClauseResultParts["+nesting+"] calling getJoinEvaluatePart for ["+part+"]");
							
							String p1=getJoinEvaluatePart(nesting, part, inputRecordSetList, codes, recordsetIndex, rowindex, dstColumns,true,true);
							
							if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":getJoinWhereClauseResultParts["+nesting+"] called getJoinEvaluatePart returned ["+p1+"]");
							sResult=p1;
							runPos=nestingIndent+":dtpi2926("+c+")";
							//logManager.error("DTP-I(2896) returned ["+part+"]=="+p1+" .... giving "+sResult);
						}
					}
					else {
						sResult="";
					}
				}
				else {
					if ( c0 < joinWhere.length() ) {
						String joinWhere0=joinWhere;
						sResult=joinWhere.substring(c0-1);
						if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":(2955b)Taking joinWhere["+joinWhere0+"] from "+c0+"; sResult=["+sResult+"]");
					}
					else {
						sResult=joinWhere; 
					}
				}
				
				runPos=nestingIndent+":dtpi2894";
				// Evaluate the result part....
				// TODO do I need this evaluate....
				sResult=sResult.trim();
				logManager.debug("\n\n"+nestingIndent+":DTP-I(2909)["+nesting+"]RESULT="+sResult+"  <====================================================");
				if ( !sResult.equals("") ) {
					if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":(2969)getJoinWhereClauseResultParts["+nesting+"] calling getJoinEvaluatePart again for ["+sResult+"]");
					sResult=getJoinEvaluatePart(nesting, sResult, inputRecordSetList, codes, recordsetIndex, rowindex, dstColumns, false, true).trim();
					if ( ltrace ) outputRecordSet.addWarningMessage("getJoinWhereClauseResultParts["+nesting+"] calling getJoinEvaluatePart again returned ["+sResult+"]");
				}
				logManager.debug(nestingIndent+":DTP-I(2901): ["+nesting+"]RESULT="+sResult);
			}
			catch (VermilionNonFatalException e) {
				throw new VermilionNonFatalException(e.getMessage(),1);
			}
			catch (Exception e) {
				logManager.error(nestingIndent+":(2924)Unexpected error "+e.getMessage()+" at "+runPos);
				throw new VermilionFatalException(nestingIndent+":Unexpected error "+e.getMessage()+" in "+runPos,1);
			}
			
			if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":getJoinWhereClauseResultParts["+nesting+"] finally returned ["+sResult+"] from ["+joinWhere+"]");
			return " "+sResult+" ";
		}
		
		// --------------------------------------------------------------------------------------------------------------
		public String getJoinWhereFieldName(String fieldName, String[] codes, String[][] dstColumns) 
				throws VermilionFatalException, VermilionNonFatalException {
			// fieldName may be a field or a value
			// fields may be a table.column or just column
			String fieldName0=fieldName;
			//logManager.error("************ getJoinWhereFieldName2932:"+fieldName);
			if (isNumeric(fieldName) ) {
				return fieldName;
			}
			
			if ( fieldName.startsWith("'") || fieldName.trim().equals("")) {
				return fieldName;
			}
			
			if (fieldName.endsWith("]") ) {
				fieldName=fieldName.substring(0, fieldName.length()-1);
				fieldName=fieldName.replace("[","");
			}
			
			if ( fieldName.indexOf(".")==-1 ) {
				int seen=0;
				String seenOptions="";
				String fieldName1="";
				String fieldNameX="";
				for(int c1=0; c1<codes.length; c1++) {
					for(int c2=0; c2< dstColumns[c1].length; c2++) {
						//logManager.error("Comparing ["+fieldName+"] to ["+dstColumns[c1][c2]+"]");
						fieldNameX=dstColumns[c1][c2].substring(0, dstColumns[c1][c2].indexOf("/"));
						//logManager.error("Comparing ["+fieldName+"] to ["+dstColumns[c1][c2]+"]["+fieldNameX+"]");
						if ( fieldName.equals(fieldNameX) ) {
							if ( seen==0 ) {
								seenOptions=codes[c1]+".["+fieldName+"]";
							}
							else {
								seenOptions=seenOptions+" or "+codes[c1]+".["+fieldName+"]";
							}
							seen++;
							fieldName1=codes[c1]+"."+fieldName;
						}
					}
				}
				if ( seen>1 ) {
					throw new VermilionNonFatalException("Ambiguous column "+fieldName+"; could be "+seenOptions,1);
				}
				if ( !fieldName1.equals("") ) {
					fieldName=fieldName1;
					if ( fieldName.indexOf(".")==-1 ) {
						outputRecordSet.addWarningMessage("getJoinWhereFieldName("+fieldName+") has no table");
					}
				}
				//throw VermilionNonFatalException("Do not rec")
//			}
//			else {
//				outputRecordSet.addWarningMessage("getJoinWhereFieldName("+fieldName+") has no table - assuming it's coming from [0]="+codes[0]);
//				fieldName=codes[0]+"."+fieldName;
			}
			if ( fieldName.indexOf(".")==-1 ) {
				fieldName=codes[0]+"."+fieldName;
			}
			//outputRecordSet.addWarningMessage("getJoinWhereFieldName="+fieldName0+"="+fieldName+" ("+codes[0]+")");
			//logManager.error("getJoinWhereFieldname2979 ="+fieldName);
			return fieldName;
		}
		
		// --------------------------------------------------------------------------------------------------------------
		public String getJoinEvaluatePart(int nesting, String part, RecordSetList inputRecordSetList, String[] codes, int[] recordsetIndex, 
				int[] rowindex, String[][] dstColumns, boolean showTrace, boolean recurse)
				throws VermilionFatalException, VermilionNonFatalException {
			// a part can be one or many same-level tests, e.g. "a=b and c=d", or "a<>b or c<>d".
			boolean ltrace			= false;
			boolean ltrace2			= false;
			String sResult			= "";
			String sResultTrace		= "";
			String nestingIndent    = "==============".substring(0,nesting*2);
			
			part=part.replace("&nbsp;"," ").trim();
			if ( part.equals("0") || part.equals("1") ) {
					return part;
			}
			//outputRecordSet.addWarningMessage("Starting getJoinEvaluatePart["+part+"]");
			String runPos=nestingIndent+":dtpi2960";
			try {
				 /*
				  * regex patterns reminder
				  * abc  Letters
				  * 123  Digits
				  * \d  Any Digit
				  * \D  Any Non-digit character
				  * .  Any Character
				  * \.  Period
				  * [abc]  Only a, b, or c
				  * [^abc]  Not a, b, nor c
				  * [a-z]  Characters a to z
				  * [0-9]  Numbers 0 to 9
				  * \w  Any Alphanumeric character
				  * \W  Any Non-alphanumeric character
				  * {m}  m Repetitions
				  * {m,n}  m to n Repetitions
				  * *  Zero or more repetitions
				  * +  One or more repetitions
				  * ?  Optional character
				  * \s  Any Whitespace
				  * \S  Any Non-whitespace character
				  * ^$  Starts and ends
				  * ()  Capture Group
				  * (a(bc))  Capture Sub-group
				  * (.*)  Capture all
				  * (abc|def)  Matches abc or def  
				 */
				logManager.debug(nestingIndent+":============getJoinEvaluatePart ["+part+"]");
				String matchquotedtext     = "((\"[^\"]*\"|'[^']*'))";
				String matchtest           = "(<>)|(!=)|(>)|(<)|(=)"; // group these pairs together
				String matchsquarebrackets = "((\\[[^\\[]*\\]))";
				String matchnospaces 	   = "[^\\s]";
				String matchalphanumerics  = "\\w";
				String matchnumbers        = "|(^\\d+\\.\\d+$)|(\\[.*?\\])";
				
				String sRegex = matchquotedtext + "" 
								+"|"+matchsquarebrackets
								+"|"+matchtest // group these pairs together
								+"|("+matchnospaces+matchalphanumerics+"*)"
								+"|"+matchnumbers // quoted blocks are ok - but the rest are individual characters
								;
				
				if ( !part.trim().equals("") ) {
					//part="("+part+")";
					
					logManager.debug(nestingIndent+":(3049)Considering part: ["+part.trim().toLowerCase().replaceAll(" +"," ")+"]");
					if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":(3049)Considering part: ["+part.trim().toLowerCase().replaceAll(" +"," ")+"]");
					if ( part.trim().toLowerCase().replaceAll(" +"," ").startsWith("0 and ")     // replaceAll(" +"," ") = replace multiple spaces with one
					||   part.trim().toLowerCase().replaceAll(" +"," ").startsWith("1 and ") ) {
						if ( ltrace ) outputRecordSet.addWarningMessage("Seen 1 and ....");
						int p=part.toLowerCase().indexOf(" and ")+5;
						if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":Reconsidering "+part.substring(0,p)+" ... "+part.substring(p));
						part=part.substring(0,p)+getJoinEvaluatePart(nesting, part.substring(p), inputRecordSetList, codes, recordsetIndex, rowindex, dstColumns, false, true);
					}
					if ( part.trim().toLowerCase().replaceAll(" +"," ").startsWith("0 or ")
					||   part.trim().toLowerCase().replaceAll(" +"," ").startsWith("1 or ") ) {
						int p=part.toLowerCase().indexOf(" or ")+4;
						if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":Reconsidering "+part.substring(0,p)+" ... "+part.substring(p));
						part=part.substring(0,p)+getJoinEvaluatePart(nesting, part.substring(p), inputRecordSetList, codes, recordsetIndex, rowindex, dstColumns, false, true);
					}
					
					runPos=nestingIndent+":dtpi2980";
					Pattern regex = Pattern.compile(sRegex);
					Matcher regexMatcher = regex. matcher(part);
					ArrayList<String> results=new ArrayList<String>();
					String sParsed;
					results.add("");
					int x=1;
					while (regexMatcher.find()) {
						sParsed = regexMatcher.group();
						if ( ltrace ) outputRecordSet.addWarningMessage("Read "+sParsed);
						if ( sParsed.startsWith(".") ) {
							if ( ltrace ) outputRecordSet.addWarningMessage("Read dot; Storing="+results.get(x-1)+sParsed);
							results.add( results.get(x-1)+sParsed );
							results.set(x-1, "");
							x++;
						}
						else {
							if ( sParsed.startsWith("[") ) {
								if ( ltrace ) outputRecordSet.addWarningMessage("Read square bracket; storing="+results.get(x-1)+sParsed);
								results.add( results.get(x-1)+sParsed );
								results.set(x-1, "");
								x++;
							}
							else {
								if ( !sParsed.equals("") ) {
									if ( ltrace ) outputRecordSet.addWarningMessage("Storing "+sParsed);
									results.add(sParsed);
									x++;
								}
							}
						}
					}
					x--;
					
					runPos=nestingIndent+":dtpi2993";
					if ( ltrace ) logManager.error(runPos);
					for(int i=x; i>=0; i--) {
						if ( results.get(i).equals("") || results.get(i).equals("(") || results.get(i).equals(")") ) {
							results.remove(i);
						}
					}
					x=results.size()-1;
					
					String sTrace1="";
					if ( ltrace ) outputRecordSet.addWarningMessage("[6579]Processing "+x+" parts");
					for(int i=x; i>=0; i--) {
						sTrace1="["+i+":"+results.get(i)+"] ..... "+sTrace1;
					}
					if ( ltrace ) outputRecordSet.addWarningMessage("[6583]Processing "+sTrace1);
					
					String sTrace2="rows ";
					for(int c=0;c<rowindex.length;c++) {
						sTrace2+=codes[c]+"="+rowindex[recordsetIndex[c]]+", ";
					}
					
					runPos=nestingIndent+":dtpi3005 "+sTrace2;
					logManager.debug(runPos);
					boolean keepParsing=true;
					boolean prevTestRes,thisTestResult=false;
					boolean thisTest0Result;
					int i=0;
					int res=0;
					int safety=1000;
					String lhs, operator, rhs, andor="", seen="", prevlhsrhs="", prevTestResult="";
					sResult="";
					operator="";
					while (keepParsing && safety-- > 0) {
						runPos=nestingIndent+":dtpi3013";
						logManager.debug(runPos);
						if ( results.size()<=i) {
							throw new Exception("Was expecting another Test and Righthand side after \""+seen);
						}
						lhs=getJoinWhereFieldName(results.get(i++).trim(), codes, dstColumns);
						sResultTrace=sResultTrace+"[g:"+lhs+"]";
						if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":[6608]Read LHS:["+lhs+"]");
						seen+=" "+lhs;
						if ( ltrace ) logManager.error(nestingIndent+":DTP-I(2977):[6608]Read LHS=\""+lhs+"\"");
						
						runPos=nestingIndent+":dtpi3020";
						logManager.debug(runPos);
						if ( results.size()>i) {
							operator=results.get(i++).trim();
							sResultTrace=sResultTrace+"[g:"+operator+"]";
							if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":[6617]Read TST:["+operator+"]");
							seen+=" "+operator;
							if ( ltrace ) logManager.error(nestingIndent+":DTP-I(2983):[6617]Read test=\""+operator+"\"");
						}
						else {
							if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":[6622]Removed operator ["+operator+"] as "+results.size()+" > "+i);
							operator="";
						}
						
						if ( ( lhs.equals("0") || lhs.equals("1") ) 
							&& operator.equalsIgnoreCase("and") || operator.equalsIgnoreCase("or") ) {
							//outputRecordSet.addWarningMessage(nestingIndent+":[6629]Skip request as boolean logic detected on the LHS");
							// This is a previously resolved part of the clause - skip it and grab the next
							sResult+= lhs+" "+operator+" ";
							continue;
						}
						runPos=nestingIndent+":dtpi3027";
						logManager.debug(runPos);
						if ( results.size()>i) {
							rhs=getJoinWhereFieldName(results.get(i++).trim(), codes, dstColumns);
							sResultTrace=sResultTrace+"[g:"+rhs+"]";
							if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":[6635]Read RHS:["+rhs+"]");
							if ( ltrace ) logManager.error(nestingIndent+":(DTP-I)2992:[6635]Read RHS=\""+rhs+"\"");
							seen+=" "+rhs;
						}
						else {
							rhs="";
							seen="";
						}
						
						runPos=nestingIndent+":dtpi3018";
						logManager.debug(runPos);
						if ( ltrace ) logManager.error(nestingIndent+":getJoinEvaluatePart2997:Considering "+lhs+" + "+operator + " + "+rhs+ " (size="+results.size()+") "+sTrace2+" ...."+seen);
						if ( !lhs.equals("") ) {
							runPos=nestingIndent+":dtpi3019";
							String lhs0=lhs;
							lhs=getJoinResolveVariables(lhs, inputRecordSetList, codes, recordsetIndex, rowindex, dstColumns);
							logManager.debug(nestingIndent+":Done3240 ("+lhs0+"="+lhs+")");
							sResultTrace=sResultTrace+"[L:"+lhs0+"::"+lhs+"]";
						}
						if ( !rhs.equals("") ) {
							runPos=nestingIndent+":dtpi2996";
							String rhs0=rhs;
							rhs=getJoinResolveVariables(rhs, inputRecordSetList, codes, recordsetIndex, rowindex, dstColumns);
							logManager.debug(nestingIndent+":Done3247 ("+rhs0+"="+rhs+")");
							sResultTrace=sResultTrace+"["+operator+"][R:"+rhs0+"::"+rhs+"]";
						}
						if ( ltrace ) logManager.error(nestingIndent+":getJoinEvaluatePart3000:Considering "+lhs+" + "+operator + " + "+rhs+ " (size="+results.size()+") "+sTrace2);
						runPos=nestingIndent+":dtpi3023";
						
						//if ( lhs.equals("") || rhs.equals("") ) {
						if ( lhs.equals("") ) {
							String msgTx="Unexpected blank operand provided LHS:["+lhs+"] RHS:["+rhs+"]";
							logManager.error(msgTx);
							throw new VermilionNonFatalException(msgTx,1);
						}
						//rawResult+=" "+lhs+" "+test+" "+rhs;
						lhs=getFormattedDate(lhs);
						rhs=getFormattedDate(rhs);
						res=lhs.compareTo(rhs);
						//logManager.error(nestingIndent+":dtpi3010:Comparing ["+lhs+"]["+test+"]["+rhs+"]="+res);
						if ( operator.equals("and") || operator.equals("or") ) {
							if ( ! isNumeric(lhs) ) {
								String msgTx="Unexpected ("+operator.trim()+") test after \""+lhs+"\" (was expecting one of the following: =,!=,<>,<,> ) [ERR3249]";
								logManager.error(msgTx);
								throw new VermilionNonFatalException(msgTx,1);
							}
						}
						if ( operator.equals("=")) {
							if ( isNumeric(lhs) && isNumeric(rhs) ) {
								if ( new Float(lhs).floatValue() == new Float(rhs).floatValue() ) {
									sResult += " 1 ";
									thisTestResult=true;
								}
								else {
									sResult += " 0 ";
									thisTestResult=false;
								}
							}
							else {
								if ( res==0 ) {
									sResult += " 1 ";
									thisTestResult=true;
								}
								else {
									sResult += " 0 ";
									thisTestResult=false;
								}
							}
						}
						else {
							if ( operator.equals("!=") || operator.equals("<>") ) {
								if ( res!=0 ) {
									sResult += " 1 ";
									thisTestResult=true;
								}
								else {
									sResult += " 0 ";
									thisTestResult=false;
									
								}
							}
							else { 
								if ( operator.equals(">") ) {
									if ( isNumeric(lhs) && isNumeric(rhs) ) {
										if ( new Float(lhs).floatValue() > new Float(rhs).floatValue() ) {
											sResult += " 1 ";
											thisTestResult=true;
										}
										else {
											sResult += " 0 ";
											thisTestResult=false;
										}
									}
									else {
										if ( res>0 ) {
											sResult += " 1 ";
											thisTestResult=true;
										}
										else {
											sResult += " 0 ";
											thisTestResult=false;
										}
									}
								}
								else { 
									if ( operator.equals("<") ) {
										if ( isNumeric(lhs) && isNumeric(rhs) ) {
											if ( new Float(lhs).floatValue() < new Float(rhs).floatValue() ) {
												sResult += " 1 ";
												thisTestResult=true;
											}
											else {
												sResult += " 0 ";
												thisTestResult=false;
											}
										}
										else {
											if ( res<0 ) {
												sResult += " 1 ";
												thisTestResult=true;
											}
											else {
												sResult += " 0 ";
												thisTestResult=false;
											}
										}
									}
									else {
										//outputRecordSet.addWarningMessage("[6763]Seeing "+operator);
										if ( operator.equalsIgnoreCase("or") ) {
											if ( lhs.equals("1") || rhs.equals("1") ) {
												sResult += " 1 ";
												thisTestResult=true;
											}
											else {
												sResult += " 0 ";
												thisTestResult=false;
											}
										}
										else {
											if ( operator.equalsIgnoreCase("and") ) {
												if ( lhs.equals("1") && rhs.equals("1") ) {
													sResult += " 1 ";
													thisTestResult=true;
												}
												else {
													sResult += " 0 ";
													thisTestResult=false;
												}
											}
											else {
												if ( operator.equals("") ) {
													sResult += " "+lhs+" ";
												}
												else {
													logManager.error(nestingIndent+":DTP-I(3046): Did not recognise Test \""+operator+"\" during "+seen+" (sTrace="+sTrace1+")");
													throw new VermilionNonFatalException("Did not recognise Test '"+operator+"'",1);
												}
											}
										}
									}
								}
							}			
						}
						
						runPos=nestingIndent+":dtpi3078";
						// andor is from the previous read (i.e. we've just seen "2=3" from the "1=2 OR 2=3")
						thisTest0Result=thisTestResult;
						if ( prevTestResult.equals("") ) {
							prevTestRes=thisTestResult;
						}
						else {
							if ( prevTestResult.equals("1") ) {
								prevTestRes=true;
							}
							else {
								prevTestRes=false;
							}
						}
						
						if ( andor.equalsIgnoreCase("and") ) {
							if ( prevTestRes && thisTestResult ) {
								thisTestResult=true;
							}
							else {
								thisTestResult=false;
							}
						}
						else {
							if ( andor.equalsIgnoreCase("or") ) {
								if ( prevTestRes || thisTestResult ) {
									thisTestResult=true;
								}
								else {
									thisTestResult=false;
								}
							}
							else {
								if ( !andor.equals("")) {
									logManager.error(nestingIndent+":DTP-I(3073)Did not recognise '"+andor+"' (expecting AND or OR) (sTrace="+sTrace1+")"); // see below for andor
									throw new VermilionNonFatalException("Did not recognise '"+andor+"' (expecting AND or OR)",1);
								}
							}
						}
						
						if ( ltrace ) logManager.error(nestingIndent+":DTP-I(3075)"+seen);
						if ( ltrace ) logManager.error(nestingIndent+":DTP-I(3077)"+prevlhsrhs+" ... "+andor+" ... "+lhs+" "+operator+" "+rhs+" ("+prevTestResult+" "+andor+" "+thisTest0Result+"="+thisTestResult+")");
						prevlhsrhs    +=" "+lhs+" "+operator+" "+rhs;
						prevTestResult =""+thisTestResult;
					
						runPos=nestingIndent+":dtpi3107";
						if ( results.size() <= i ) {
							keepParsing=false;
						}
						else {
							andor=results.get(i++).trim();
							sResultTrace=sResultTrace+"{g:"+andor+"}";
		
							logManager.debug(nestingIndent+":3135:Read "+andor);
							seen+=" "+andor;
							sResult+=andor;
							if ( ltrace) logManager.error(nestingIndent+":DTP-I(3088) '"+sResult+"' = '"+prevTestResult+"' vs '"+thisTestResult+"'");
							//throw new VermilionFatalException(sResult+"="+prevTest+" vs "+thisTest,1);
						}
					} // while (keepParsing) {
		
					sResultTrace=sResultTrace+" == ["+sResult.trim()+"]";
					// Boil down the results so far....
					logManager.debug(nestingIndent+":******************************** Preresult is "+sResult);
					sResult=sResult.trim().toLowerCase();
					String sResult0="";
					if ( ltrace || ltrace2 ) outputRecordSet.addWarningMessage("\n"+nestingIndent+":[6865]evaluating ["+part+"]=["+sResult+"]");
					while ( !sResult.equals(sResult0) ) {
						sResult0=sResult.trim();
						sResult=sResult.trim();
						sResult=sResult.replaceAll("  ", " ");
						sResult=sResult.replaceAll("1 and 1", "1");
						sResult=sResult.replaceAll("1 and 0", "0");
						sResult=sResult.replaceAll("0 and 1", "0");
						sResult=sResult.replaceAll("0 and 0", "0");
			
						sResult=sResult.replaceAll("1 or 1", "1");
						sResult=sResult.replaceAll("1 or 0", "1");
						sResult=sResult.replaceAll("0 or 1", "1");
						if ( ltrace || ltrace2 ) outputRecordSet.addWarningMessage(nestingIndent+":............................................................... ["+sResult+"] vs previously ["+sResult0+"]");
					}
					runPos=nestingIndent+":dtpi3118";
					logManager.debug(nestingIndent+":******************************** Result is "+sResult);
					
					if ( recurse ) {
						boolean keepLooping=true;
						safety=50;
						String sResult1;
						if ( !sResult.equals("1") && !sResult.equals("0") ) {
							while ( keepLooping && safety-->0 ) {
								if ( !sResult.equals("") ) {
									if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":Nesting a call for ["+sResult+"]");
									sResult1=getJoinEvaluatePart(nesting, sResult, inputRecordSetList, codes, recordsetIndex, rowindex, dstColumns, false, false);
									if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+":Nesting a call returned ["+sResult+"]");
									if ( sResult.equals(sResult1) ) {
										keepLooping=false;
									}
									sResult=sResult1;
								}
								else {
									sResult="?";
									keepLooping=false;
								}
							}
						}
					}
				}
				else {
					sResult="1";
				}
			}
			catch (VermilionNonFatalException e) {
				throw new VermilionNonFatalException(e.getMessage(),1);
			}
			catch (Exception e) {
				logManager.error(nestingIndent+":(3195) "+e.toString()+" (at err"+runPos+")");
				throw new VermilionFatalException(e.toString()+" (err"+runPos+")",1);
			}

			if ( ltrace ) outputRecordSet.addWarningMessage(nestingIndent+"=="+sResultTrace+" =["+sResult.trim()+"]");
			logManager.debug(nestingIndent+":============getJoinEvaluatePart ["+part+"]");
			return sResult.trim();
		}
		
		// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
		public String getFormattedDate(String str)
		{
			String str0=str;
			
			// If it's not quoted then it's not a string....
			if ( !str.startsWith("'") || !str.endsWith("'") ) {
				return str;
			}
			
			str=str.replace("'", "");
			
			// If the string is 9999-99-99 then return 9999-99-99 00:00:00
			if ( str.matches("\\d{4}-\\d{2}-\\d{2}")) {
				str=str+" 00:00:00";
				return "'"+str+"'";
			}
			
			// if the string is 99999999 then return 9999-99-99 00:00:00
			if ( str.matches("\\d{8}")) {
				str=str.substring(0,4)+"-"+str.substring(5,2)+"-"+str.substring(7,2)+" 00:00:00";
			}
			
			// If str is a valid date then format it to yyyy-mm-dd hh:mm:ss
			boolean done=false;
			SimpleDateFormat odf=new SimpleDateFormat("yyyy-MM-dd 00:00:00");
			
			String[] formats = {"dd MMMM yyyy", "dd MMM yyyy", "dd MMMM yy", "dd MMM yy"};
			SimpleDateFormat sdf;
			Date dt;
			
			for(String tryFormat: formats) {
				sdf=new SimpleDateFormat(tryFormat);
				try {
					dt=sdf.parse(str);
					str=odf.format(dt);
					done=true;
				}
				catch (Exception e) {
					logManager.debug("----------------------------Could not format "+str+" as "+tryFormat);
				}
				if ( done ) {
					logManager.debug("----------------------------Format for "+str0+" is "+tryFormat+" = "+str);
					if ( str.matches("\\w\\d{2}:\\d{2}:\\d{2}$") ) { // if it already ends in a time then just return it
						return "'"+str+"'";
					}
					return "'"+str+"'";
				}
			}		
		
			return "'"+str+"'";
		}
		
		// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
		public static boolean isNumeric(String str)
		{
			try {
				double d = Double.parseDouble(str);
			}
			catch(NumberFormatException nfe) {
				//System.out.println("Could not convert "+str+" to double");
				return false;
			}
			return true;
		}
		
		// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
		public String getJoinResolveVariables(String part, RecordSetList inputRecordSetList, String[] codes, 
												int[] recordsetIndex, int[] rowindex, String[][] dstColumns)
			throws VermilionFatalException, VermilionNonFatalException {
			boolean ltrace=false;
			String runPos="dtpi3215";
			String sTable;
			String sColumn;
			String sSeenList="; consider using ";
			String sColNm="";
			try {
				logManager.debug(runPos);
				if ( isNumeric(part) ) {
					return part;
				}
				if ( part.startsWith("'") || part.startsWith("\"")) {
					return part;
				}
				int p=part.indexOf(".");
				if ( p>0 ) {
					runPos="dtpi3226";
					sTable=part.substring(0,p);
					if ( p < part.length() ) {
						sColumn=part.substring(p+1);
					}
					else {
						throw new VermilionFatalException("no column name for table "+sTable+" in "+part,1);
					}
					runPos="dtpi3235";
					logManager.debug("getJoinResolveVariables3120: Seen t:"+sTable+" c:"+sColumn);
					for(int c=0; c<codes.length; c++) {
						logManager.debug("getJoinResolveVariables3122: considering table "+codes[c]+" (vs t:"+sTable+")");
						runPos="dtpi3240("+c+")";
						if ( sTable.equalsIgnoreCase(codes[c])) {
							logManager.debug("getJoinResolveVariables3124: Seen table "+sTable);
							runPos="dtpi3243";
							for(int f=0; f<dstColumns[c].length; f++) {
								runPos="dtpi3245";
								sColNm=dstColumns[c][f].substring(0, dstColumns[c][f].indexOf("/"));
								logManager.debug("getJoinResolveVariables3126: considering column "+sColNm+" (vs c:"+sColumn+")");
								if ( sColumn.equalsIgnoreCase(sColNm)) {
									runPos="dtpi3239";
									logManager.debug("getJoinResolveVariables3127: Seen column "+sColumn+" ("+c+")");
									int rs=recordsetIndex[c];
									logManager.debug("getJoinResolveVariables3183: Seen column "+sColumn+" (recordset "+rs+" row "+rowindex[rs]+" column "+f+")");
									String value=inputRecordSetList.getRecordSet(rs).getRow(rowindex[rs]).getString(f);
									if ( ! isNumeric(value) ) {
										value="'"+value+"'";
									}
									logManager.debug("getJoinResolveVariables3131: Returning value "+sTable+"."+sColumn+"="+value);
									return value;
								}
								else {
									logManager.debug("3138:no");
								}
							}
						}
						else {
							logManager.debug("3143:no");
						}
					}
				}
			}
			catch (Exception e) {
				logManager.error("(3260) Unexpected error at "+runPos);
				throw new VermilionFatalException("Unexpected error "+e.toString()+" at "+runPos,1);
			}
			
			for(int c=0; c<codes.length; c++) {
				for(int f=0; f<dstColumns[c].length; f++) {
					sColNm=dstColumns[c][f].substring(0, dstColumns[c][f].indexOf("/"));
					if ( sColNm.indexOf(" ")>0 ) {
						sColNm="["+sColNm+"]";
					}
					sSeenList+="["+codes[c]+"."+sColNm+"], ";
				}
			}
			
			sSeenList=sSeenList.substring(0, sSeenList.length()-2);
			String sMsg="Did not recognise \""+part+"\""+sSeenList+".";
			logManager.error("DTP-I3427 Exception: "+sMsg);
			throw new VermilionNonFatalException(sMsg,1);
		}

		// --------------------------------------------------------------------------------------------------------
		public boolean getJoinIsColumnFieldValid(String code, String field, RecordSetList irs, String[] codes) 
			throws VermilionFatalException, VermilionNonFatalException {
			RecordSetMetaData md;
			boolean seen=false;
			String sFieldNm, sFieldList="";
			for(int rs=0; rs<irs.getRecordSetCount(); rs++) {
				if ( codes[rs].equalsIgnoreCase(code) ) {
					md=irs.getRecordSet(rs).getMetaData();
					for(int f=0; f<md.getColumnCount(); f++) {
						sFieldNm       = md.getField(f).getFieldName();
						sFieldList=sFieldList+", "+sFieldNm;
						if ( sFieldNm.equals(field) ) {
							seen=true;
						}
					}
				}
			}
			if ( seen ) {
				return true;
			}
			else {
				sFieldList=sFieldList.substring(2);
				throw new VermilionFatalException ("Did not find "+code+"."+field+" (but did see "+sFieldList+" in "+code+")",1);
			}
			
		}
		
		// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
		public RecordSet getJoinMergeRows(RecordSet outputRecordSet, RecordRow mergeThisRow, int newRow, String[] orsColumns, String[] dstColumns) // , RecordSetMetaData mdA, RecordSetMetaData mdB)
				throws VermilionFatalException, VermilionNonFatalException {
				boolean ltrace=false;
				boolean validType;
				// If columns in A match columns in B then move the values from B rowB to A rowA
				logManager.debug("(2908) getJoinMergeRows start");
				int colCountA=outputRecordSet.getColumnCount();
				int colCountB=mergeThisRow.getColumnCount();

				String runPos="";
				try {
					for(int cA=0; cA < colCountA; cA++) {
						logManager.debug("(2910)   getJoinMergeRows start rowA "+cA);
						for(int cB=0; cB<colCountB; cB++) {
							logManager.debug("(2912)    getJoinMergeRows start rowB "+cB+" from "+orsColumns[cA]+" vs "+dstColumns[cB]+";");
							if ( orsColumns[cA].equals(dstColumns[cB]) ) {
								logManager.debug("(2924)    getJoinMergeRows ("+orsColumns[cA]+" from row "+cB+")...");
								validType=false;
								
								if ( orsColumns[cA].endsWith("integer") ) {
									if ( !mergeThisRow.isEmpty(cB) ) {
										logManager.debug("(2926)    getJoinMergeRows integer from "+0+","+cB);
										runPos="merging integer (2926."+cB+")";
										runPos="merging integer (2926."+cB+"):'"+mergeThisRow.getString(cB)+"'";
										outputRecordSet.setValue(newRow, cA, mergeThisRow.getString(cB));
									}
									validType=true;
								}
								if ( orsColumns[cA].endsWith("decimal")) {
									if ( !mergeThisRow.isEmpty(cB) ) {
										logManager.debug("(2930)    getJoinMergeRows decimal from "+0+","+cB);
										runPos="merging decimal (2930."+cB+")";
										runPos="merging decimal (2930."+cB+"):'"+mergeThisRow.getString(cB)+"'";
										runPos="merging decimal (2930."+cB+"):'"+mergeThisRow.getDecimal(cB).toPlainString()+"'";
										outputRecordSet.setValue(newRow, cA, mergeThisRow.getDecimal(cB).toPlainString());
									}
									validType=true;
								}
								if ( orsColumns[cA].endsWith("datetime") ) {
									if ( !mergeThisRow.isEmpty(cB) ) {
										logManager.debug("(2934)    getJoinMergeRows datetime from "+0+","+cB);
										runPos="merging datetime (2934."+cB+")";
										runPos="merging datetime (2934."+cB+"):'"+mergeThisRow.getString(cB)+"'";
										outputRecordSet.setValue(newRow, cA, mergeThisRow.getString(cB));
									}
									validType=true;
								}
								if ( orsColumns[cA].endsWith("html") ) {
									if ( !mergeThisRow.isEmpty(cB) ) {
										logManager.debug("(2938)    getJoinMergeRows html from "+0+","+cB);
										runPos="merging html (2938."+cB+")";
										runPos="merging html (2938."+cB+"):'"+mergeThisRow.getString(cB)+"'";
										outputRecordSet.setValue(newRow, cA, mergeThisRow.getString(cB));
									}
									validType=true;
								}
								if ( orsColumns[cA].endsWith("string") ) {
									if ( !mergeThisRow.isEmpty(cB) ) {
										logManager.debug("(2942)    getJoinMergeRows string from "+0+","+cB);
										runPos="merging string (2942."+cB+")";
										runPos="merging string (2942."+cB+"):'"+mergeThisRow.getString(cB)+"'";
										outputRecordSet.setValue(newRow, cA, mergeThisRow.getString(cB));
									}
									validType=true;
								}
								if ( orsColumns[cA].endsWith("boolean") ) {
									if ( !mergeThisRow.isEmpty(cB) ) {
										logManager.debug("(2946)    getJoinMergeRows boolean from "+0+","+cB);
										runPos="merging boolean (2946."+cB+")";
										runPos="merging boolean (2946."+cB+"):'"+mergeThisRow.getString(cB)+"'";
										outputRecordSet.setValue(newRow, cA, mergeThisRow.getString(cB));
									}
									validType=true;
								}
								if ( ! validType ) {
									outputRecordSet.addWarningMessage("DTP join instruction did not recognise field type "+orsColumns[cA]);
									throw new VermilionFatalException("DTP join instruction did not recognise field type "+orsColumns[cA],1);
								}
							}
						}
					}
				}
				catch (Exception e) {
					throw new VermilionFatalException ("Problems performing getJoinMergeRows for "+runPos,1);
				}
				logManager.debug("(2941) getJoinMergeRows fin");
				return outputRecordSet;
			}
		


	// -----------------------------------------------------------------------------------------------------------------------------------------------------------------
	
}
